           (*-------------------------------------------*
            |        CSP-Prover on Isabelle2004         |
            |               November 2004               |
            |                   June 2005  (modified)   |
            |                                           |
            |        CSP-Prover on Isabelle2005         |
            |                October 2005  (modified)   |
            |                                           |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_syntax = Infra:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  Union (B ` A) = (UN x:A. B x)                      *)
(*                  Inter (B ` A) = (INT x:A. B x)                     *)

declare Union_image_eq [simp del]
declare Inter_image_eq [simp del]

(*-----------------------------------------------------------*
 |                                                           |
 |    Process Type Definitions                               |
 |                                                           |
 |             'a proc : type of process expressions         |
 |                       'a : event                          |
 |                                                           |
 *-----------------------------------------------------------*)

(*********************************************************************
                       process expression
 *********************************************************************)

datatype 
 'a proc
    = STOP

    | SKIP

    | DIV

    | Act_prefix     "'a" "'a proc"      ("(1_ /-> _)" [150,80] 80)

    | Ext_pre_choice "'a set" "'a => 'a proc"
                                               ("(1? :_ /-> _)" [900,80] 80)
    | Ext_choice     "'a proc" "'a proc"  
                                               ("(1_ /[+] _)" [72,73] 72)
    | Int_choice     "'a proc" "'a proc"  
                                               ("(1_ /|~| _)" [64,65] 64)
    | Rep_int_choice "(('a,'a set, nat) tri_type) set"
                     "(('a,'a set, nat) tri_type) => 'a proc"
                                               ("(1!! :_ .. /_)" [900,68] 68) 
    | Conditional    "bool" "'a proc" "'a proc"
                                 ("(0IF _ /THEN _ /ELSE _)" [900,60,60] 88)
    | Parallel       "'a proc" "'a set" "'a proc"  
                                               ("(1_ /|[_]| _)" [76,0,77] 76)
    | Hiding         "'a proc" "'a set"   ("(1_ /-- _)" [84,85] 84)

    | Renaming       "'a proc" "('a * 'a) set"
                                               ("(1_ /[[_]])" [84,0] 84)
    | Seq_compo      "'a proc" "'a proc"  
                                               ("(1_ /;; _)" [79,78] 78)

(*--------------------------------------------------------------------*
 |                                                                    |
 | (1) binding power:                                                 |
 |      Proc_name > Conditional > {Hiding, Renaming} >                |
 |      {Act_prefix, Ext_pre_choice} > Seq_compo > Parallel >         |
 |      Ext_choice > Rep_int_choice > int_choice                      |
 |                                                                    |
 |   The binding power 90 is higher than ones of processes ops.       |
 |                                                                    |
 | (2) DIV is a short notation as follows: DIV = ! {} .. (%x. STOP)   |
 |                                                                    |
 | (3) About ("(1_ /-> _)" [150,80] 80)                               |
 |      The nth operator ! has the power 101.                         |
 |      So, 150 > 101 is needed.                                      |
 |                                                                    |
 | (4) natural number is defaultly available in replicated internal   |
 |     choice. it is necessary for defining fixed points.             |
 |                                                                    |
 *--------------------------------------------------------------------*)

(*** external prefix ***)

syntax
  "@Ext_pre_choice"  :: "pttrn => 'a set => 'a proc 
                => 'a proc"  ("(1? _:_ /-> _)" [900,900,80] 80)

translations
  "? x:X -> P"  == "? :X -> (%x. P)"

(*** internal choice ***)

consts
  Rep_int_choice_fun :: "('b => (('a,'a set, nat) tri_type))
                         => 'b set => ('b => 'a proc) => 'a proc"
                                      ("(1!!<_> :_ .. /_)" [0,900,68] 68)
defs
  Rep_int_choice_fun_def : 
        "!!<f> :X .. Pf == !! :(f ` X) .. (%x. (Pf ((inv f) x)))"

(*** internal choice ***)

syntax
  "@Rep_int_choice":: 
      "pttrn => (('a,'a set, nat) tri_type) set => 'a proc => 'a proc"  
                               ("(1!! _:_ .. /_)" [900,900,68] 68)
  "@Rep_int_choice_UNIV":: 
      "pttrn => 'a proc => 'a proc"  
                               ("(1!! _ .. /_)" [900,68] 68)

  "@Rep_int_choice_fun":: 
  "('b => (('a,'a set, nat) tri_type)) => pttrn => 'b set => 'a proc => 'a proc"
                               ("(1!!<_> _:_ .. /_)" [0,900,900,68] 68)

  "@Rep_int_choice_UNIV_fun":: 
  "('b => (('a,'a set, nat) tri_type)) => pttrn => 'a proc => 'a proc"
                               ("(1!!<_> _ .. /_)" [0,900,68] 68)

translations
  "!! x:X .. P"    == "!! :X .. (%x. P)"
  "!! x .. P"      == "!! x:UNIV .. P"
  "!!<f> x:X .. P" == "!!<f> :X .. (%x. P)"
  "!!<f> x .. P"   == "!!<f> x:UNIV .. P"

syntax
  "@Rep_int_choice_fun_Tri1" :: "'a set => ('a => 'a proc) => 'a proc"
                                      ("(1! :_ .. /_)" [900,68] 68)
  "@Rep_int_choice_fun_Tri2" :: "('a set) set => ('a set => 'a proc) => 'a proc"
                                      ("(1!set :_ .. /_)" [900,68] 68)
  "@Rep_int_choice_fun_Tri3" :: "nat set => (nat => 'a proc) => 'a proc"
                                      ("(1!nat :_ .. /_)" [900,68] 68)

translations
           "! :X .. Pf" == "!!<Tri1> :X .. Pf"
        "!set :X .. Pf" == "!!<Tri2> :X .. Pf"
        "!nat :X .. Pf" == "!!<Tri3> :X .. Pf"

consts
  Rep_int_choice_f :: "('b => 'a) => 'b set => ('b => 'a proc) => 'a proc"
                                      ("(1!<_> :_ .. /_)" [0,900,68] 68)

defs
 Rep_int_choice_f_def : 
        "!<f> :X .. Pf == ! :(f ` X) .. (%x. (Pf ((inv f) x)))"

syntax
  "@Rep_int_choice_Tri1" :: "pttrn => 'a set => ('a => 'a proc) => 'a proc"
                                      ("(1! _:_ .. /_)" [900,900,68] 68)
  "@Rep_int_choice_Tri2" :: "pttrn => ('a set) set => ('a set => 'a proc) => 'a proc"
                                      ("(1!set _:_ .. /_)" [900,900,68] 68)
  "@Rep_int_choice_Tri3" :: "pttrn => nat set => (nat => 'a proc) => 'a proc"
                                      ("(1!nat _:_ .. /_)" [900,900,68] 68)
  "@Rep_int_choice_compo":: "('b => 'a) => pttrn => 'b set => 'a proc => 'a proc"
                                      ("(1!<_> _:_ .. /_)" [0,900,900,68] 68)

translations
       "! x:X .. P" == "! :X .. (%x. P)"
    "!set x:X .. P" == "!set :X .. (%x. P)"
    "!nat x:X .. P" == "!nat :X .. (%x. P)"
    "!<f> x:X .. P" == "!<f> :X .. (%x. P)"

syntax
  "@Rep_int_choice_Tri1_UNIV" :: "pttrn => ('a => 'a proc) => 'a proc"
                                      ("(1! _ .. /_)" [900,68] 68)
  "@Rep_int_choice_Tri2_UNIV" :: "pttrn => ('a set => 'a proc) => 'a proc"
                                      ("(1!set _ .. /_)" [900,68] 68)
  "@Rep_int_choice_Tri3_UNIV" :: "pttrn => (nat => 'a proc) => 'a proc"
                                      ("(1!nat _ .. /_)" [900,68] 68)
  "@Rep_int_choice_compo_UNIV":: "('b => 'a) => pttrn => 'a proc => 'a proc"
                                      ("(1!<_> _ .. /_)" [900,900,68] 68)

translations
         "! x .. P" == "! x:UNIV .. P"
      "!set x .. P" == "!set x:UNIV .. P"
      "!nat x .. P" == "!nat x:UNIV .. P"
      "!<f> x .. P" == "!<f> x:UNIV .. P"

(*** internal prefix choice ***)

consts
  Int_pre_choice :: "'a set => ('a => 'a proc) => 'a proc"  
                                      ("(1! :_ /-> _)" [900,80] 80)

defs
  Int_pre_choice_def : "! :X -> Pf == ! :X .. (%x. x -> (Pf x))"

syntax
  "@Int_pre_choice"  :: "pttrn => 'a set => 'a proc 
                         => 'a proc"  ("(1! _:_ /-> _)" [900,900,80] 80)

  "@Int_pre_choice_UNIV" ::  "pttrn => 'a proc => 'a proc"
                                           ("(1! _ /-> _)" [900,80] 80)

translations
  "! x:X -> P"  == "! :X -> (%x. P)"
  "! x -> P"    == "! x:UNIV -> P"

(*** sending and receiving prefixes ***)

consts
  Send_prefix :: "('x => 'a) => 'x => 'a proc => 'a proc" 
                                      ("(1_ ! _ /-> _)" [900,1000,80] 80)
  Nondet_send_prefix :: "('x => 'a) => 'x set => 
                         ('x => 'a proc) => 'a proc" 
  Rec_prefix  :: "('x => 'a) => 'x set => 
                  ('x => 'a proc) => 'a proc" 

defs
  Send_prefix_def : 
    "a ! x -> P    == a x -> P"
  Nondet_send_prefix_def : 
    "Nondet_send_prefix f X Pf == ! :(f ` X) -> (%x. (Pf ((inv f) x)))"
  Rec_prefix_def : 
    "Rec_prefix f X Pf == ? :(f ` X) -> (%x. (Pf ((inv f) x)))"

syntax
  "@Nondet_send_prefix" :: 
     "('x => 'a) => pttrn => 'x set => 'a proc => 'a proc"
                               ("(1_ !? _:_ /-> _)" [900,900,1000,80] 80)

  "@Nondet_send_prefix_UNIV" :: 
     "('x => 'a) => pttrn => 'a proc => 'a proc"
                               ("(1_ !? _ /-> _)" [900,900,80] 80)

  "@Rec_prefix" :: 
     "('x => 'a) => pttrn => 'x set => 'a proc => 'a proc"
                               ("(1_ ? _:_ /-> _)" [900,900,1000,80] 80)
  "@Rec_Prefix_UNIV" :: 
     "('x => 'a) => pttrn => 'a proc => 'a proc"
                               ("(1_ ? _ /-> _)" [900,900,80] 80)

translations
  "a !? x:X -> P" == "Nondet_send_prefix a X (%x. P)"
  "a !? x -> P"   == "a !? x:UNIV -> P"

  "a ? x:X -> P"  == "Rec_prefix a X (%x. P)"
  "a ? x -> P"    == "a? x:UNIV -> P"

(*** unfolding syntactic sugar ***)

lemmas csp_prefix_ss_def = Int_pre_choice_def Send_prefix_def Nondet_send_prefix_def
                           Rec_prefix_def

(*** parallel ***)

syntax
  "_Interleave"    :: "'a proc => 'a proc
                    => 'a proc"  ("(1_ /||| _)" [76,77] 76)
  "_Synchro"       :: "'a proc => 'a proc
                    => 'a proc"  ("(1_ /|| _)" [76,77] 76)

translations
  "P ||| Q"     == "P |[{}]| Q"
  "P || Q"      == "P |[UNIV]| Q"

consts
  Alpha_parallel :: "'a proc => 'a set => 'a set => 'a proc
                      => 'a proc"  ("(1_ /(2|[_,/_]|)/ _)" [76,0,0,77] 76)
defs
  Alpha_parallel_def :
   "P |[X,Y]| Q  == (P |[- X]| SKIP) |[X Int Y]| (Q |[- Y]| SKIP)"

consts
  Rep_parallel :: "('a proc * 'a set) list
                   => 'a proc"  ("(1[||] _)" [77] 77)

primrec
  "[||] [] = SKIP"
  "[||] (PX # PXs) = (fst PX) |[snd PX, Union (snd ` set PXs)]| ([||] PXs)"

consts
  Index_parallel :: "'i set => ('i => ('a proc * 'a set))
                      => 'a proc"  ("(1[||]:_ /_)" [90,77] 77)

defs
  Index_parallel_def :
   "[||]:I PXf == [||] (map PXf (SOME Is. Is isListOf I))"

syntax
  "@Index_parallel" :: 
     "pttrn => 'i set => ('a proc * 'a set)
                      => 'a proc"  ("(1[||] _:_ .. /_)" [900,90,77] 77)

translations
  "[||] i:I .. PX"  == "[||]:I (%i. PX)"

(************************************
 |           empty Index            |
 ************************************)

lemma Index_parallel_empty[simp]: "I = {} ==> [||]:I PXf = SKIP"
by (simp add: Index_parallel_def)

(************************************
 |            one Index             |
 ************************************)

lemma Index_parallel_one: 
  "I = {i} ==> [||]:I PXf = (fst (PXf i)) |[snd (PXf i), {}]| SKIP"
by (simp add: Index_parallel_def)

(*** timeout ***)

syntax
  "_Timeout"  :: "'a proc => 'a proc 
               => 'a proc"  ("(1_ /[> _)" [73,74] 73)
translations
  "P [> Q"      == "(P |~| STOP) [+] Q"

(*------------------------------------*
 |                                    |
 |              X-Symbols             |
 |                                    |
 *------------------------------------*)

syntax (output)
  "_PrefixX"         :: "'a => 'a proc => 'a proc"
                                   ("(1_ /-> _)" [150,80] 80)
  "_Ext_pre_choiceX" :: "pttrn => 'a set => 'a proc => 'a proc"  
                                   ("(1? _:_ /-> _)" [900,900,80] 80)
  "_Ext_choiceX"     :: "'a proc => 'a proc => 'a proc"
                                   ("(1_ /[+] _)" [72,73] 72)
  "_Int_choiceX"     :: "'a proc => 'a proc => 'a proc"
                                   ("(1_ /|~| _)" [64,65] 64)
  "_Int_pre_choiceX" :: "pttrn => 'a set => 'a proc => 'a proc"  
                                   ("(1! _:_ /-> _)" [900,900,80] 80)

syntax (xsymbols)
  "_PrefixX"         :: "'a => 'a proc => 'a proc"
                          ("(1_ /\<rightarrow> _)" [150,80] 80)
  "_Ext_pre_choiceX" :: "pttrn => 'a set => 'a proc => 'a proc"  
                          ("(1? _:_ /\<rightarrow> _)" [900,900,80] 80)
  "_Ext_choiceX"     :: "'a proc => 'a proc => 'a proc"
                          ("(1_ /\<box> _)" [72,73] 72)
  "_Int_choiceX"     :: "'a proc => 'a proc => 'a proc"
                          ("(1_ /\<sqinter> _)" [64,65] 64)

  "_Int_pre_choiceX" :: "pttrn => 'a set => 'a proc => 'a proc"  
                          ("(1! _:_ /\<rightarrow> _)" [900,900,80] 80)

translations
  "a \<rightarrow> P"     == "a -> P"
  "? x:X \<rightarrow> P" == "? x:X -> P"
  "P \<box> Q"            == "P [+] Q"
  "P \<sqinter> Q"        == "P |~| Q"
  "! x:X \<rightarrow> P" == "! x:X -> P"

syntax (output)
  "_Nondet_send_prefixX"      :: "('x => 'a) => pttrn => 'x set => 
                                  'a proc => 'a proc"
                                  ("(1_ !? _:_ /-> _)" [900,900,1000,80] 80)
  "_Nondet_send_prefix_UNIVX" :: "('x => 'a) => pttrn => 'a proc => 'a proc"
                               ("(1_ !? _ /-> _)" [900,900,80] 80)

  "_Send_prefixX"     :: "('x => 'a) => 'x => 'a proc => 'a proc" 
                          ("(1_ ! _ /-> _)" [900,1000,80] 80)

  "_Rec_prefixX"      :: "('x => 'a) => pttrn => 'x set => 
                          'a proc => 'a proc"
                          ("(1_ ? _:_ /-> _)" [900,900,1000,80] 80)
  "_Rec_prefix_UNIVX" :: "('x => 'a) => pttrn =>
                          'a proc => 'a proc"
                          ("(1_ ? _ /-> _)" [900,900,80] 80)

syntax (xsymbols)
  "_Nondet_send_prefixX"      :: "('x => 'a) => pttrn => 'x set => 
                                  'a proc => 'a proc"
                               ("(1_ !? _:_ /\<rightarrow> _)" [900,900,1000,80] 80)
  "_Nondet_send_prefix_UNIVX" :: "('x => 'a) => pttrn => 'a proc => 'a proc"
                               ("(1_ !? _ /\<rightarrow> _)" [900,900,80] 80)

  "_Send_prefixX"     :: "('x => 'a) => 'x => 'a proc => 'a proc" 
                          ("(1_ ! _ /\<rightarrow> _)" [900,1000,80] 80)

  "_Rec_prefixX"      :: "('x => 'a) => pttrn => 'x set => 
                          'a proc => 'a proc"
                          ("(1_ ? _:_ /\<rightarrow> _)" [900,900,1000,80] 80)
  "_Rec_prefix_UNIVX" :: "('x => 'a) => pttrn =>
                          'a proc => 'a proc"
                          ("(1_ ? _ /\<rightarrow> _)" [900,900,80] 80)

translations
  "a !? x:X \<rightarrow> P" == "a !? x:X -> P"
  "a !? x \<rightarrow> P"   == "a !? x:UNIV \<rightarrow> P"

  "a ! x \<rightarrow> P"    == "a ! x -> P"
  "a ? x:X \<rightarrow> P"  == "a ? x:X -> P"
  "a ? x \<rightarrow> P"    == "a ? x:UNIV \<rightarrow> P"

syntax (output)
  "_Proc_nameX" :: "'n => 'a proc" ("<_>" [0] 90)
  "_TimeoutX"   :: "'a proc => 'a proc 
                 => 'a proc"  ("(1_ /[> _)" [73,74] 73)
  "_HidingX"    :: "'a proc => 'a set => 'a proc"  
                                        ("(1_ /-- _)" [84,85] 84)
  "_RenamingX"  :: "'a proc => ('a * 'a) set => 'a proc"  
                                        ("(1_ /[[_]])" [84,0] 84)

syntax (xsymbols)
  "_Proc_nameX" :: "'n => 'a proc" ("\<langle>_\<rangle>" [0] 90)
  "_TimeoutX"   :: "'a proc => 'a proc 
                 => 'a proc"  ("(1_ /\<rhd> _)" [73,74] 73)
  "_HidingX"    :: "'a proc => 'a set => 'a proc"  
                                   ("(1_ /\<midarrow> _)" [84,85] 84)
  "_RenamingX"  :: "'a proc => ('a * 'a) set => 'a proc"  
                                        ("(1_ /\<lbrakk>_\<rbrakk>)" [84,0] 84)

translations
  "P \<rhd> Q"            == "(P |~| STOP) [+] Q"
  "P \<midarrow> X"       == "P -- X"
  "P \<lbrakk>r\<rbrakk>" == "P [[r]]"

(*********************************************************************
             recursive process (fixed point) and
         process expression (process with variable)
 *********************************************************************)

types ('pn,'a) procfun = "('pn => 'a proc) => 'a proc"
types ('pn,'a) procFun = "('pn => 'a proc) => ('pn => 'a proc)"
types ('pn,'a) procDef = "(('pn => 'a proc) * 'pn) => 'a proc"

consts
 FIXn :: "nat => ('pn,'a) procFun => ('pn => 'a proc)"  ("FIX[_] _" [0,1000] 55)
 FIX  :: "('pn,'a) procFun => ('pn => 'a proc)"  ("FIX _" [1000] 55)

defs
 FIXn_def :  "FIX[n] PF == (PF^n) (%pm. DIV)"
 FIX_def  :     "FIX PF == (%pn. (!nat n .. ((FIX[n] PF) pn)))"

(*** Procx ***)

consts
  Procfun :: "('pn,'a) procfun set"
  ProcFun :: "('pn,'a) procFun set"

defs
  ProcFun_def: "ProcFun == {PF. (ALL pn. (%X. PF X pn) : Procfun)}"

inductive "Procfun"
intros
Procfun_pn:
  "(%x. x pn) : Procfun"
Procfun_STOP: 
  "(%x. STOP) : Procfun"
Procfun_SKIP: 
  "(%x. SKIP) : Procfun"
Procfun_DIV:
  "(%x. DIV) : Procfun"
Procfun_Act_prefix:
  "Pf : Procfun ==> (%x. a -> (Pf x)) : Procfun"
Procfun_Ext_pre_choice:
  "[| ALL a. Pff a : Procfun |] ==> (%x. ? a:X -> (Pff a x)) : Procfun"
Procfun_Ext_choice:
  "[| Pf : Procfun ; Qf : Procfun |] ==> (%x. (Pf x) [+] (Qf x)) : Procfun"
Procfun_Int_choice:
  "[| Pf : Procfun ; Qf : Procfun |] ==> (%x. (Pf x) |~| (Qf x)) : Procfun"
Procfun_Rep_int_choice_Tri:
  "[| ALL a. Pff a : Procfun |] ==> (%x. !! a:X .. (Pff a x)) : Procfun"
Procfun_Conditional:
  "[| Pf : Procfun ; Qf : Procfun |] ==> (%x. IF b THEN (Pf x) ELSE (Qf x)) : Procfun"
Procfun_Parallel:
  "[| Pf : Procfun ; Qf : Procfun |] ==> (%x. (Pf x) |[X]| (Qf x)) : Procfun"
Procfun_Hiding:
  "Pf : Procfun ==> (%x. (Pf x) -- X) : Procfun"
Procfun_Renaming:
  "Pf : Procfun ==> (%x. (Pf x) [[r]]) : Procfun"
Procfun_Seq_compo:
  "[| Pf : Procfun ; Qf : Procfun |] ==> (%x. (Pf x) ;; (Qf x)) : Procfun"

lemma Procfun_Rep_int_choice_fun:
  "[| ALL a. Pff a : Procfun |] 
   ==> (%x. !!<f> a:X .. (Pff a x)) : Procfun"
by (simp add: Rep_int_choice_fun_def Procfun_Rep_int_choice_Tri)

lemma Procfun_Rep_int_choice_f:
  "[| ALL a. Pff a : Procfun |] 
   ==> (%x. !<f> a:X .. (Pff a x)) : Procfun"
apply (simp add: Rep_int_choice_f_def)
apply (rule Procfun_Rep_int_choice_fun)
by (auto)

lemmas Procfun_Rep_int_choice = Procfun_Rep_int_choice_Tri
                                Procfun_Rep_int_choice_fun
                                Procfun_Rep_int_choice_f

lemmas Procfun_basic_def = 
  Procfun_pn Procfun_STOP  Procfun_SKIP        Procfun_DIV
  Procfun_Act_prefix   Procfun_Ext_pre_choice   Procfun_Ext_choice 
  Procfun_Int_choice   Procfun_Rep_int_choice   Procfun_Conditional
  Procfun_Parallel     Procfun_Hiding           Procfun_Renaming
  Procfun_Seq_compo

(* test
lemma Procfun_basic_def_test:
  "[| ALL a. Pff a : Procfun |] ==> (%x. ! a:X .. (Pff a x)) : Procfun"
by (simp add: Procfun_basic_def)
*)

(*------------------------------------------------------*
 |  "Procfun.induct" is available.                      |
 |                                                      |
 |      apply (rule Procfun.induct[of P PR])            |
 |                                                      |
 *------------------------------------------------------*)

lemma compo_in_Procfun1:
  "[| Pf : Procfun ; QF : ProcFun |] ==> (%x. Pf (QF x)) : Procfun"
apply (rule Procfun.induct[of Pf "(%Pf. (%x. Pf (QF x)) : Procfun)"])
apply (simp)
apply (simp add: ProcFun_def)
apply (simp_all add: Procfun_basic_def)
done

lemma compo_in_Procfun2:
  "[| Pf : Procfun ; QF : ProcFun |] ==> Pf o QF : Procfun"
by (simp add: comp_def compo_in_Procfun1)

lemmas compo_in_Procfun = compo_in_Procfun1 compo_in_Procfun2

lemma compo_in_Procfun_ALL:
  "ALL Pf QF. (Pf : Procfun & QF : ProcFun ) --> (%x. Pf (QF x)) : Procfun"
by (simp add: compo_in_Procfun)

lemma compo_in_ProcFun1:
  "[| PF : ProcFun ; QF : ProcFun |] ==> (%x. PF (QF x)): ProcFun"
apply (simp add: ProcFun_def)
apply (intro allI)
apply (drule_tac x="pn" in spec)
apply (insert compo_in_Procfun_ALL)
apply (drule_tac x="(%X. PF X pn)" in spec)
apply (drule_tac x="QF" in spec)
apply (simp add: ProcFun_def)
done

lemma compo_in_ProcFun2:
  "[| PF : ProcFun ; QF : ProcFun |] ==> PF o QF : ProcFun"
by (simp add: comp_def compo_in_ProcFun1)

lemmas compo_in_ProcFun = compo_in_ProcFun1 compo_in_ProcFun2

lemma iteration_in_ProcFun:
  "PF : ProcFun ==> PF ^ n : ProcFun"
apply (induct_tac n)
apply (simp add: ProcFun_def)
apply (simp add: Procfun_basic_def)
apply (simp add: compo_in_ProcFun)
done

lemma iteration_in_Procfun:
  "PF : ProcFun ==> (%X. (PF ^ n) X pn) : Procfun"
apply (insert iteration_in_ProcFun[of PF n])
apply (simp)
apply (simp add: ProcFun_def)
done

(*------------------------------------------------------*
 |                                                      |
 |            single recursion by MU operator           |
 |                                                      |
 *------------------------------------------------------*)

datatype 
 mupn = MUpn

consts
  MUX :: "('a proc => 'a proc) => 'a proc"

defs
  MUX_def :
   "MUX PX == (FIX (%x. (%n. (PX (x MUpn))))) MUpn"

syntax
  "@MU"  :: "pttrn => 'a proc => 'a proc"  ("(1MU _ .. _)" [900,62] 62)

translations
  "MU X .. P"  == "MUX (%X. P)"

consts
  ProcX :: "('a proc => 'a proc) set"

defs
  ProcX_def: "ProcX == {PX. (%X. PX (X MUpn)) : Procfun}"

(*** Procfun lemma ***)

lemma Procfun_MU:
  "(ALL f:Procfun. (%X. (PXY X (f X))) : Procfun) ==> (%X. MUX (PXY X)) : Procfun"
apply (simp add: MUX_def FIX_def)
apply (rule Procfun_basic_def)
apply (rule allI)
apply (simp add: FIXn_def)
apply (induct_tac n)
apply (simp add: Procfun_basic_def)
apply (simp)
done

lemmas Procfun_def = Procfun_basic_def Procfun_MU

(*** Procfun test ***)

lemma "(%Z. (MU X .. a -> (MU Y .. b -> X [+] c -> Y |~| d -> Z))) : ProcX"
apply (simp add: ProcX_def)
apply (simp add: Procfun_def)
done

(*************************************************************
                   Syntactical functions
 *************************************************************)

(*** guarded (rec_proc_fun) ***)

consts
  gTickfun  :: "('pn,'a) procfun set"
  guardfun  :: "('pn,'a) procfun set"
  nohidefun :: "('pn,'a) procfun set"

inductive "gTickfun"
intros
gTickfun_STOP: 
  "(%x. STOP) : gTickfun"
gTickfun_DIV:
  "(%x. DIV) : gTickfun"
gTickfun_Act_prefix:
  "(%x. a -> (Pf x)) : gTickfun"
gTickfun_Ext_pre_choice:
  "(%x. ? a:X -> (Pff a x)) : gTickfun"
gTickfun_Ext_choice:
  "[| Pf : gTickfun ; Qf : gTickfun |] ==> (%x. (Pf x) [+] (Qf x)) : gTickfun"
gTickfun_Int_choice:
  "[| Pf : gTickfun ; Qf : gTickfun |] ==> (%x. (Pf x) |~| (Qf x)) : gTickfun"
gTickfun_Rep_int_choice_Tri:
  "[| ALL a. Pff a : gTickfun |] ==> (%x. !! a:X .. (Pff a x)) : gTickfun"
gTickfun_Conditional:
  "[| Pf : gTickfun ; Qf : gTickfun |] ==> (%x. IF b THEN (Pf x) ELSE (Qf x)) : gTickfun"
gTickfun_Parallel:
  "[| Pf : gTickfun | Qf : gTickfun |] ==> (%x. (Pf x) |[X]| (Qf x)) : gTickfun"
gTickfun_Renaming:
  "Pf : gTickfun ==> (%x. (Pf x) [[r]]) : gTickfun"
gTickfun_Seq_compo:
  "[| Pf : gTickfun | Qf : gTickfun |] 
   ==> (%x. (Pf x) ;; (Qf x)) : gTickfun"

lemmas gTickfun_basic_def = 
  gTickfun_STOP         gTickfun_DIV
  gTickfun_Act_prefix   gTickfun_Ext_pre_choice       gTickfun_Ext_choice 
  gTickfun_Int_choice   gTickfun_Rep_int_choice_Tri   gTickfun_Conditional
  gTickfun_Parallel     gTickfun_Renaming             gTickfun_Seq_compo

inductive "guardfun"
intros
guardfun_STOP: 
  "(%x. STOP) : guardfun"
guardfun_SKIP: 
  "(%x. SKIP) : guardfun"
guardfun_DIV:
  "(%x. DIV) : guardfun"
guardfun_Act_prefix:
  "(%x. a -> (Pf x)) : guardfun"
guardfun_Ext_pre_choice:
  "(%x. ? a:X -> (Pff a x)) : guardfun"
guardfun_Ext_choice:
  "[| Pf : guardfun ; Qf : guardfun |] ==> (%x. (Pf x) [+] (Qf x)) : guardfun"
guardfun_Int_choice:
  "[| Pf : guardfun ; Qf : guardfun |] ==> (%x. (Pf x) |~| (Qf x)) : guardfun"
guardfun_Rep_int_choice_Tri:
  "[| ALL a. Pff a : guardfun |] ==> (%x. !! a:X .. (Pff a x)) : guardfun"
guardfun_Conditional:
  "[| Pf : guardfun ; Qf : guardfun |] ==> (%x. IF b THEN (Pf x) ELSE (Qf x)) : guardfun"
guardfun_Parallel:
  "[| Pf : guardfun ; Qf : guardfun |] ==> (%x. (Pf x) |[X]| (Qf x)) : guardfun"
guardfun_Renaming:
  "Pf : guardfun ==> (%x. (Pf x) [[r]]) : guardfun"
guardfun_Seq_compo:
  "[| Pf : guardfun ; Pf : gTickfun | Qf : guardfun |] 
   ==> (%x. (Pf x) ;; (Qf x)) : guardfun"

lemmas guardfun_basic_def = 
  guardfun_STOP         guardfun_SKIP                guardfun_DIV
  guardfun_Act_prefix   guardfun_Ext_pre_choice      guardfun_Ext_choice 
  guardfun_Int_choice   guardfun_Rep_int_choice_Tri  guardfun_Conditional
  guardfun_Parallel     guardfun_Renaming            guardfun_Seq_compo

inductive "nohidefun"
intros
nohidefun_pn:
  "(%x. x pn) : nohidefun"
nohidefun_STOP: 
  "(%x. STOP) : nohidefun"
nohidefun_SKIP: 
  "(%x. SKIP) : nohidefun"
nohidefun_DIV:
  "(%x. DIV) : nohidefun"
nohidefun_Act_prefix:
  "[| Pf : nohidefun |] ==>  (%x. a -> (Pf x)) : nohidefun"
nohidefun_Ext_pre_choice:
  "[| ALL a. Pff a : nohidefun |] ==> (%x. ? a:X -> (Pff a x)) : nohidefun"
nohidefun_Ext_choice:
  "[| Pf : nohidefun ; Qf : nohidefun |] ==> (%x. (Pf x) [+] (Qf x)) : nohidefun"
nohidefun_Int_choice:
  "[| Pf : nohidefun ; Qf : nohidefun |] ==> (%x. (Pf x) |~| (Qf x)) : nohidefun"
nohidefun_Rep_int_choice_Tri:
  "[| ALL a. Pff a : nohidefun |] ==> (%x. !! a:X .. (Pff a x)) : nohidefun"
nohidefun_Conditional:
  "[| Pf : nohidefun ; Qf : nohidefun |] 
   ==> (%x. IF b THEN (Pf x) ELSE (Qf x)) : nohidefun"
nohidefun_Parallel:
  "[| Pf : nohidefun ; Qf : nohidefun |] ==> (%x. (Pf x) |[X]| (Qf x)) : nohidefun"
nohidefun_Renaming:
  "Pf : nohidefun ==> (%x. (Pf x) [[r]]) : nohidefun"
nohidefun_Seq_compo:
  "[| Pf : nohidefun ; Qf : nohidefun |] ==> (%x. (Pf x) ;; (Qf x)) : nohidefun"

lemmas nohidefun_basic_def = 
  nohidefun_pn
  nohidefun_STOP         nohidefun_SKIP                nohidefun_DIV
  nohidefun_Act_prefix   nohidefun_Ext_pre_choice      nohidefun_Ext_choice 
  nohidefun_Int_choice   nohidefun_Rep_int_choice_Tri  nohidefun_Conditional
  nohidefun_Parallel     nohidefun_Renaming            nohidefun_Seq_compo

(*** simp for gTickfun ***)

lemma gTickfun_pn_iff: "(%x. x pn) : gTickfun = False"
apply (rule)
apply (erule gTickfun.elims)
apply (simp_all add: expand_fun_eq)
apply (drule_tac x="(%x. SKIP)" in spec, simp)
apply (drule_tac x="(%x. SKIP)" in spec, simp)
apply (drule_tac x="(%x. SKIP)" in spec, simp)
apply (drule_tac x="(%x. SKIP)" in spec, simp)
apply (drule_tac x="(%x. SKIP)" in spec, simp)
apply (drule_tac x="(%x. SKIP)" in spec, simp)
apply (drule_tac x="(%x. SKIP)" in spec, simp)
apply (drule_tac x="(%x. SKIP)" in spec, simp)
apply (drule_tac x="(%x. SKIP)" in spec, simp)
apply (drule_tac x="(%x. SKIP)" in spec, simp)
apply (drule_tac x="(%x. SKIP)" in spec, simp)
done

lemma gTickfun_STOP_iff: "(%x. STOP) : gTickfun"
by (simp add: gTickfun_basic_def)

lemma gTickfun_SKIP_iff: "(%x. SKIP) : gTickfun = False"
apply (rule)
apply (erule gTickfun.elims)
apply (simp_all add: expand_fun_eq)
done

lemma gTickfun_DIV_iff: "(%x. DIV) : gTickfun"
by (simp add: gTickfun_basic_def)

lemma gTickfun_Act_prefix_iff: "(%x. a -> (Pf x)) : gTickfun"
by (simp add: gTickfun_basic_def)

lemma gTickfun_Ext_pre_choice_iff: "(%x. ? a:X -> (Pff a x)) : gTickfun"
by (simp add: gTickfun_basic_def)

lemma gTickfun_Ext_choice_iff: 
  "(%x. (Pf x) [+] (Qf x)) : gTickfun = (Pf : gTickfun & Qf : gTickfun)"
apply (rule)
apply (erule gTickfun.elims)
apply (simp_all add: expand_fun_eq)
apply (subgoal_tac "Pf = Pfa & Qf = Qfa")
apply (simp)
apply (simp add: expand_fun_eq)
by (simp add: gTickfun_basic_def)

lemma gTickfun_Int_choice_iff: 
  "(%x. (Pf x) |~| (Qf x)) : gTickfun = (Pf : gTickfun & Qf : gTickfun)"
apply (rule)
apply (erule gTickfun.elims)
apply (simp_all add: expand_fun_eq)
apply (subgoal_tac "Pf = Pfa & Qf = Qfa")
apply (simp)
apply (simp add: expand_fun_eq)
by (simp add: gTickfun_basic_def)

lemma gTickfun_Rep_int_choice_Tri_iff: 
  "(%x. !! a:X .. (Pff a x)) : gTickfun = (ALL a. Pff a : gTickfun)"
apply (rule)
apply (erule gTickfun.elims)
apply (simp_all add: expand_fun_eq)
apply (subgoal_tac "Pff = Pffa")
apply (simp)
apply (simp add: expand_fun_eq)
by (simp add: gTickfun_basic_def)

lemma gTickfun_Conditional_iff:
  "(%x. IF b THEN (Pf x) ELSE (Qf x)) : gTickfun = (Pf : gTickfun & Qf : gTickfun)"
apply (rule)
apply (erule gTickfun.elims)
apply (simp_all add: expand_fun_eq)
apply (subgoal_tac "Pf = Pfa & Qf = Qfa")
apply (simp)
apply (simp add: expand_fun_eq)
by (simp add: gTickfun_basic_def)

lemma gTickfun_Parallel_iff: 
  "(%x. (Pf x) |[X]| (Qf x)) : gTickfun = (Pf : gTickfun | Qf : gTickfun)"
apply (rule)
apply (erule gTickfun.elims)
apply (simp_all add: expand_fun_eq)
apply (subgoal_tac "Pf = Pfa & Qf = Qfa")
apply (simp)
apply (simp add: expand_fun_eq)
by (simp add: gTickfun_basic_def)

lemma gTickfun_Hiding_iff: "(%x. Pf x -- X) : gTickfun = False"
apply (rule)
apply (erule gTickfun.elims)
apply (simp_all add: expand_fun_eq)
done

lemma gTickfun_Renaming_iff: 
  "(%x. (Pf x) [[r]]) : gTickfun = (Pf : gTickfun)"
apply (rule)
apply (erule gTickfun.elims)
apply (simp_all add: expand_fun_eq)
apply (subgoal_tac "Pf = Pfa")
apply (simp)
apply (simp add: expand_fun_eq)
by (simp add: gTickfun_basic_def)

lemma gTickfun_Seq_compo_iff: 
  "(%x. (Pf x) ;; (Qf x)) : gTickfun = (Pf : gTickfun | Qf : gTickfun)"
apply (rule)
apply (erule gTickfun.elims)
apply (simp_all add: expand_fun_eq)
apply (subgoal_tac "Pf = Pfa & Qf = Qfa")
apply (simp)
apply (simp add: expand_fun_eq)
by (simp add: gTickfun_basic_def)

lemmas gTickfun_iff =
  gTickfun_pn_iff  gTickfun_SKIP_iff  gTickfun_STOP_iff       gTickfun_DIV_iff
  gTickfun_Act_prefix_iff   gTickfun_Ext_pre_choice_iff       gTickfun_Ext_choice_iff 
  gTickfun_Int_choice_iff   gTickfun_Rep_int_choice_Tri_iff   gTickfun_Conditional_iff
  gTickfun_Parallel_iff     gTickfun_Hiding_iff
  gTickfun_Renaming_iff     gTickfun_Seq_compo_iff

declare gTickfun_iff [simp]

(*** simp for guardfun ***)

lemma guardfun_pn_iff: "(%x. x pn) : guardfun = False"
apply (rule)
apply (erule guardfun.elims)
apply (simp_all add: expand_fun_eq)
apply (drule_tac x="(%x. SKIP)" in spec, simp)
apply (drule_tac x="(%x. STOP)" in spec, simp)
apply (drule_tac x="(%x. SKIP)" in spec, simp)
apply (drule_tac x="(%x. SKIP)" in spec, simp)
apply (drule_tac x="(%x. SKIP)" in spec, simp)
apply (drule_tac x="(%x. SKIP)" in spec, simp)
apply (drule_tac x="(%x. SKIP)" in spec, simp)
apply (drule_tac x="(%x. SKIP)" in spec, simp)
apply (drule_tac x="(%x. SKIP)" in spec, simp)
apply (drule_tac x="(%x. SKIP)" in spec, simp)
apply (drule_tac x="(%x. SKIP)" in spec, simp)
apply (drule_tac x="(%x. SKIP)" in spec, simp)
done

lemma guardfun_STOP_iff: "(%x. STOP) : guardfun"
by (simp add: guardfun_basic_def)

lemma guardfun_SKIP_iff: "(%x. SKIP) : guardfun"
by (simp add: guardfun_basic_def)

lemma guardfun_DIV_iff: "(%x. DIV) : guardfun"
by (simp add: guardfun_basic_def)

lemma guardfun_Act_prefix_iff: "(%x. a -> (Pf x)) : guardfun"
by (simp add: guardfun_basic_def)

lemma guardfun_Ext_pre_choice_iff: "(%x. ? a:X -> (Pff a x)) : guardfun"
by (simp add: guardfun_basic_def)

lemma guardfun_Ext_choice_iff: 
  "(%x. (Pf x) [+] (Qf x)) : guardfun = (Pf : guardfun & Qf : guardfun)"
apply (rule)
apply (erule guardfun.elims)
apply (simp_all add: expand_fun_eq)
apply (subgoal_tac "Pf = Pfa & Qf = Qfa")
apply (simp)
apply (simp add: expand_fun_eq)
by (simp add: guardfun_basic_def)

lemma guardfun_Int_choice_iff: 
  "(%x. (Pf x) |~| (Qf x)) : guardfun = (Pf : guardfun & Qf : guardfun)"
apply (rule)
apply (erule guardfun.elims)
apply (simp_all add: expand_fun_eq)
apply (subgoal_tac "Pf = Pfa & Qf = Qfa")
apply (simp)
apply (simp add: expand_fun_eq)
by (simp add: guardfun_basic_def)

lemma guardfun_Rep_int_choice_Tri_iff: 
  "(%x. !! a:X .. (Pff a x)) : guardfun = (ALL a. Pff a : guardfun)"
apply (rule)
apply (erule guardfun.elims)
apply (simp_all add: expand_fun_eq)
apply (subgoal_tac "Pff = Pffa")
apply (simp)
apply (simp add: expand_fun_eq)
by (simp add: guardfun_basic_def)

lemma guardfun_Conditional_iff:
  "(%x. IF b THEN (Pf x) ELSE (Qf x)) : guardfun = (Pf : guardfun & Qf : guardfun)"
apply (rule)
apply (erule guardfun.elims)
apply (simp_all add: expand_fun_eq)
apply (subgoal_tac "Pf = Pfa & Qf = Qfa")
apply (simp)
apply (simp add: expand_fun_eq)
by (simp add: guardfun_basic_def)

lemma guardfun_Parallel_iff: 
  "(%x. (Pf x) |[X]| (Qf x)) : guardfun = (Pf : guardfun & Qf : guardfun)"
apply (rule)
apply (erule guardfun.elims)
apply (simp_all add: expand_fun_eq)
apply (subgoal_tac "Pf = Pfa & Qf = Qfa")
apply (simp)
apply (simp add: expand_fun_eq)
by (simp add: guardfun_basic_def)

lemma guardfun_Hiding_iff: "(%x. Pf x -- X) : guardfun = False"
apply (rule)
apply (erule guardfun.elims)
apply (simp_all add: expand_fun_eq)
done

lemma guardfun_Renaming_iff: 
  "(%x. (Pf x) [[r]]) : guardfun = (Pf : guardfun)"
apply (rule)
apply (erule guardfun.elims)
apply (simp_all add: expand_fun_eq)
apply (subgoal_tac "Pf = Pfa")
apply (simp)
apply (simp add: expand_fun_eq)
by (simp add: guardfun_basic_def)

lemma guardfun_Seq_compo_iff: 
  "(%x. (Pf x) ;; (Qf x)) : guardfun
 = (Pf : guardfun & (Pf : gTickfun | Qf : guardfun))"
apply (rule)
apply (erule guardfun.elims)
apply (simp_all add: expand_fun_eq)
apply (subgoal_tac "Pf = Pfa & Qf = Qfa")
apply (simp)
apply (simp add: expand_fun_eq)
by (simp add: guardfun_basic_def)

lemmas guardfun_iff =
  guardfun_pn_iff  guardfun_SKIP_iff  guardfun_STOP_iff       guardfun_DIV_iff
  guardfun_Act_prefix_iff   guardfun_Ext_pre_choice_iff       guardfun_Ext_choice_iff 
  guardfun_Int_choice_iff   guardfun_Rep_int_choice_Tri_iff   guardfun_Conditional_iff
  guardfun_Parallel_iff     guardfun_Hiding_iff
  guardfun_Renaming_iff     guardfun_Seq_compo_iff

declare guardfun_iff [simp]

(*** simp for nohidefun ***)

lemma nohidefun_pn_iff: "(%x. x pn) : nohidefun"
by (simp add: nohidefun_basic_def)

lemma nohidefun_STOP_iff: "(%x. STOP) : nohidefun"
by (simp add: nohidefun_basic_def)

lemma nohidefun_SKIP_iff: "(%x. SKIP) : nohidefun"
by (simp add: nohidefun_basic_def)

lemma nohidefun_DIV_iff: "(%x. DIV) : nohidefun"
by (simp add: nohidefun_basic_def)

lemma nohidefun_Act_prefix_iff: 
  "(%x. a -> (Pf x)) : nohidefun = (Pf : nohidefun)"
apply (rule)
apply (erule nohidefun.elims)
apply (simp_all add: expand_fun_eq)
apply (drule_tac x="(%x. STOP)" in spec, simp)
apply (subgoal_tac "Pf = Pfa")
apply (simp)
apply (simp add: expand_fun_eq)
by (simp add: nohidefun_basic_def)

lemma nohidefun_Ext_pre_choice_iff:
  "(%x. ? a:X -> (Pff a x)) : nohidefun = (ALL a. Pff a : nohidefun)"
apply (rule)
apply (erule nohidefun.elims)
apply (simp_all add: expand_fun_eq)
apply (drule_tac x="(%x. STOP)" in spec, simp)
apply (subgoal_tac "Pff = Pffa")
apply (simp)
apply (simp add: expand_fun_eq)
by (simp add: nohidefun_basic_def)

lemma nohidefun_Ext_choice_iff: 
  "(%x. (Pf x) [+] (Qf x)) : nohidefun = (Pf : nohidefun & Qf : nohidefun)"
apply (rule)
apply (erule nohidefun.elims)
apply (simp_all add: expand_fun_eq)
apply (drule_tac x="(%x. STOP)" in spec, simp)
apply (subgoal_tac "Pf = Pfa & Qf = Qfa")
apply (simp)
apply (simp add: expand_fun_eq)
by (simp add: nohidefun_basic_def)

lemma nohidefun_Int_choice_iff: 
  "(%x. (Pf x) |~| (Qf x)) : nohidefun = (Pf : nohidefun & Qf : nohidefun)"
apply (rule)
apply (erule nohidefun.elims)
apply (simp_all add: expand_fun_eq)
apply (drule_tac x="(%x. STOP)" in spec, simp)
apply (subgoal_tac "Pf = Pfa & Qf = Qfa")
apply (simp)
apply (simp add: expand_fun_eq)
by (simp add: nohidefun_basic_def)

lemma nohidefun_Rep_int_choice_Tri_iff: 
  "(%x. !! a:X .. (Pff a x)) : nohidefun = (ALL a. Pff a : nohidefun)"
apply (rule)
apply (erule nohidefun.elims)
apply (simp_all add: expand_fun_eq)
apply (drule_tac x="(%x. STOP)" in spec, simp)
apply (subgoal_tac "Pff = Pffa")
apply (simp)
apply (simp add: expand_fun_eq)
by (simp add: nohidefun_basic_def)

lemma nohidefun_Conditional_iff:
  "(%x. IF b THEN (Pf x) ELSE (Qf x)) : nohidefun = (Pf : nohidefun & Qf : nohidefun)"
apply (rule)
apply (erule nohidefun.elims)
apply (simp_all add: expand_fun_eq)
apply (drule_tac x="(%x. STOP)" in spec, simp)
apply (subgoal_tac "Pf = Pfa & Qf = Qfa")
apply (simp)
apply (simp add: expand_fun_eq)
by (simp add: nohidefun_basic_def)

lemma nohidefun_Parallel_iff: 
  "(%x. (Pf x) |[X]| (Qf x)) : nohidefun = (Pf : nohidefun & Qf : nohidefun)"
apply (rule)
apply (erule nohidefun.elims)
apply (simp_all add: expand_fun_eq)
apply (drule_tac x="(%x. STOP)" in spec, simp)
apply (subgoal_tac "Pf = Pfa & Qf = Qfa")
apply (simp)
apply (simp add: expand_fun_eq)
by (simp add: nohidefun_basic_def)

lemma nohidefun_Hiding_iff: "(%x. Pf x -- X) : nohidefun = False"
apply (rule)
apply (erule nohidefun.elims)
apply (simp_all add: expand_fun_eq)
apply (drule_tac x="(%x. STOP)" in spec, simp)
done

lemma nohidefun_Renaming_iff: 
  "(%x. (Pf x) [[r]]) : nohidefun = (Pf : nohidefun)"
apply (rule)
apply (erule nohidefun.elims)
apply (simp_all add: expand_fun_eq)
apply (drule_tac x="(%x. STOP)" in spec, simp)
apply (subgoal_tac "Pf = Pfa")
apply (simp)
apply (simp add: expand_fun_eq)
by (simp add: nohidefun_basic_def)

lemma nohidefun_Seq_compo_iff: 
  "(%x. (Pf x) ;; (Qf x)) : nohidefun = (Pf : nohidefun & Qf : nohidefun)"
apply (rule)
apply (erule nohidefun.elims)
apply (simp_all add: expand_fun_eq)
apply (drule_tac x="(%x. STOP)" in spec, simp)
apply (subgoal_tac "Pf = Pfa & Qf = Qfa")
apply (simp)
apply (simp add: expand_fun_eq)
by (simp add: nohidefun_basic_def)

lemmas nohidefun_iff =
  nohidefun_pn_iff    nohidefun_SKIP_iff   nohidefun_STOP_iff   nohidefun_DIV_iff
  nohidefun_Act_prefix_iff   nohidefun_Ext_pre_choice_iff       nohidefun_Ext_choice_iff 
  nohidefun_Int_choice_iff   nohidefun_Rep_int_choice_Tri_iff   nohidefun_Conditional_iff
  nohidefun_Parallel_iff     nohidefun_Hiding_iff
  nohidefun_Renaming_iff     nohidefun_Seq_compo_iff

declare nohidefun_iff [simp]

(****** short notations ******)

(* gTick *)

lemma gTickfun_Rep_int_choice_fun_iff[simp]:
  "inj f ==> (%x. !!<f> a:X .. (Pff a x)) : gTickfun = (ALL a. Pff a : gTickfun)"
apply (auto simp add: Rep_int_choice_fun_def)
apply (drule_tac x="f a" in spec)
apply (simp)
done

lemma gTickfun_Rep_int_choice_f_iff[simp]:
  "inj f ==> (%x. !<f> a:X .. (Pff a x)) : gTickfun = (ALL a. Pff a : gTickfun)"
apply (simp add: Rep_int_choice_f_def)
apply (auto)
apply (drule_tac x="f a" in spec)
apply (simp)
done

lemma gTickfun_Alpha_parallel_iff[simp]:
  "(%x. (Pf x) |[X,Y]| (Qf x)) : gTickfun = (Pf : gTickfun | Qf : gTickfun)"
apply (simp add: Alpha_parallel_def)
done

lemma gTickfun_MU[simp]:
  "(ALL f: gTickfun. (%X. (PXY X (f X))) : gTickfun)
    ==> (%X. MUX (PXY X)) : gTickfun"
apply (simp add: MUX_def FIX_def)
apply (rule allI)
apply (simp add: FIXn_def)
apply (induct_tac n)
apply (simp)
apply (simp)
done

lemmas gTickfun_def = gTickfun_basic_def 
                      gTickfun_MU

(*** gTickfun test ***)

lemma "(%Z. (MU X .. a -> (MU Y .. b -> X [+] c -> Y |~| d -> Z i))) : gTickfun"
apply (simp)
done

lemma "ALL a. P a : gTickfun ==> (%x. (!set a:X .. P a x)) : gTickfun"
by (simp)

(* guard *)

lemma guardfun_Rep_int_choice_fun_iff[simp]:
  "inj f ==> (%x. !!<f> a:X .. (Pff a x)) : guardfun = (ALL a. Pff a : guardfun)"
apply (auto simp add: Rep_int_choice_fun_def)
apply (drule_tac x="f a" in spec)
apply (simp)
done

lemma guardfun_Rep_int_choice_f_iff[simp]:
  "inj f ==> (%x. !<f> a:X .. (Pff a x)) : guardfun = (ALL a. Pff a : guardfun)"
apply (simp add: Rep_int_choice_f_def)
apply (auto)
apply (drule_tac x="f a" in spec)
apply (simp)
done

lemma guardfun_Alpha_parallel_iff[simp]:
  "(%x. (Pf x) |[X,Y]| (Qf x)) : guardfun = (Pf : guardfun & Qf : guardfun)"
apply (simp add: Alpha_parallel_def)
done

lemma guardfun_MU[simp]:
  "(ALL f: guardfun. (%X. (PXY X (f X))) : guardfun)
    ==> (%X. MUX (PXY X)) : guardfun"
apply (simp add: MUX_def FIX_def)
apply (rule allI)
apply (simp add: FIXn_def)
apply (induct_tac n)
apply (simp)
apply (simp)
done

lemmas guardfun_def = guardfun_basic_def 
                      guardfun_MU

(*** guardfun test ***)

lemma "(%Z. (MU X .. a -> (MU Y .. b -> X [+] c -> Y |~| d -> Z i))) : guardfun"
apply (simp)
done

lemma "(%Z. (MU X .. a -> (MU Y .. b -> X [+] c -> Y |~| d -> Z i)) |~| Z j) : guardfun
       = False"
apply (simp)
done

(* nohide *)

lemma nohidefun_Rep_int_choice_fun_iff[simp]:
  "inj f ==> (%x. !!<f> a:X .. (Pff a x)) : nohidefun = (ALL a. Pff a : nohidefun)"
apply (auto simp add: Rep_int_choice_fun_def)
apply (drule_tac x="f a" in spec)
apply (simp)
done

lemma nohidefun_Rep_int_choice_f_iff[simp]:
  "inj f ==> (%x. !<f> a:X .. (Pff a x)) : nohidefun = (ALL a. Pff a : nohidefun)"
apply (simp add: Rep_int_choice_f_def)
apply (auto)
apply (drule_tac x="f a" in spec)
apply (simp)
done

lemma nohidefun_Alpha_parallel_iff[simp]:
  "(%x. (Pf x) |[X,Y]| (Qf x)) : nohidefun = (Pf : nohidefun & Qf : nohidefun)"
apply (simp add: Alpha_parallel_def)
done

lemma nohidefun_MU[simp]:
  "(ALL f: nohidefun. (%X. (PXY X (f X))) : nohidefun)
    ==> (%X. MUX (PXY X)) : nohidefun"
apply (simp add: MUX_def FIX_def)
apply (rule allI)
apply (simp add: FIXn_def)
apply (induct_tac n)
apply (simp)
apply (simp)
done

lemmas nohidefun_def = nohidefun_basic_def 
                       nohidefun_MU

(*** nohidefun test ***)

lemma "(%Z. (MU X .. a -> (MU Y .. b -> X [+] c -> Y |~| d -> Z i))) : nohidefun"
apply (simp)
done

(************ Fun ************)

consts
  gTickFun  :: "('pn,'a) procFun set"
  guardFun  :: "('pn,'a) procFun set"
  nohideFun :: "('pn,'a) procFun set"

defs 
  gTickFun_def : 
    "gTickFun == {PF. (ALL pn. (%x. (PF x pn)) : gTickfun)}"

  guardFun_def : 
    "guardFun == {PF. (ALL pn. (%x. (PF x pn)) : guardfun)}"

  nohideFun_def : 
    "nohideFun == {PF. (ALL pn. (%x. (PF x pn)) : nohidefun)}"

(************ for MU ************)

consts
  gTickX  :: "('a proc => 'a proc) set"
  guardX  :: "('a proc => 'a proc) set"
  nohideX :: "('a proc => 'a proc) set"

defs
  gTickX_def : "gTickX  == {PX. (%X. PX (X MUpn)) : gTickfun}"
  guardX_def : "guardX  == {PX. (%X. PX (X MUpn)) : guardfun}"
  nohideX_def: "nohideX == {PX. (%X. PX (X MUpn)) : nohidefun}"

(*********************************************************
             termination relation for proc
 *********************************************************)

consts
  procterm :: "('a proc * 'a proc) set"

inductive procterm
intros
  "(P, a -> P)             : procterm"
  "(Pf a, ? :X -> Pf)      : procterm"
  "(P, P [+] Q)            : procterm"
  "(Q, P [+] Q)            : procterm"
  "(P, P |~| Q)            : procterm"
  "(Q, P |~| Q)            : procterm"
  "(Pf a, !! :X .. Pf)     : procterm"
  "(P, IF b THEN P ELSE Q) : procterm"
  "(Q, IF b THEN P ELSE Q) : procterm"
  "(P, P |[X]| Q)          : procterm"
  "(Q, P |[X]| Q)          : procterm"
  "(P, P -- X)             : procterm"
  "(P, P [[r]])            : procterm"
  "(P, P ;; Q)             : procterm"
  "(Q, P ;; Q)             : procterm"

lemma wf_procterm: "wf procterm"
  apply (unfold wf_def)
  apply (intro strip)
  apply (induct_tac x)
  apply (erule allE, erule mp, intro strip,
         erule procterm.elims,
         simp, simp, simp, simp, simp, simp,
         simp, simp, simp, simp, simp, simp,
         simp, simp, simp)+
done

(*********************************************************
           expanding expressions (for step laws)
 *********************************************************)

consts
  Ext_choice_step  :: 
      "'a set => 'a set => ('a => 'a proc) 
                        => ('a => 'a proc) => 'a proc"

  Parallel_step ::
      "'a set => 'a set => 'a set
          => ('a => 'a proc) => ('a => 'a proc)
          => 'a proc"

  Hiding_step ::
      "'a set => 'a set => ('a => 'a proc) => 'a proc"

  Renaming_step ::
      "('a * 'a) set => 'a set => ('a => 'a proc) => 'a proc"

  Seq_compo_step ::
      "'a set => ('a => 'a proc) => 'a proc => 'a proc"

defs
  Ext_choice_step_def :
   "Ext_choice_step X Y Pf Qf ==
      ? x:(X Un Y) -> 
                (IF (x : X & x : Y) THEN Pf x |~| Qf x
                 ELSE IF (x : X) THEN Pf x
                 ELSE Qf x)"
defs
  Parallel_step_def :
   "Parallel_step X Y Z Pf Qf ==
      ? x:((X Int Y Int Z) Un (Y - X) Un (Z - X))
         -> IF (x : X) THEN (Pf x |[X]| Qf x)
               ELSE IF (x : Y & x : Z) THEN ((Pf x |[X]| ? x:Z -> Qf x)
                                     |~| (? x:Y -> Pf x |[X]| Qf x))
               ELSE IF (x : Y) THEN (Pf x |[X]| ? x:Z -> Qf x)
               ELSE (? x:Y -> Pf x |[X]| Qf x)"
defs
  Hiding_step_def :
   "Hiding_step X Y Pf ==
      IF (Y Int X = {}) THEN (? x:Y -> (Pf x -- X))
                         ELSE ((? x:(Y-X) -> (Pf x -- X)) 
                               [> (! x:(Y Int X) .. (Pf x -- X)))"
defs
  Renaming_step_def :
   "Renaming_step r X Pf ==
      ? y:(r `` X) -> (! x:{x. x:X & (x,y):r} .. (Pf x)[[r]])"

defs
  Seq_compo_step_def :
   "Seq_compo_step X Pf Q == ? x:X -> (Pf x ;; Q)"

(*-------------------------------------------------------* 
 |                                                       |
 |      decompostion controlled by Not_Decompo_Flag      |
 |                                                       |
 *-------------------------------------------------------*)

consts Not_Decompo_Flag :: bool
defs   Not_Decompo_Flag_def : "Not_Decompo_Flag == True"

lemma on_Not_Decompo_Flag: "Not_Decompo_Flag & R ==> R"
by (simp)

lemma off_Not_Decompo_Flag: "R ==> Not_Decompo_Flag & R"
by (simp add: Not_Decompo_Flag_def)

lemma off_Not_Decompo_Flag_True: "Not_Decompo_Flag"
by (simp add: Not_Decompo_Flag_def)

ML_setup {*
    val ON_Not_Decompo_Flag       = thms "on_Not_Decompo_Flag";
    val OFF_Not_Decompo_Flag      = thms "off_Not_Decompo_Flag";
    val OFF_Not_Decompo_Flag_True = thms "off_Not_Decompo_Flag_True";
*}

(*-------------------------------------------------------* 
 |                                                       |
 |                  unfold PF: ***Fun                    |
 |                                                       |
 *-------------------------------------------------------*)

lemmas anyFun = ProcFun_def gTickFun_def guardFun_def nohideFun_def

(****************** to add them again ******************)

declare Union_image_eq [simp]
declare Inter_image_eq [simp]

end
