           (*-------------------------------------------*
            |        CSP-Prover on Isabelle2004         |
            |               November 2004               |
            |                   June 2005  (modified)   |
            |                   July 2005  (modified)   |
            |                                           |
            |        CSP-Prover on Isabelle2005         |
            |                October 2005  (modified)   |
            |                                           |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Infra_type = Infra_common: 

(*****************************************************
              Infinite Sequence type
 *****************************************************)

types 'a infinite_seq = "nat => 'a"           (* synonym *)

(********************************************************** 
                     OR type
 **********************************************************)

datatype
  ('a,'b) or_type = OR1 'a | OR2 'b

lemma inj_OR1[simp]: "inj OR1"
by (simp add: inj_on_def)

lemma inj_OR2[simp]: "inj OR2"
by (simp add: inj_on_def)

lemma OR1_or_OR2: "ALL x. EX y. (x = OR1 y | x = OR2 y)"
apply (intro allI)
apply (induct_tac x)
by (auto)

lemma ALL_or_type_decompo: 
  "(ALL x:X. P x) = ((ALL x:(OR1 ` X). P ((inv OR1) x)) &
                     (ALL x:(OR2 ` X). P ((inv OR2) x)))"
by (auto)

lemma ALL_UNIV_or_type_decompo: 
  "(ALL x. P x) = ((ALL x:range OR1. P ((inv OR1) x)) &
                   (ALL x:range OR2. P ((inv OR2) x)))"
by (simp)

lemma P_inv_OR1:
  "[| ALL x:X. P x ; ox : (OR1 ` X) |] ==> P ((inv OR1) ox)"
by (auto)

lemma P_inv_OR2:
  "[| ALL x:X. P x ; ox : (OR2 ` X) |] ==> P ((inv OR2) ox)"
by (auto)

lemma P_OR1:
  "(ALL x. P (inv OR1 x)) = (ALL ox. P ox)"
apply (auto)
apply (drule_tac x="OR1 ox" in spec)
by (simp)

lemma P_OR1I:
  "ALL x. P (inv OR1 x) ==> P ox"
by (simp add: P_OR1)

lemma P_OR2:
  "(ALL x. P (inv OR2 x)) = (ALL ox. P ox)"
apply (auto)
apply (drule_tac x="OR2 ox" in spec)
by (simp)

lemma P_OR2I:
  "ALL x. P (inv OR2 x) ==> P ox"
by (simp add: P_OR2)

(********************************************************** 
                     Triple type
 **********************************************************)

datatype
  ('a,'b,'c) tri_type = Tri1 'a | Tri2 'b | Tri3 'c

lemma inj_Tri1[simp]: "inj Tri1"
by (simp add: inj_on_def)

lemma inj_Tri2[simp]: "inj Tri2"
by (simp add: inj_on_def)

lemma inj_Tri3[simp]: "inj Tri3"
by (simp add: inj_on_def)

lemma OR_tri_type: "ALL x. EX y. (x = Tri1 y | x = Tri2 y | x = Tri3 y)"
apply (intro allI)
apply (induct_tac x)
by (auto)

lemma ALL_tri_type_decompo: 
  "(ALL x:X. P x) = ((ALL x:(Tri1 ` X). P ((inv Tri1) x)) &
                     (ALL x:(Tri2 ` X). P ((inv Tri2) x)) &
                     (ALL x:(Tri3 ` X). P ((inv Tri3) x)))"
by (auto)

lemma ALL_UNIV_tri_type_decompo: 
  "(ALL x. P x) = ((ALL x:range Tri1. P ((inv Tri1) x)) &
                   (ALL x:range Tri2. P ((inv Tri2) x)) &
                   (ALL x:range Tri3. P ((inv Tri3) x)))"
by (simp)

lemma P_inv_Tri1:
  "[| ALL x:X. P x ; y : (Tri1 ` X) |] ==> P ((inv Tri1) y)"
by (auto)

lemma P_inv_Tri2:
  "[| ALL x:X. P x ; y : (Tri2 ` X) |] ==> P ((inv Tri2) y)"
by (auto)

lemma P_inv_Tri3:
  "[| ALL x:X. P x ; y : (Tri3 ` X) |] ==> P ((inv Tri3) y)"
by (auto)

lemma P_Tri1:
  "(ALL x. P (inv Tri1 x)) = (ALL y. P y)"
apply (auto)
apply (drule_tac x="Tri1 y" in spec)
by (simp)

lemma P_Tri1I:
  "ALL x. P (inv Tri1 x) ==> P y"
by (simp add: P_Tri1)

lemma P_Tri2:
  "(ALL x. P (inv Tri2 x)) = (ALL y. P y)"
apply (auto)
apply (drule_tac x="Tri2 y" in spec)
by (simp)

lemma P_Tri2I:
  "ALL x. P (inv Tri2 x) ==> P y"
by (simp add: P_Tri2)

lemma P_Tri3:
  "(ALL x. P (inv Tri3 x)) = (ALL y. P y)"
apply (auto)
apply (drule_tac x="Tri3 y" in spec)
by (simp)

lemma P_Tri3I:
  "ALL x. P (inv Tri3 x) ==> P y"
by (simp add: P_Tri3)

end
