           (*-------------------------------------------*
            |        CSP-Prover on Isabelle2004         |
            |               December 2004               |
            |                   July 2005 (modified)    |
            |                 August 2005 (modified)    |
            |                                           |
            |        CSP-Prover on Isabelle2005         |
            |                October 2005  (modified)   |
            |                                           |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_F_failuresfun = CSP_T_tracesfun + CSP_F_full_abs:

(*****************************************************************

         1. semantics for process functions
         2. 
         3. 
         4. 

 *****************************************************************)

(*---------------------------------*
 |                                 |
 |       traces , invT, invF       |
 |                                 |
 *---------------------------------*)

lemma traces_Pf_inv_semT_semF:
  "Pf : Procfun ==>
  traces (Pf (%pn. inv[[fstF (x pn)]]T)) = traces (Pf (%pn. inv[[x pn]]F))"
apply (simp add: tracesfun_X)
apply (simp add: traces_inv_semT_semF)
done

(*********************************************************
                      semantics
 *********************************************************)

consts
  failuresfun  :: "('pn,'a) procfun => (('pn => 'a domF) => 'a setF)"
  failuresFun  :: "('pn,'a) procFun => (('pn => 'a domF) => ('pn => 'a setF))"

  semFfun    :: "('pn,'a) procfun => (('pn => 'a domF) => 'a domF)"  ("[[_]]Ffun")
  semFFun    :: "('pn,'a) procFun => 
                 (('pn => 'a domF) => ('pn => 'a domF))"  ("[[_]]FFun")
defs
  failuresfun_def : "failuresfun Pf == (%X. failures (Pf (%pn. inv[[X pn]]F)))"
  failuresFun_def : "failuresFun PF == (%X. (%pn. failuresfun (%X. (PF X) pn) X))"

  semFfun_def : "[[Pf]]Ffun == (%X. [[Pf (%pn. inv[[X pn]]F)]]F)"
  semFFun_def : "[[PF]]FFun == (%X. (%pn. [[(%X. (PF X) pn)]]Ffun X))"

(*********************************************************
              semantics for each operators
 *********************************************************)

lemma failuresfun_pn: 
  "failuresfun (%x. x pn) = (%x. sndF (x pn))"
apply (simp add: failuresfun_def)
apply (simp add: failures_inv_semF)
done

lemma failuresfun_STOP: 
  "failuresfun (%x. STOP) = (%x. {f. EX X. f = ([]t, X) }f)"
by (simp add: failuresfun_def failures_def)

lemma failuresfun_SKIP: 
  "failuresfun (%x. SKIP) = (%x. {f. (EX X. f = ([]t, X) & X <= Evset) |
                              (EX X. f = ([Tick]t, X)) }f)"
by (simp add: failuresfun_def failures_def)

lemma failuresfun_DIV: 
  "failuresfun (%x. DIV) = (%x. {}f)"
by (simp add: failuresfun_def failures_def)

lemma failuresfun_Act_prefix:
  "failuresfun (%x. a -> (Pf x)) = 
  (%x. {f. (EX X.   f = ([]t,X) & Ev a ~: X) |
           (EX s X. f = ([Ev a]t @t s, X) & (s,X) :f failuresfun Pf x ) }f)"
by (simp add: failuresfun_def failures_def)

lemma failuresfun_Ext_pre_choice:
  "failuresfun (%x. ? a:X -> (Pff a x)) = 
                     (%x. {f. (EX Y.     f = ([]t,Y) & Ev`X Int Y = {}) |
                              (EX a s Y. f = ([Ev a]t @t s, Y) & 
                                   (s,Y) :f failuresfun (Pff a) x & a : X) }f)"
by (simp add: failuresfun_def failures_def)

lemma failuresfun_Ext_choice:
  "[| Pf : Procfun ; Qf : Procfun |] ==> 
   failuresfun (%x. (Pf x) [+] (Qf x)) = 
   (%x. {f. ((EX X. f = ([]t, X)) & f :f failuresfun Pf x & f :f failuresfun Qf x |
     (EX s. (EX X. f = (s, X)) &
            (f :f failuresfun Pf x | f :f failuresfun Qf x) & s ~= []t) |
     (EX X. f = ([]t, X) &
            ([Tick]t :t tracesfun Pf (fstF o x) | [Tick]t :t tracesfun Qf (fstF o x)) &
            X <= Evset)) }f)"
apply (simp add: failuresfun_def)
apply (simp add: failures_def)
apply (simp add: tracesfun_def)
apply (simp add: tracesfun_X)
apply (simp add: traces_inv_semT)
apply (simp add: traces_inv_semF)
done

lemma failuresfun_Int_choice:
  "failuresfun (%x. (Pf x) |~| (Qf x)) = (%x. failuresfun Pf x UnF failuresfun Qf x)"
by (simp add: failuresfun_def failures_def)

lemma failuresfun_Rep_int_choice_Tri:
  "failuresfun (%x. !! a:X .. (Pff a x)) = 
  (%x. {f. EX a:X. f :f failuresfun (Pff a) x}f)"
by (simp add: failuresfun_def failures_def)

lemma failuresfun_Rep_int_choice_fun:
  "inj f ==> failuresfun (%x. !!<f> a:X .. (Pff a x)) = 
  (%x. {f. EX a:X. f :f failuresfun (Pff a) x}f)"
by (simp add: failuresfun_def failures_def)

lemmas failuresfun_Rep_int_choice = failuresfun_Rep_int_choice_Tri
                                    failuresfun_Rep_int_choice_fun

lemma failuresfun_Conditional:
  "failuresfun (%x. IF b THEN (Pf x) ELSE (Qf x))
   = (%x. if b then failuresfun Pf x else failuresfun Qf x)"
apply (case_tac "b")
by (simp_all add: failuresfun_def failures_def)

lemma failuresfun_Parallel:
  "failuresfun (%x. (Pf x) |[X]| (Qf x))
   = (%x. {f. EX u Y Z. f = (u, Y Un Z) &
              Y-((Ev`X) Un {Tick})= Z-((Ev`X) Un {Tick}) &
              (EX s t. u : s |[X]|tr t & (s,Y) :f failuresfun Pf x & 
                                         (t,Z) :f failuresfun Qf x) }f)"
by (simp add: failuresfun_def failures_def)

lemma failuresfun_Hiding:
  "failuresfun (%x. (Pf x) -- X)
   = (%x. {f. EX s Y. f = (s --tr X, Y) & 
                      (s,(Ev`X) Un Y) :f failuresfun Pf x }f)"
by (simp add: failuresfun_def failures_def)

lemma failuresfun_Renaming:
  "failuresfun (%x. (Pf x) [[r]])
   = (%x. {f. EX s t X. f = (t,X) & s [[r]]tr t & 
                       (s, [[r]]inv X) :f failuresfun Pf x }f)"
by (simp add: failuresfun_def failures_def)

lemma failuresfun_Seq_compo:
  "Pf : Procfun ==>
   failuresfun (%x. (Pf x) ;; (Qf x))
   = (%x. {f. (EX t X. f = (t,X) & (t, X Un {Tick}) :f failuresfun Pf x &
                       noTick t) |
              (EX s t X. f = (s @t t,X) & s @t [Tick]t :t tracesfun Pf (fstF o x) & 
                      (t, X) :f failuresfun Qf x & noTick s) }f)"
apply (simp add: failuresfun_def)
apply (simp add: failures_def)
apply (simp add: tracesfun_def)
apply (simp add: tracesfun_X)
apply (simp add: traces_inv_semT)
apply (simp add: traces_inv_semF)
done

lemmas failuresfun_simp =
  failuresfun_pn failuresfun_STOP failuresfun_SKIP failuresfun_DIV
  failuresfun_Act_prefix failuresfun_Ext_pre_choice 
  failuresfun_Ext_choice failuresfun_Int_choice
  failuresfun_Rep_int_choice failuresfun_Conditional
  failuresfun_Parallel failuresfun_Hiding
  failuresfun_Renaming failuresfun_Seq_compo

(*********************************************************
                   in failuresfun
 *********************************************************)

lemma in_failuresfun_pn: 
  "(f :f failuresfun (%x. x pn) x) = (f :f sndF (x pn))"
apply (simp add: failuresfun_def)
apply (simp add: failures_inv_semF)
done

lemma in_failuresfun_STOP: 
  "(f :f failuresfun (%x. STOP) x) = (EX X. f = ([]t, X))"
by (simp add: failuresfun_def in_failures)

lemma in_failuresfun_SKIP: 
  "(f :f failuresfun (%x. SKIP) x) = 
   ((EX X. f = ([]t, X) & X <= Evset) |
    (EX X. f = ([Tick]t, X)))"
by (simp add: failuresfun_def in_failures)

lemma in_failuresfun_DIV: 
  "(f ~:f failuresfun (%x. DIV) x)"
by (simp add: failuresfun_def in_failures)

lemma in_failuresfun_Act_prefix: 
  "(f :f failuresfun (%x. a -> (Pf x)) x) = 
   ((EX X.   f = ([]t,X) & Ev a ~: X) |
    (EX s X. f = ([Ev a]t @t s, X) & (s,X) :f failuresfun Pf x ))"
by (simp add: failuresfun_def in_failures)

lemma in_failuresfun_Ext_pre_choice:
  "(f :f failuresfun (%x. ? a:X -> (Pff a x)) x) = 
   ((EX Y.     f = ([]t,Y) & Ev`X Int Y = {}) |
    (EX a s Y. f = ([Ev a]t @t s, Y) & 
               (s,Y) :f failuresfun (Pff a) x & a : X))"
by (simp add: failuresfun_def in_failures)

lemma in_failuresfun_Ext_choice:
  "[| Pf : Procfun ; Qf : Procfun |] ==> 
   (f :f failuresfun (%x. (Pf x) [+] (Qf x)) x) = 
   (((EX X. f = ([]t, X)) & f :f failuresfun Pf x & f :f failuresfun Qf x |
     (EX s. (EX X. f = (s, X)) &
            (f :f failuresfun Pf x | f :f failuresfun Qf x) & s ~= []t) |
     (EX X. f = ([]t, X) &
            ([Tick]t :t tracesfun Pf (fstF o x) | [Tick]t :t tracesfun Qf (fstF o x)) &
            X <= Evset)))"
apply (simp add: failuresfun_def in_failures)
apply (simp add: tracesfun_def)
apply (simp add: tracesfun_X)
apply (simp add: traces_inv_semT)
apply (simp add: traces_inv_semF)
done

lemma in_failuresfun_Int_choice:
  "(f :f failuresfun (%x. (Pf x) |~| (Qf x)) x) = 
   (f :f failuresfun Pf x | f :f failuresfun Qf x)"
by (simp add: failuresfun_def in_failures)

lemma in_failuresfun_Rep_int_choice_Tri:
  "(f :f failuresfun (%x. !! a:X .. (Pff a x)) x) = 
   (EX a:X. f :f failuresfun (Pff a) x)"
by (simp add: failuresfun_def in_failures)

lemma in_failuresfun_Rep_int_choice_fun:
  "inj g ==> (f :f failuresfun (%x. !!<g> a:X .. (Pff a x)) x) = 
   (EX a:X. f :f failuresfun (Pff a) x)"
by (simp add: failuresfun_def in_failures)

lemmas in_failuresfun_Rep_int_choice = in_failuresfun_Rep_int_choice_Tri
                                       in_failuresfun_Rep_int_choice_fun

lemma in_failuresfun_Conditional:
  "(f :f failuresfun (%x. IF b THEN (Pf x) ELSE (Qf x)) x) =
   (if b then f :f failuresfun Pf x else f :f failuresfun Qf x)"
apply (case_tac "b")
by (simp_all add: failuresfun_def failures_def)

lemma in_failuresfun_Parallel:
  "(f :f failuresfun (%x. (Pf x) |[X]| (Qf x)) x) =
   (EX u Y Z. f = (u, Y Un Z) &
              Y-((Ev`X) Un {Tick})= Z-((Ev`X) Un {Tick}) &
              (EX s t. u : s |[X]|tr t & (s,Y) :f failuresfun Pf x & 
                                         (t,Z) :f failuresfun Qf x))"
by (simp add: failuresfun_def in_failures)

lemma in_failuresfun_Hiding:
  "(f :f failuresfun (%x. (Pf x) -- X) x) =
   (EX s Y. f = (s --tr X, Y) & 
                      (s,(Ev`X) Un Y) :f failuresfun Pf x)"
by (simp add: failuresfun_def in_failures)

lemma in_failuresfun_Renaming:
  "(f :f failuresfun (%x. (Pf x) [[r]]) x) =
   (EX s t X. f = (t,X) & s [[r]]tr t & 
             (s, [[r]]inv X) :f failuresfun Pf x)"
by (simp add: failuresfun_def in_failures)

lemma in_failuresfun_Seq_compo:
  "Pf : Procfun ==>
   (f :f failuresfun (%x. Pf x ;; Qf x) x) =
   ((EX t X. f = (t, X) & (t, insert Tick X) :f failuresfun Pf x & noTick t) |
    (EX s t X. f = (s @t t, X) & s @t [Tick]t :t tracesfun Pf (fstF o x) &
             (t, X) :f failuresfun Qf x & noTick s))"
apply (simp add: failuresfun_def in_failures)
apply (simp add: tracesfun_def)
apply (simp add: tracesfun_X)
apply (simp add: traces_inv_semT)
apply (simp add: traces_inv_semF)
done

lemmas in_failuresfun =
  in_failuresfun_pn in_failuresfun_STOP in_failuresfun_SKIP in_failuresfun_DIV
  in_failuresfun_Act_prefix in_failuresfun_Ext_pre_choice 
  in_failuresfun_Ext_choice in_failuresfun_Int_choice
  in_failuresfun_Rep_int_choice in_failuresfun_Conditional
  in_failuresfun_Parallel in_failuresfun_Hiding
  in_failuresfun_Renaming in_failuresfun_Seq_compo

(*************************************************
            failuresfun, failuresFun
 *************************************************)

lemma proj_failuresFun_failuresfun:
 "(proj_fun pn o failuresFun PF) = failuresfun (%X. PF X pn)"
apply (simp add: failuresFun_def)
apply (simp add: proj_fun_def)
apply (simp add: expand_fun_eq)
done

lemma proj_semFFun_semFfun:
 "(proj_fun pn o [[PF]]FFun) = [[%X. PF X pn]]Ffun"
apply (simp add: semFFun_def)
apply (simp add: proj_fun_def)
apply (simp add: expand_fun_eq)
done

(**************************************************
        failures failuresfun failuresFun
 **************************************************)

lemma failuresfun_X:
  "Pf : Procfun ==> 
   failures (Pf X) = 
   failuresfun (Pf) (%pn. (traces(X pn),, failures(X pn)))"
apply (simp add: failuresfun_def)
apply (rule Procfun.induct[of Pf])
apply (simp_all add: failures_def failures_inv_semF)
apply (simp_all add: tracesfun_X traces_inv_semF)
done

lemma failuresFun_X_pn:
  "PF : ProcFun ==> 
   failures(PF X pn) = 
   failuresFun(PF) (%pn. (traces(X pn),, failures(X pn))) pn"
apply (simp add: failuresFun_def)
apply (simp add: ProcFun_def)
apply (simp add: failuresfun_X[THEN sym])
done

(**************************************************
        [[ ]]Ffun [[ ]]FFun traces failures
 **************************************************)

lemma semFfun_traces_failures:
  "Pf : Procfun 
   ==> [[Pf]]Ffun = (%x. (traces (Pf (%pn. inv[[x pn]]F)) ,, 
                        failures (Pf (%pn. inv[[x pn]]F))))"
apply (simp add: semFfun_def)
apply (simp add: semF_def)
done

lemma semFfun_tracesfun_failuresfun:
  "Pf : Procfun ==> [[Pf]]Ffun = (%X. (tracesfun Pf (fstF o X) ,, failuresfun Pf X))"
apply (simp add: failuresfun_def tracesfun_def)
apply (simp add: semFfun_traces_failures)
apply (simp add: traces_Pf_inv_semT_semF)
done

lemma semFFun_traces_failures:
  "PF : ProcFun 
   ==> [[PF]]FFun = (%x pn. (traces (PF (%pn. inv[[x pn]]F) pn) ,, 
                           failures (PF (%pn. inv[[x pn]]F) pn)))"
apply (simp add: semFFun_def)
apply (simp add: ProcFun_def)
apply (simp add: semFfun_traces_failures)
done

lemma semFFun_tracesFun_failuresFun:
  "PF : ProcFun ==> 
   [[PF]]FFun = 
   (%X pn. (tracesFun PF (fstF o X) pn,, failuresFun PF X pn))"
apply (simp add: tracesFun_def failuresFun_def semFFun_def)
apply (simp add: ProcFun_def)
apply (simp add: semFfun_tracesfun_failuresfun)
done

lemmas semF_decompo = semF_def
                      semFfun_traces_failures
                      semFFun_traces_failures

lemmas semF_decompo_fun = semF_def
                          semFfun_tracesfun_failuresfun
                          semFFun_tracesFun_failuresFun

(*** semFfun_variable ***)

lemma semFfun_variable: "[[%x. x pn]]Ffun = (%x. x pn)"
apply (simp add: Procfun_def semF_decompo_fun)
apply (simp add: tracesfun_simp)
apply (simp add: failuresfun_simp)
done

(**************************************************
           failuresfun composition
 **************************************************)

lemma failuresfun_compo_failuresFun_X:
   "[| Pf : Procfun ; QF : ProcFun |] ==>
    failuresfun(Pf o QF) X = 
    failuresfun(Pf) (%pn. (tracesFun(QF) (fstF o X) pn ,, failuresFun(QF) X pn))"
apply (simp add: failuresfun_def)
apply (simp add: failuresfun_X)
apply (simp add: tracesFun_X_pn failuresFun_X_pn)
apply (simp add: failures_inv_semF traces_inv_semF)
apply (simp add: comp_def)
done

lemma failuresfun_compo_failuresFun_fun:
   "[| Pf : Procfun ; QF : ProcFun |] ==>
    failuresfun(Pf o QF) = 
    failuresfun(Pf) o (%X pn. (tracesFun(QF) (fstF o X) pn ,, failuresFun(QF) X pn))"
apply (simp add: expand_fun_eq)
apply (simp add: failuresfun_compo_failuresFun_X)
done

lemmas failuresfun_compo_failuresFun = failuresfun_compo_failuresFun_X 
                                       failuresfun_compo_failuresFun_fun 
                      failuresfun_compo_failuresFun_X[simplified comp_def]
                      failuresfun_compo_failuresFun_fun[simplified comp_def]

(*** ProcFun ***)

lemma failuresFun_compo_failuresFun_X_lm:
   "[| PF : ProcFun ; QF : ProcFun |] ==>
    failuresfun (%x. PF (QF x) pn) X =
    failuresfun (%X. PF X pn)
    (%pn. (tracesFun(QF) (fstF o X) pn ,, failuresFun(QF) X pn))"
apply (insert failuresfun_compo_failuresFun_X[of "(%X. PF X pn)" QF X])
apply (simp add: ProcFun_def comp_def)
done

lemma failuresFun_compo_failuresFun_X:
   "[| PF : ProcFun ; QF : ProcFun |] ==>
    failuresFun(PF o QF) X = 
    failuresFun(PF) (%pn. (tracesFun(QF) (fstF o X) pn ,, failuresFun(QF) X pn))"
apply (simp add: failuresFun_def[of PF])
apply (simp add: failuresFun_def[of "(PF o QF)"])
apply (simp add: expand_fun_eq)
apply (simp add: failuresFun_compo_failuresFun_X_lm)
done

(*** composition of process functions ***)

lemma failuresFun_compo_failuresFun_fun:
   "[| PF : ProcFun ; QF : ProcFun |] ==>
    failuresFun(PF o QF) = 
    failuresFun(PF) o (%X pn. (tracesFun(QF) (fstF o X) pn ,, failuresFun(QF) X pn))"
apply (simp add: expand_fun_eq)
apply (simp add: failuresFun_compo_failuresFun_X)
done

lemmas failuresFun_compo_failuresFun = failuresFun_compo_failuresFun_X 
                                       failuresFun_compo_failuresFun_fun
                  failuresFun_compo_failuresFun_X[simplified comp_def]
                  failuresFun_compo_failuresFun_fun[simplified comp_def]

(**************************************************
           [[P]]F, [[P]]Ffun, [[P]]FFun
 **************************************************)

lemma semFfun_X:
  "Pf : Procfun ==> [[Pf X]]F = [[Pf]]Ffun (%pn. [[X pn]]F)"
apply (simp add: eqF_decompo)
apply (simp add: semFfun_def)
apply (simp add: tracesfun_X failuresfun_X)
apply (simp add: traces_inv_semF failures_inv_semF)
done

lemma semFFun_X_pn:
  "PF : ProcFun ==> [[PF X pn]]F = [[PF]]FFun (%pn. [[X pn]]F) pn"
apply (simp add: eqF_decompo)
apply (simp add: semFFun_def semFfun_def)
apply (simp add: tracesFun_X_pn failuresFun_X_pn)
apply (simp add: traces_inv_semF failures_inv_semF)
done

lemma semFfun_compo_semFFun_X:
   "[| Pf : Procfun ; QF : ProcFun |] ==>
    [[Pf o QF]]Ffun X = [[Pf]]Ffun ([[QF]]FFun X)"
apply (simp add: eqF_decompo)
apply (simp add: semFFun_def semFfun_def)
apply (simp add: tracesfun_X failuresfun_X)
apply (simp add: traces_inv_semF failures_inv_semF)
done

lemma semFfun_compo_semFFun_fun:
   "[| Pf : Procfun ; QF : ProcFun |] ==>
    [[Pf o QF]]Ffun = [[Pf]]Ffun o [[QF]]FFun"
apply (simp add: expand_fun_eq)
apply (simp add: semFfun_compo_semFFun_X)
done

lemmas semFfun_compo_ProcFun = semFfun_compo_semFFun_X semFfun_compo_semFFun_fun
                               semFfun_compo_semFFun_X[simplified comp_def]
                               semFfun_compo_semFFun_fun[simplified comp_def]

(*** semFFun ***)

lemma semFFun_compo_semFFun_X_lm:
   "[| PF : ProcFun; QF : ProcFun |]
         ==> [[%X. PF (QF X) pn]]Ffun X =
             [[%X. PF X pn]]Ffun (%pn. [[%X. QF X pn]]Ffun X)"
apply (insert semFfun_compo_semFFun_fun[of "(%X. PF X pn)" QF])
apply (simp add: ProcFun_def comp_def)
apply (simp add: semFFun_def)
done

lemma semFFun_compo_semFFun_X:
   "[| PF : ProcFun ; QF : ProcFun |] ==>
    [[PF o QF]]FFun X = [[PF]]FFun ([[QF]]FFun X)"
apply (simp add: semFFun_def)
apply (simp add: semFFun_compo_semFFun_X_lm)
done

(*** composition of process functions ***)

lemma semFFun_compo_semFFun_fun:
   "[| PF : ProcFun ; QF : ProcFun |] ==>
    [[PF o QF]]FFun = [[PF]]FFun o [[QF]]FFun"
apply (simp add: expand_fun_eq)
apply (simp add: semFFun_compo_semFFun_X)
done

lemmas semFFun_compo_semFFun = semFFun_compo_semFFun_X semFFun_compo_semFFun_fun
                               semFFun_compo_semFFun_X[simplified comp_def]
                               semFFun_compo_semFFun_fun[simplified comp_def]

lemma semFFun_compo_n_X:
   "PF : ProcFun ==> [[PF ^ n]]FFun X = ([[PF]]FFun ^ n) X"
apply (induct_tac n)

apply (simp add: semFFun_def semFfun_def)
apply (simp add: semF_inv_semF)
apply (simp add: semFFun_compo_semFFun iteration_in_ProcFun)
done

lemma semFFun_compo_n:
   "PF : ProcFun ==> [[PF ^ n]]FFun = ([[PF]]FFun ^ n)"
apply (simp add: expand_fun_eq)
apply (simp add: semFFun_compo_n_X)
done

(**************************************************
          tracesfun failuresfun in domF
 **************************************************)

lemma tracesfun_failuresfun_in_domF1:
  "Pf : Procfun ==> (tracesfun Pf (%pn. fstF (X pn)) , failuresfun Pf X) : domF"
apply (subgoal_tac 
  "(traces (Pf (%pn. inv[[X pn]]F)) , failures (Pf (%pn. inv[[X pn]]F))) : domF")
apply (simp only: tracesfun_X failuresfun_X)
apply (simp add: traces_inv_semF failures_inv_semF)
apply (simp)
done

lemma tracesfun_failuresfun_in_domF2:
  "Pf : Procfun ==> (tracesfun Pf (fstF o X) , failuresfun Pf X) : domF"
by (simp add: comp_def tracesfun_failuresfun_in_domF1)

lemmas tracesfun_failuresfun_in_domF = tracesfun_failuresfun_in_domF1
                                       tracesfun_failuresfun_in_domF2

lemma tracesFun_failuresFun_in_domF1:
  "PF : ProcFun ==> (tracesFun PF (%pn. fstF (X pn)) pn , failuresFun PF X pn) : domF"
apply (simp add: tracesFun_def)
apply (simp add: failuresFun_def)
apply (simp add: ProcFun_def)
apply (simp add: tracesfun_failuresfun_in_domF)
done

lemma tracesFun_failuresFun_in_domF2:
  "PF : ProcFun ==> (tracesFun PF (fstF o X) pn , failuresFun PF X pn) : domF"
by (simp add: comp_def tracesFun_failuresFun_in_domF1)

lemmas tracesFun_failuresFun_in_domF = tracesFun_failuresFun_in_domF1
                                       tracesFun_failuresFun_in_domF2

lemmas fun_in_domF = tracesfun_failuresfun_in_domF
                     tracesFun_failuresFun_in_domF

(**************************************************
                 fstF and sndF
 **************************************************)

lemma fstF_semFfun: 
  "Pf : Procfun ==> fstF ([[Pf]]Ffun X) = tracesfun Pf (fstF o X)"
apply (simp add: semF_decompo_fun)
apply (simp add: tracesfun_failuresfun_in_domF pairF)
done

lemma sndF_semFfun: 
  "Pf : Procfun ==> sndF ([[Pf]]Ffun X) = failuresfun Pf X"
apply (simp add: semF_decompo_fun)
apply (simp add: tracesfun_failuresfun_in_domF pairF)
done

lemma fstF_semFFun: 
  "PF : ProcFun ==> fstF ([[PF]]FFun X pn) = tracesFun PF (fstF o X) pn"
apply (simp add: semF_decompo_fun)
apply (simp add: tracesFun_failuresFun_in_domF pairF)
done

lemma sndF_semFFun: 
  "PF : ProcFun ==> sndF ([[PF]]FFun X pn) = failuresFun PF X pn"
apply (simp add: semF_decompo_fun)
apply (simp add: tracesFun_failuresFun_in_domF pairF)
done

lemmas pairF_semF = pairF fstF_semFfun sndF_semFfun fstF_semFFun sndF_semFFun

end
