           (*-------------------------------------------*
            |        CSP-Prover on Isabelle2004         |
            |                 August 2005 (modified)    |
            |                                           |
            |        CSP-Prover on Isabelle2005         |
            |                October 2005  (modified)   |
            |                                           |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_F_full_abs = CSP_F_domain + CSP_F_proc + CSP_T_full_abs:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  disj_not1: (~ P | Q) = (P --> Q)                   *)

declare disj_not1 [simp del]

(*********************************************************
            inverse function : DomT => proc
 *********************************************************)

consts
  head_failures :: "'a setF => 'a set"
  tail_failures :: "'a setF => 'a => 'a setF"

  proc_setF   :: "nat => 'a domF => 'a proc"

  inv_semF    :: "'a domF => 'a proc"      ("inv[[_]]F")

defs
  head_failures_def :
  "head_failures F == {a. EX t X. ([Ev a]t @t t, X) :f F}"
  tail_failures_def :
  "tail_failures F == (%a. {(t,X). ([Ev a]t @t t, X) :f F}f)"

primrec
  "proc_setF 0 = (%DF. !set X:{X. EX Y. ([]t,Y) :f sndF DF & 
                                  Ev ` X = Evset - Y & Tick : Y &
                                  (ALL a:X. [Ev a]t :t fstF DF)} .. 
                        (? x:X -> DIV))"
  "proc_setF (Suc n)
     = (%DF. (! a:(head_failures (sndF DF)) .. a -> 
              proc_setF n (tail_traces (fstF DF) a ,,
                           tail_failures (sndF DF) a) [+] DIV))"

defs
  inv_semF_def :
   "inv[[DF]]F == inv[[fstF DF]]T |~| (!nat n .. proc_setF n DF)"

(*********************************************************
                     lemmas
 *********************************************************)

(* tail in setF *)

lemma tail_failures_setF: "{(t,X). ([Ev a]t @t t, X) :f F} : setF"
apply (simp add: setF_def)
apply (simp add: HC_F2_def)
apply (intro allI impI)
apply (rule memF_F2)
by (auto)

(* tail *)

lemma in_tail_failures: 
  "((s,X) :f tail_failures F a) = (([Ev a]t @t s, X) :f F)"
apply (simp add: tail_failures_def)
apply (simp add: tail_failures_setF CollectF_open_memF)
done

(* head & tail *)

lemma head_tail_failures_only_if:
   "([Ev a]t @t s, X) :f F ==> a : head_failures F & (s,X) :f tail_failures F a"
apply (simp add: in_tail_failures)
apply (auto simp add: head_failures_def)
done

(* iff *)

lemma head_tail_failures:
   "([Ev a]t @t s, X) :f F = (a : head_failures F & (s,X) :f tail_failures F a)"
apply (rule iffI)
apply (simp add: head_tail_failures_only_if)
apply (simp add: in_tail_failures)
done

(*** domF ***)

(* head *)

lemma head_failures_traces: 
  "a : head_failures (sndF DF) ==> a : head_traces (fstF DF)"
apply (simp add: head_failures_def)
apply (simp add: head_traces_def)
apply (elim exE)
apply (rule_tac x="t" in exI)
apply (simp add: pairF_domF_T2)
done

(* T2 *)

lemma tail_traces_failures_T2: 
  "HC_T2 (tail_traces (fstF DF) a , tail_failures (sndF DF) a)"
apply (simp add: HC_T2_def in_traces in_failures)
apply (intro allI impI)
apply (elim exE)
apply (simp add: in_tail_failures)
apply (subgoal_tac "a : head_traces (fstF DF)")
apply (simp add: in_tail_traces)
apply (simp add: pairF_domF_T2)

apply (simp add: head_tail_failures)
apply (simp add: head_failures_traces)
done

(* F3 *)

lemma tail_traces_failures_F3: 
  "a : head_traces (fstF DF) ==>
   HC_F3 (tail_traces (fstF DF) a , tail_failures (sndF DF) a)"
apply (simp add: HC_F3_def in_traces in_failures)
apply (intro allI impI)
apply (elim conjE)

apply (simp add: in_tail_failures)
apply (simp add: in_tail_traces)
apply (rule pairF_domF_F3)
apply (simp)
apply (simp)
apply (simp add: appt_assoc_sym)
done

(* T3_F4 *)

lemma tail_traces_failures_T3_F4: 
  "a : head_traces (fstF DF) ==>
   HC_T3_F4 (tail_traces (fstF DF) a , tail_failures (sndF DF) a)"
apply (simp add: HC_T3_F4_def in_traces in_failures)
apply (intro allI impI)
apply (elim conjE)

apply (simp add: in_tail_failures)
apply (simp add: in_tail_traces)
apply (simp add: appt_assoc_sym)
apply (simp add: pairF_domF_F4)
apply (simp add: pairF_domF_T3)
done

lemma tail_traces_failures_domF: 
  "a : head_traces (fstF DF) | a : head_failures (sndF DF) 
   ==> (tail_traces (fstF DF) a , tail_failures (sndF DF) a) : domF"
apply (simp add: domF_iff)
apply (subgoal_tac "a : head_traces (fstF DF)")
apply (simp add: tail_traces_failures_T2)
apply (simp add: tail_traces_failures_F3)
apply (simp add: tail_traces_failures_T3_F4)
apply (auto simp add: head_failures_traces)
done

(*-------------------------------------*
 |   failures (proc_domT n) --> Tick   |
 *-------------------------------------*)

lemma failures_proc_domT_noTick_lm: 
  "ALL s T. ((s, X) :f failures (proc_domT n T) & noTick s)
   --> (EX t. s @t t @t [Tick]t :t T & noTick t)"
apply (induct_tac n)

(* base *)
 apply (simp add: in_traces)
 apply (simp add: in_failures)

(* step *)
 apply (intro allI impI)
 apply (simp add: in_failures)
 apply (elim conjE exE disjE bexE)
 apply (simp_all)

  apply (drule_tac x="sa" in spec)
  apply (drule_tac x="(tail_traces T a)" in spec)
  apply (simp)
  apply (elim conjE exE)
  apply (rule_tac x="t" in exI)
  apply (simp add: appt_assoc)
  apply (simp add: head_tail_traces)

  apply (simp add: in_traces)
  apply (simp add: in_traces)

  apply (case_tac "[Tick]t :t T")

   apply (simp add: in_failures)
   apply (erule disjE)
    apply (rule_tac x="[]t" in exI)
    apply (simp)
    apply (simp)

   apply (simp add: in_failures)
done

lemma failures_proc_domT_noTick: 
  "[| (s, X) :f failures (proc_domT n T) ; noTick s|]
   ==> (EX t. s @t t @t [Tick]t :t T & noTick t)"
by (simp add: failures_proc_domT_noTick_lm)

lemma failures_proc_domT_lm: 
  "ALL s T. ((s, X) :f failures (proc_domT n T))
   --> s :t T"
apply (induct_tac n)

(* base *)
 apply (simp add: in_traces)
 apply (simp add: in_failures)

(* step *)
 apply (intro allI impI)
 apply (simp add: in_failures)
 apply (elim conjE exE disjE bexE)
 apply (simp_all)

  apply (drule_tac x="sa" in spec)
  apply (drule_tac x="(tail_traces T a)" in spec)
  apply (simp add: head_tail_traces)

  apply (case_tac "[Tick]t :t T")
  apply (auto simp add: in_failures)
done

lemma failures_proc_domT: 
  "(s, X) :f failures (proc_domT n T) ==> s :t T"
by (simp add: failures_proc_domT_lm)

lemma failures_proc_domT_T3: 
  "[| (s, X) :f failures (proc_domT n (fstF DF)) ; noTick s |]
   ==> (EX t. (s @t t @t [Tick]t,Y) :f (sndF DF) & noTick t)"
apply (insert failures_proc_domT_noTick_lm[of X n])
apply (drule_tac x="s" in spec)
apply (drule_tac x="fstF DF" in spec)
apply (simp)
apply (elim conjE exE)
apply (rule_tac x="t" in exI)
apply (simp)
apply (simp add: appt_assoc_sym)
apply (simp add: pairF_domF_T3)
done

(*** head T --> head F **)

lemma head_traces_failures_noTick:
  "[| a : head_traces (fstF DF);
     (s, X) :f failures (proc_domT n (tail_traces (fstF DF) a)) ; noTick s|]
   ==> a : head_failures (sndF DF)"
apply (insert failures_proc_domT_noTick_lm[of X n])
apply (drule_tac x="s" in spec)
apply (drule_tac x="tail_traces (fstF DF) a" in spec)
apply (simp)
apply (elim conjE exE)
apply (simp add: in_tail_traces)

apply (simp add: head_failures_def)
apply (rule_tac x="s @t t @t [Tick]t" in exI)
apply (rule_tac x="X" in exI)
apply (simp add: appt_assoc_sym)
apply (simp add: pairF_domF_T3)
done

(*** head_traces_failures ***)

lemma head_traces_failures:
  "[| a : head_traces (fstF DF);
     (s, X) :f failures (proc_domT n (tail_traces (fstF DF) a)) |]
   ==> a : head_failures (sndF DF)"
apply (case_tac "noTick s")
apply (simp add: head_traces_failures_noTick)
apply (simp add: noTick_def)

apply (simp add: Tick_in_sett)
apply (elim conjE exE)

apply (insert failures_proc_domT_lm[of X n])
apply (drule_tac x="s" in spec)
apply (drule_tac x="tail_traces (fstF DF) a" in spec)

apply (simp add: in_tail_traces)
apply (simp add: appt_assoc_sym)

apply (simp add: head_failures_def)
apply (rule_tac x="t @t [Tick]t" in exI)
apply (rule_tac x="X" in exI)
apply (simp add: appt_assoc_sym)
apply (simp add: pairF_domF_T3)
done

(*----------------------------*
 |       inv_semT lemma       |
 *----------------------------*)

(* traces(proc_setF) => fst DF (lm) *)

lemma inv_semF_to_T_lm:
   "ALL DF t. t :t traces (proc_setF n DF) --> t :t fstF DF"
apply (induct_tac n)

(* base *)
 apply (simp add: in_traces)
 apply (intro allI impI)
 apply (elim conjE exE)

 apply (erule disjE, simp)
 apply (elim conjE exE)
 apply (simp)

(* step *)
 apply (intro allI impI)
 apply (simp add: in_traces)
 apply (elim conjE exE disjE bexE)
 apply (simp_all)
 apply (drule_tac x="(tail_traces (fstF DF) a ,, tail_failures (sndF DF) a)" in spec)
 apply (drule_tac x="s" in spec)
 apply (simp)
 apply (simp add: tail_traces_failures_domF pairF_fstF)
 apply (simp add: in_tail_traces head_failures_traces)
done

(* traces(proc_setF) => fst DF *)

lemma inv_semF_to_T:
   "t :t traces (proc_setF n DF) ==> t :t fstF DF"
by (simp add: inv_semF_to_T_lm)

(* traces(proc_setF) => fst DF (lm) *)

lemma inv_semT_to_F_lm:
   "ALL DF s X.
    (s, X) :f failures (proc_domT n (fstF DF)) --> (s, X) :f sndF DF"
apply (induct_tac n)

(* base *)
 apply (simp add: in_failures)

(* step *)
 apply (intro allI impI)
 apply (simp add: in_failures)
 apply (elim conjE exE disjE bexE)
 apply (simp_all)

  apply (drule_tac x="(tail_traces (fstF DF) a ,, tail_failures (sndF DF) a)" in spec)
  apply (drule_tac x="sa" in spec)
  apply (drule_tac x="X" in spec)
  apply (simp add: tail_traces_failures_domF pairF_fstF)
  apply (simp add: tail_traces_failures_domF pairF_sndF)
  apply (simp add: head_tail_failures)
  apply (simp add: head_traces_failures)

  apply (simp add: in_traces)
  apply (simp add: in_traces)

  apply (case_tac "[Tick]t :t fstF DF")
   apply (simp add: in_failures)
   apply (erule disjE)
    apply (simp add: pairF_domF_F2_F4)
    apply (simp add: pairF_domF_T3_Tick)

  apply (simp add: in_failures)
done

(* traces(proc_setF) => fst DF *)

lemma inv_semT_to_F:
  "(s, X) :f failures (proc_domT n (fstF DF)) ==> (s, X) :f sndF DF"
by (simp add: inv_semT_to_F_lm)

(* failures(proc_setF) => snd DF (lm) *)

lemma inv_semF_to_F_lm:
   "ALL DF s X. (s, X) :f failures (proc_setF n DF) --> (s, X) :f sndF DF"
apply (induct_tac n)

(* base *)
 apply (simp add: in_failures)
 apply (intro allI impI)
 apply (elim conjE exE)
 apply (rule memF_F2)
 apply (simp)
 apply (simp add: Evset_def)
 apply (blast)

(* step *)
 apply (intro allI impI)
 apply (simp add: in_failures)

 apply (elim conjE exE bexE disjE)
 apply (simp_all)

  apply (drule_tac x="(tail_traces (fstF DF) a ,, tail_failures (sndF DF) a)" in spec)
  apply (drule_tac x="sa" in spec)
  apply (drule_tac x="X" in spec)
  apply (simp add: tail_traces_failures_domF pairF_fstF)
  apply (simp add: tail_traces_failures_domF pairF_sndF)
  apply (simp add: head_tail_failures)

  apply (simp add: in_traces)
  apply (simp add: in_traces)
done

lemma inv_semF_to_F:
   "(s, X) :f failures (proc_setF n DF) ==> (s, X) :f sndF DF"
by (simp add: inv_semF_to_F_lm)

(* sndF DF => failures (proc_setF) lm *)

lemma F_inv_semF_lm:
   "ALL DF X. ((s, X) :f sndF DF & noTick s & (Tick : X | s @t [Tick]t ~:t fstF DF))
             --> (s, X) :f failures (proc_setF (lengtht s) DF)"
apply (induct_tac s rule: induct_trace)

 (* []t *)
 apply (intro allI impI)
 apply (simp add: in_failures)
 apply (rule_tac x="{a. Ev a ~: X & [Ev a]t :t fstF DF}" in exI)
 apply (rule)

  apply (elim conjE)
  apply (erule disjE)

   (* Tick : X *)
   apply (rule_tac x="X Un {Ev a|a. [Ev a]t ~:t fstF DF}" in exI)
   apply (simp)
   apply (rule)

    apply (rule pairF_domF_F3)
     apply (simp)
     apply (simp)
     apply (force)
    apply (simp add: Evset_def not_Tick_to_Ev)
    apply (force)

   (* Tick : X *)
   apply (rule_tac x="X Un ({Tick} Un {Ev a|a. [Ev a]t ~:t fstF DF})" in exI)
   apply (rule)

    apply (rule pairF_domF_F3)
     apply (simp)
     apply (simp)
     apply (force)

    apply (simp add: Evset_def not_Tick_to_Ev)
    apply (force)
    apply (force)

 (* [Tick]t *)
 apply (simp)

 (* [Ev a]t @t s *)
 apply (intro allI impI)
 apply (simp add: in_failures)
 apply (simp add: head_tail_failures)

 apply (drule_tac x="(tail_traces (fstF DF) a ,, tail_failures (sndF DF) a)" in spec)
 apply (drule_tac x="X" in spec)
 apply (simp add: tail_traces_failures_domF pairF_fstF pairF_sndF)
 apply (elim conjE exE disjE)
 apply (simp)

 apply (simp add: appt_assoc)
 apply (simp add: head_failures_traces head_tail_traces)
done

(* sndF DF => failures (proc_setF) *)

lemma F_inv_semF:
   "[| (s, X) :f sndF DF ; noTick s ; Tick : X | s @t [Tick]t ~:t fstF DF |]
    ==> (s, X) :f failures (proc_setF (lengtht s) DF)"
by (simp add: F_inv_semF_lm)

(* sndF DF => failures (proc_domT) (noTick) lm *)

declare lengtht_app_event_Suc_head [simp del]
declare lengtht_app_decompo1       [simp del]
declare lengtht_app_decompo2       [simp del]
declare proc_domT.simps            [simp del]

lemma F_inv_semT_noTick_lm:
   "ALL DF X. (s, X) :f sndF DF & noTick s & Tick ~: X & s @t [Tick]t :t fstF DF
    --> (s, X) :f failures (proc_domT (Suc (lengtht s)) (fstF DF))"
apply (induct_tac s rule: induct_trace)

 (* []t *)
 apply (intro allI impI)
 apply (simp add: proc_domT.simps in_failures)
 apply (simp add: Evset_def)
 apply (fast)

 (* [Tick]t *)
 apply (intro allI impI)
 apply (simp add: proc_domT.simps in_failures)

 (* [Ev a]t @t s *)
 apply (intro allI impI)
 apply (simp (no_asm) add: proc_domT.simps)
 apply (simp add: in_failures)
 apply (elim conjE)
 apply (simp add: appt_assoc)
 apply (simp add: head_tail_traces)

 apply (drule_tac x="(tail_traces (fstF DF) a ,, tail_failures (sndF DF) a)" in spec)
 apply (drule_tac x="X" in spec)
 apply (simp)
 apply (elim conjE)
 apply (simp add: tail_traces_failures_domF pairF_fstF pairF_sndF)
 apply (simp add: head_tail_failures)
 apply (simp add: lengtht_app_event_Suc_head)
done

(* sndF DF => failures (proc_domT) noTick *)

lemma F_inv_semT_noTick:
   "[| (s, X) :f sndF DF ; noTick s ; Tick ~: X ; s @t [Tick]t :t fstF DF |]
    ==> (s, X) :f failures (proc_domT (Suc (lengtht s)) (fstF DF))"
by (simp add: F_inv_semT_noTick_lm)

declare lengtht_app_event_Suc_head [simp]
declare lengtht_app_decompo1       [simp]
declare lengtht_app_decompo2       [simp]
declare proc_domT.simps            [simp]

(* sndF DF => failures (proc_domT) (Tick) lm *)

lemma F_inv_semT_Tick_lm:
   "ALL DF X. ((s, X) :f sndF DF & ~ noTick s)
    --> (s, X) :f failures (proc_domT (lengtht s) (fstF DF))"
apply (induct_tac s rule: induct_trace)

 (* []t *)
 apply (intro allI impI)
 apply (simp add: in_failures)

 (* [Tick]t *)
 apply (intro allI impI)
 apply (simp add: in_failures)
 apply (simp add: pairF_domF_T2)

 (* [Ev a]t @t s *)
 apply (intro allI impI)
 apply (simp add: in_failures)
 apply (elim conjE)
 apply (simp add: head_tail_failures)

 apply (drule_tac x="(tail_traces (fstF DF) a ,, tail_failures (sndF DF) a)" in spec)
 apply (drule_tac x="X" in spec)
 apply (simp add: tail_traces_failures_domF pairF_fstF pairF_sndF)
 apply (elim conjE)
 apply (simp add: head_failures_traces)
done

(* sndF DF => failures (proc_domT) noTick *)

lemma F_inv_semT_Tick:
   "[| (s, X) :f sndF DF; ~ noTick s |]
           ==> (s, X) :f failures (proc_domT (lengtht s) (fstF DF))"
by (simp add: F_inv_semT_Tick_lm)

(*==================================================*
 |              inv_semF lemma (main)               |
 *==================================================*)

lemma semF_inv_semF: "[[ inv[[DF]]F ]]F = DF"
apply (simp add: inv_semF_def)
apply (rule order_antisym)

 (* <= *)
 apply (simp add: subdomF_decompo)
 apply (rule conjI)

  (* inv_sem <= fstF DF *)
  apply (rule)
  apply (simp add: in_traces)
  apply (simp add: semT_inv_semT[simplified semT_def])
  apply (erule disjE, simp)
  apply (erule disjE, simp)
  apply (elim conjE exE)
  apply (simp add: inv_semF_to_T)

  (* inv_sem <= sndF DF *)
  apply (rule)
  apply (simp add: in_failures)
  apply (erule disjE)

   (* inv_sem T <= sndF DF *)
   apply (simp add: inv_semT_def)
   apply (simp add: in_failures)
   apply (erule exE)
   apply (simp add: inv_semT_to_F)

   (* inv_sem F <= sndF DF *)
   apply (erule exE)
   apply (simp add: inv_semF_to_F)

 (* <= *)
 apply (simp add: subdomF_decompo)
 apply (rule conjI)

  (* fstF DF <= inv_sem *)
  apply (rule)
  apply (simp add: in_traces)
  apply (simp add: semT_inv_semT[simplified semT_def])

  (* sndF DF <= inv_sem *)
  apply (rule)
  apply (simp add: in_failures)
  apply (case_tac "noTick s & (Tick : X | s @t [Tick]t ~:t fstF DF)")
   apply (rule disjI2)
   apply (rule_tac x="lengtht s" in exI)
   apply (simp add: F_inv_semF)

   apply (simp)
   apply (rule disjI1)
   apply (simp add: inv_semT_def)
   apply (simp add: in_failures)
   apply (case_tac "noTick s")

    (* noTick s & s @t [Tick]t :t fstF DF *)
    apply (simp)
    apply (rule_tac x="Suc (lengtht s)" in exI)
    apply (simp add: F_inv_semT_noTick del: proc_domT.simps)

    (* ~ noTick s *)
    apply (rule_tac x="lengtht s" in exI)
    apply (simp add: F_inv_semT_Tick)
done

(*----------------------------*
 |      Full abstraction      |
 *----------------------------*)

theorem full_abs_domF: "ALL DF. EX P. [[P]]F = DF"
apply (rule allI)
apply (rule_tac x="inv[[DF]]F" in exI)
apply (simp add: semF_inv_semF)
done

(*----------------------------*
 |   failures and inv[[DF]]F  |
 *----------------------------*)

lemma failures_inv_semF: "failures inv[[DF]]F = sndF DF"
apply (insert semF_inv_semF[of DF])
apply (simp add: semF_def)
apply (simp add: eqF_decompo)
done

(*----------------------------*
 |    traces and inv[[DF]]F   |
 *----------------------------*)

lemma traces_inv_semF: "traces inv[[DF]]F = fstF DF"
apply (insert semF_inv_semF[of DF])
apply (simp add: semF_def)
apply (simp add: eqF_decompo)
done

lemma traces_inv_semT_semF:
  "traces inv[[fstF DF]]T = traces inv[[DF]]F"
apply (simp add: traces_inv_semT)
apply (simp add: traces_inv_semF)
done

(****************** to add them again ******************)

declare disj_not1   [simp]

end
