           (*-------------------------------------------*
            |        CSP-Prover on Isabelle2004         |
            |               December 2004               |
            |                   June 2005  (modified)   |
            |              September 2005  (modified)   |
            |                                           |
            |        CSP-Prover on Isabelle2005         |
            |               November 2005  (modified)   |
            |                                           |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_F_law = CSP_F_law_SKIP  + CSP_F_law_ref       +
                   CSP_F_law_dist  + CSP_F_law_alpha_par +
                   CSP_F_law_step  + CSP_F_law_index_par +
                   CSP_F_law_fp    + CSP_T_law:

(*****************************************************************

         1. Conditional
         2. Internal ... simp
         3. External ... simp
         4. Mix      ... simp
         5. Extended reflex

 *****************************************************************)

(*****************************************************************
                Extended reflex by empty sets
 *****************************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspF_Ext_pre_choice_empty: 
  "? :{} -> Pf =F ? x:{} -> DIV"
apply (simp add: cspF_cspT_semantics)
apply (simp add: failures_def)
done

lemma cspF_Ext_pre_choice_STOP:
   "? :{} -> Pf =F STOP"
apply (rule cspF_rw_right)
apply (rule cspF_STOP_step)
apply (rule cspF_Ext_pre_choice_empty)
done

lemmas cspF_Ext_pre_choice_STOP_sym = cspF_Ext_pre_choice_STOP[THEN cspF_sym]

lemma cspF_Rep_int_choice0_DIV:
   "!! :{} .. Pf =F DIV"
apply (simp add: cspF_cspT_semantics)
apply (simp add: failures_def)
apply (simp add: empF_def)
done

lemma cspF_Rep_int_choice_fun_DIV:
   "inj f ==> !!<f> :{} .. Pf =F DIV"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_Rep_int_choice_fun_DIV)
apply (simp add: failures_def)
apply (simp add: empF_def)
done

lemma cspF_Rep_int_choice1_DIV:
   "! :{} .. Pf =F DIV"
by (simp add: cspF_Rep_int_choice_fun_DIV)

lemma cspF_Rep_int_choice2_DIV:
   "!set :{} .. Pf =F DIV"
by (simp add: cspF_Rep_int_choice_fun_DIV)

lemma cspF_Rep_int_choice3_DIV:
   "!nat :{} .. Pf =F DIV"
by (simp add: cspF_Rep_int_choice_fun_DIV)

lemmas cspF_Rep_int_choice_DIV = cspF_Rep_int_choice0_DIV
                                 cspF_Rep_int_choice1_DIV
                                 cspF_Rep_int_choice2_DIV
                                 cspF_Rep_int_choice3_DIV

lemmas cspF_Rep_int_choice_DIV_sym = cspF_Rep_int_choice0_DIV[THEN cspF_sym]

lemma cspF_Ext_pre_choice_empty_reflex:
   "? :{} -> Pf =F ? :{} -> Qf"
apply (rule cspF_trans[of _ "STOP"])
apply (simp add: cspF_Ext_pre_choice_STOP)
apply (simp add: cspF_Ext_pre_choice_STOP_sym)
done

lemma cspF_Rep_int_choice0_empty_reflex:
   "!! :{} .. Pf =F !! :{} .. Qf"
apply (rule cspF_trans[of _ "DIV"])
apply (simp add: cspF_Rep_int_choice_DIV)
apply (simp add: cspF_Rep_int_choice_DIV_sym)
done

lemma cspF_Rep_int_choice_fun_empty_reflex:
   "inj f ==> !!<f> :{} .. Pf =F !!<f> :{} .. Qf"
apply (simp add: Rep_int_choice_fun_def)
apply (simp add: cspF_Rep_int_choice0_empty_reflex)
done

lemma cspF_Rep_int_choice1_empty_reflex:
   "! :{} .. Pf =F ! :{} .. Qf"
by (simp add: cspF_Rep_int_choice_fun_empty_reflex)

lemma cspF_Rep_int_choice2_empty_reflex:
   "!set :{} .. Pf =F !set :{} .. Qf"
by (simp add: cspF_Rep_int_choice_fun_empty_reflex)

lemma cspF_Rep_int_choice3_empty_reflex:
   "!nat :{} .. Pf =F !nat :{} .. Qf"
by (simp add: cspF_Rep_int_choice_fun_empty_reflex)

lemmas cspF_Rep_int_choice_empty_reflex 
     = cspF_Rep_int_choice0_empty_reflex
       cspF_Rep_int_choice1_empty_reflex
       cspF_Rep_int_choice2_empty_reflex
       cspF_Rep_int_choice3_empty_reflex

lemmas cspF_reflex_ex = cspF_Ext_pre_choice_STOP
                        cspF_Ext_pre_choice_STOP_sym
                        cspF_Rep_int_choice_DIV
                        cspF_Rep_int_choice_DIV_sym
                        cspF_Ext_pre_choice_empty_reflex
                        cspF_Rep_int_choice_empty_reflex

declare cspF_reflex_ex [simp]

(*********************************************************
                       IF bool
 *********************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspF_Conditional_split: 
  "IF b THEN P ELSE Q =F (if b then P else Q)"
apply (simp add: cspF_semantics)
apply (simp add: traces_def)
apply (simp add: failures_def)
done

lemma cspF_Conditional_True:
  "IF True THEN P ELSE Q =F P"
apply (rule cspF_rw_left)
apply (rule cspF_Conditional_split)
by (simp)

lemma cspF_Conditional_False:
  "IF False THEN P ELSE Q =F Q"
apply (rule cspF_rw_left)
apply (rule cspF_Conditional_split)
by (simp)

lemmas cspF_Conditional = cspF_Conditional_True cspF_Conditional_False

(*****************************************************************
                       Internal ... simp                         
 *****************************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspF_DIV_unwind:
   "DIV =F ! x:{} .. DIV"
apply (simp add: cspF_semantics)
apply (simp add: traces_def)
apply (simp add: failures_def)
apply (simp add: empF_def)
done

(*** ! :{a} ***)

lemma cspF_Rep_int_choice0_unique:
  "!! :{a} .. Pf =F Pf a"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_Rep_int_choice0_unique)
apply (rule order_antisym)

 apply (rule, simp add: in_failures)
 apply (rule, simp add: in_failures)
done

lemma cspF_Rep_int_choice_fun_unique:
  "inj f ==> !!<f> :{a} .. Pf =F Pf a"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_Rep_int_choice_fun_unique)
apply (simp add: Rep_int_choice_fun_def)
apply (rule order_antisym)

 apply (rule, simp add: in_failures)
 apply (rule, simp add: in_failures)
done

lemma cspF_Rep_int_choice1_unique:
  "! :{a} .. Pf =F Pf a"
by (simp add: cspF_Rep_int_choice_fun_unique)

lemma cspF_Rep_int_choice2_unique:
  "!set :{a} .. Pf =F Pf a"
by (simp add: cspF_Rep_int_choice_fun_unique)

lemma cspF_Rep_int_choice3_unique:
  "!nat :{a} .. Pf =F Pf a"
by (simp add: cspF_Rep_int_choice_fun_unique)

lemmas cspF_Rep_int_choice_unique = cspF_Rep_int_choice0_unique
                                    cspF_Rep_int_choice1_unique
                                    cspF_Rep_int_choice2_unique
                                    cspF_Rep_int_choice3_unique

(* Int choice idempotence *)

lemma cspF_Int_choice_idem: 
  "P |~| P =F P"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_Int_choice_idem)
apply (rule order_antisym)
 apply (rule, simp add: in_failures)+
done

lemma cspF_Int_choice_idem_weak: 
  "P =F Q ==> P |~| Q =F P"
apply (rule cspF_rw_left)
apply (rule cspF_decompo)
apply (rule cspF_reflex)
apply (erule cspF_symE)
apply (simp)
apply (simp add: cspF_Int_choice_idem)
done

lemmas cspF_Int_choice_rule = cspF_DIV_unwind 
                              cspF_Rep_int_choice_unique 
                              cspF_Int_choice_idem

lemmas cspF_Int_choice_simp = cspF_DIV_unwind 
                              cspF_Rep_int_choice_unique 
                              cspF_Int_choice_idem_weak

(*****************************************************************
                      External ... simp
 *****************************************************************)

(*** STOP [+] P ***)

lemma cspF_Ext_choice_unit_l: 
  "STOP [+] P =F P"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_Ext_choice_unit_l)
apply (rule order_antisym)
 apply (rule, simp add: in_traces in_failures)
 apply (elim conjE disjE)
 apply (simp_all)
 apply (rule proc_F2_F4)
 apply (simp_all)

 apply (rule, simp add: in_failures)
done

lemma cspF_Ext_choice_unit_l_hnf: 
  "? :{} -> Qf [+] P =F P"
apply (rule cspF_rw_left)
apply (rule cspF_decompo)
apply (rule cspF_Ext_pre_choice_STOP)
apply (rule cspF_reflex)
apply (simp add: cspF_Ext_choice_unit_l)
done

lemma cspF_Ext_choice_STOP_r: 
  "P [+] STOP =F P"
apply (rule cspF_rw_left)
apply (rule cspF_Ext_choice_commut)
apply (simp add: cspF_Ext_choice_unit_l)
done

lemma cspF_Ext_choice_STOP_r_hnf: 
  "P [+] ? :{} -> Qf =F P"
apply (rule cspF_rw_left)
apply (rule cspF_Ext_choice_commut)
apply (simp add: cspF_Ext_choice_unit_l_hnf)
done

(*** weak ***)

lemma cspF_Ext_choice_unit_l_weak: 
  "Q =F STOP 
   ==> Q [+] P =F P"
apply (rule cspF_rw_left)
apply (rule cspF_decompo)
apply (assumption)
apply (rule cspF_reflex)
apply (simp add: cspF_Ext_choice_unit_l)
done

lemma cspF_Ext_choice_unit_r_weak: 
  "Q =F STOP 
   ==> P [+] Q =F P"
apply (rule cspF_rw_left)
apply (rule cspF_commut)
apply (simp add: cspF_Ext_choice_unit_l_weak)
done

lemmas cspF_Ext_choice_rule = cspF_Ext_choice_unit_l
                              cspF_Ext_choice_unit_l_hnf
                              cspF_Ext_choice_STOP_r
                              cspF_Ext_choice_STOP_r_hnf

lemmas cspF_Ext_choice_simp = cspF_Ext_choice_unit_l_weak
                              cspF_Ext_choice_unit_r_weak

(*****************************************************************
                      Mix ... simp
 *****************************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

(*** <= Timeout ***)

lemma cspF_Timeout_right:
  "[| P <=F Q1 ; P <=F Q2 |] ==> P <=F Q1 [> Q2"
apply (rule cspF_rw_right)
apply (rule cspF_dist)
apply (rule cspF_Int_choice_right)
apply (rule cspF_Ext_choice_right, simp_all)
apply (rule cspF_rw_right)
apply (rule cspF_Ext_choice_unit_l, simp_all)
done

(*** STOP [> P  =  P ***)

lemma cspF_STOP_Timeout:
  "STOP [> P =F P"
apply (rule cspF_rw_left)
apply (rule cspF_dist)
apply (rule cspF_rw_left)
apply (rule cspF_Int_choice_idem)
apply (rule cspF_Ext_choice_unit_l_weak)
by (simp)

lemma cspF_STOP_Timeout_hnf:
  "(? :{} -> Qf1 |~| ? :{} -> Qf2) [+] Q =F Q"
apply (rule cspF_rw_left)
apply (rule cspF_decompo)
apply (rule cspF_decompo)
apply (rule cspF_Ext_pre_choice_STOP)
apply (rule cspF_Ext_pre_choice_STOP)
apply (rule cspF_reflex)
apply (simp add: cspF_STOP_Timeout)
done

lemma cspF_STOP_Timeout_weak:
  "[| P1 =F STOP ;
      P2 =F STOP |]
   ==> (P1 |~| P2) [+] Q =F Q"
apply (rule cspF_rw_left)
apply (rule cspF_decompo)
apply (rule cspF_decompo)
apply (simp_all)
apply (rule cspF_reflex)
apply (simp add: cspF_STOP_Timeout)
done

lemmas cspF_Mix_choice_rule = cspF_STOP_Timeout cspF_STOP_Timeout_hnf

lemmas cspF_Mix_choice_simp = cspF_STOP_Timeout_weak

(*-----------------------------*
 |      simp rule for cspF     |
 *-----------------------------*)

lemmas cspF_choice_rule  = cspF_Int_choice_rule cspF_Ext_choice_rule cspF_Mix_choice_rule
lemmas cspF_choice_simp  = cspF_Int_choice_simp cspF_Ext_choice_simp cspF_Mix_choice_simp

end
