           (*-------------------------------------------*
            |        CSP-Prover on Isabelle2004         |
            |               December 2004               |
            |                   June 2005  (modified)   |
            |                                           |
            |        CSP-Prover on Isabelle2005         |
            |                October 2005  (modified)   |
            |                                           |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_F_law_decompo = CSP_F_domain + CSP_T_law_decompo:

(*------------------------------------------------*
 |                                                |
 |      laws for monotonicity and congruence      |
 |                                                |
 *------------------------------------------------*)

(*********************************************************
                        Act_prefix mono
 *********************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspF_Act_prefix_mono:
  "[| a = b ; P <=F Q |] ==> a -> P <=F b -> Q"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_Act_prefix_mono)
apply (simp add: subsetF_iff)
apply (intro allI impI)
apply (simp add: in_failures)
apply (fast)
done

lemma cspF_Act_prefix_cong:
  "[| a = b ; P =F Q |] ==> a -> P =F b -> Q"
apply (simp add: cspF_eq_ref_iff)
apply (simp add: cspF_Act_prefix_mono)
done

(*********************************************************
                   Ext_pre_choice mono 
 *********************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspF_Ext_pre_choice_mono:
  "[| X = Y ; !! a. a:Y ==> Pf a <=F Qf a |] ==> ? :X -> Pf <=F ? :Y -> Qf"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_Ext_pre_choice_mono)
apply (simp add: subsetF_iff)
apply (intro allI impI)
apply (simp add: in_failures)
apply (fast)
done

lemma cspF_Ext_pre_choice_cong:
  "[| X = Y ; !! a. a:Y ==> Pf a =F Qf a |] ==> ? :X -> Pf =F ? :Y -> Qf"
by (simp add: cspF_eq_ref_iff cspF_Ext_pre_choice_mono)

(*********************************************************
                      Ext choice mono
 *********************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspF_Ext_choice_mono:
  "[| P1 <=F Q1 ; P2 <=F Q2 |] ==> P1 [+] P2 <=F Q1 [+] Q2"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_Ext_choice_mono)
apply (simp add: cspT_semantics)
apply (simp add: subdomT_iff)
apply (simp add: subsetF_iff)
apply (intro allI impI)
apply (simp add: in_failures)
apply (elim conjE disjE)
apply (force)+
done

lemma cspF_Ext_choice_cong:
  "[| P1 =F Q1 ; P2 =F Q2 |] ==> P1 [+] P2 =F Q1 [+] Q2"
by (simp add: cspF_eq_ref_iff cspF_Ext_choice_mono)

(*********************************************************
                      Int choice mono
 *********************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspF_Int_choice_mono:
  "[| P1 <=F Q1 ; P2 <=F Q2 |] ==> P1 |~| P2 <=F Q1 |~| Q2"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_Int_choice_mono)
apply (simp add: subsetF_iff)
apply (intro allI impI)
apply (simp add: in_failures)
apply (fast)
done

lemma cspF_Int_choice_cong:
  "[| P1 =F Q1 ; P2 =F Q2 |] ==> P1 |~| P2 =F Q1 |~| Q2"
by (simp add: cspF_eq_ref_iff cspF_Int_choice_mono)

(*********************************************************
               replicated internal choice
 *********************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspF_Rep_int_choice_mono0:
  "[| X = Y ; !! a. a:Y ==> Pf a <=F Qf a |]
      ==> !! :X .. Pf <=F !! :Y .. Qf"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_Rep_int_choice_mono0)
apply (simp add: subsetF_iff)
apply (intro allI impI)
apply (simp add: in_failures)
apply (fast)
done

lemma cspF_Rep_int_choice_cong0:
  "[| X = Y ; !! a. a:Y ==> Pf a =F Qf a |]
      ==> !! :X .. Pf =F !! :Y .. Qf"
by (simp add: cspF_eq_ref_iff cspF_Rep_int_choice_mono0)

(*** fun ***)

lemma cspF_Rep_int_choice_mono_fun:
  "[| inj f ; X = Y ; !! a. a:Y ==> Pf a <=F Qf a |]
      ==> !!<f> :X .. Pf <=F !!<f> :Y .. Qf"
apply (simp add: Rep_int_choice_fun_def)
apply (rule cspF_Rep_int_choice_mono0)
apply (auto)
done

lemma cspF_Rep_int_choice_cong_fun:
  "[| inj f ; X = Y ; !! a. a:Y ==> Pf a =F Qf a |]
      ==> !!<f> :X .. Pf =F !!<f> :Y .. Qf"
by (simp add: cspF_eq_ref_iff cspF_Rep_int_choice_mono_fun)

(*** Tri1 ***)

lemma cspF_Rep_int_choice_mono_Tri1:
  "[| X = Y ; !! a. a:Y ==> Pf a <=F Qf a |]
      ==> ! :X .. Pf <=F ! :Y .. Qf"
by (simp add: cspF_Rep_int_choice_mono_fun)

lemma cspF_Rep_int_choice_cong_Tri1:
  "[| X = Y ; !! a. a:Y ==> Pf a =F Qf a |]
      ==> ! :X .. Pf =F ! :Y .. Qf"
by (simp add: cspF_Rep_int_choice_cong_fun)

(*** Tri2 ***)

lemma cspF_Rep_int_choice_mono_Tri2:
  "[| X = Y ; !! a. a:Y ==> Pf a <=F Qf a |]
      ==> !set :X .. Pf <=F !set :Y .. Qf"
by (simp add: cspF_Rep_int_choice_mono_fun)

lemma cspF_Rep_int_choice_cong_Tri2:
  "[| X = Y ; !! a. a:Y ==> Pf a =F Qf a |]
      ==> !set :X .. Pf =F !set :Y .. Qf"
by (simp add: cspF_Rep_int_choice_cong_fun)

(*** Tri3 ***)

lemma cspF_Rep_int_choice_mono_Tri3:
  "[| X = Y ; !! a. a:Y ==> Pf a <=F Qf a |]
      ==> !nat :X .. Pf <=F !nat :Y .. Qf"
by (simp add: cspF_Rep_int_choice_mono_fun)

lemma cspF_Rep_int_choice_cong_Tri3:
  "[| X = Y ; !! a. a:Y ==> Pf a =F Qf a |]
      ==> !nat :X .. Pf =F !nat :Y .. Qf"
by (simp add: cspF_Rep_int_choice_cong_fun)

(*** f ***)

lemma cspF_Rep_int_choice_mono_f:
  "[| inj f ; X = Y ; !! a. a:Y ==> Pf a <=F Qf a |]
      ==> !<f> :X .. Pf <=F !<f> :Y .. Qf"
apply (simp add: Rep_int_choice_f_def)
apply (rule cspF_Rep_int_choice_mono_Tri1)
apply (auto)
done

lemma cspF_Rep_int_choice_cong_f:
  "[| inj f ; X = Y ; !! a. a:Y ==> Pf a =F Qf a |]
      ==> !<f> :X .. Pf =F !<f> :Y .. Qf"
apply (simp add: Rep_int_choice_f_def)
apply (rule cspF_Rep_int_choice_cong_Tri1)
apply (auto)
done

lemmas cspF_Rep_int_choice_mono = cspF_Rep_int_choice_mono0
                                  cspF_Rep_int_choice_mono_Tri1
                                  cspF_Rep_int_choice_mono_Tri2
                                  cspF_Rep_int_choice_mono_Tri3
                                  cspF_Rep_int_choice_mono_f

lemmas cspF_Rep_int_choice_cong = cspF_Rep_int_choice_cong0
                                  cspF_Rep_int_choice_cong_Tri1
                                  cspF_Rep_int_choice_cong_Tri2
                                  cspF_Rep_int_choice_cong_Tri3
                                  cspF_Rep_int_choice_cong_f

(*********************************************************
                   Conditional mono
 *********************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspF_Conditional_mono:
  "[| b1 = b2 ; P1 <=F Q1 ; P2 <=F Q2 |]
           ==> IF b1 THEN P1 ELSE P2 <=F
               IF b2 THEN Q1 ELSE Q2"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_Conditional_mono)
apply (simp add: subsetF_iff)
apply (intro allI impI)
apply (simp add: in_failures)
done

lemma cspF_Conditional_cong:
  "[| b1 = b2 ; P1 =F Q1 ;  P2 =F Q2 |]
           ==> IF b1 THEN P1 ELSE P2 =F
               IF b2 THEN Q1 ELSE Q2"
by (simp add: cspF_eq_ref_iff cspF_Conditional_mono)

(*********************************************************
                     Parallel mono 
 *********************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspF_Parallel_mono:
  "[| X = Y ; P1 <=F Q1 ; P2 <=F Q2 |]
           ==> P1 |[X]| P2 <=F Q1 |[Y]| Q2"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_Parallel_mono)
apply (simp add: subsetF_iff)
apply (intro allI impI)
apply (simp add: in_failures)
apply (fast)
done

lemma cspF_Parallel_cong:
  "[| X = Y ; P1 =F Q1 ; P2 =F Q2 |]
           ==> P1 |[X]| P2 =F Q1 |[Y]| Q2"
by (simp add: cspF_eq_ref_iff cspF_Parallel_mono)

(*********************************************************
                        Hiding mono
 *********************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspF_Hiding_mono:
  "[| X = Y ; P <=F Q |] ==> P -- X <=F Q -- Y"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_Hiding_mono)
apply (simp add: subsetF_iff)
apply (intro allI impI)
apply (simp add: in_failures)
apply (fast)
done

lemma cspF_Hiding_cong:
  "[| X = Y ; P =F Q |] ==> P -- X =F Q -- Y"
by (simp add: cspF_eq_ref_iff cspF_Hiding_mono)

(*********************************************************
                        Renaming mono
 *********************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspF_Renaming_mono:
  "[| r1 = r2 ; P <=F Q |] ==> P [[r1]] <=F Q [[r2]]"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_Renaming_mono)
apply (simp add: subsetF_iff)
apply (intro allI impI)
apply (simp add: in_failures)
apply (fast)
done

lemma cspF_Renaming_cong:
  "[| r1 = r2 ; P =F Q |] ==> P [[r1]] =F Q [[r2]]"
by (simp add: cspF_eq_ref_iff cspF_Renaming_mono)

(*********************************************************
               Sequential composition mono 
 *********************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspF_Seq_compo_mono:
  "[| P1 <=F Q1 ; P2 <=F Q2 |]
           ==> P1 ;; P2 <=F Q1 ;; Q2"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_Seq_compo_mono)
apply (simp add: cspT_semantics)
apply (simp add: subdomT_iff)
apply (simp add: subsetF_iff)
apply (intro allI impI)
apply (simp add: in_failures)
apply (fast)
done

lemma cspF_Seq_compo_cong:
  "[| P1 =F Q1 ; P2 =F Q2 |] ==> P1 ;; P2 =F Q1 ;; Q2"
by (simp add: cspF_eq_ref_iff cspF_Seq_compo_mono)

(*********************************************************
                        Timeout mono
 *********************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspF_Timeout_mono:
  "[| P1 <=F Q1 ; P2 <=F Q2 |]
      ==> P1 [> P2 <=F Q1 [> Q2"
apply (rule cspF_Ext_choice_mono)
apply (rule cspF_Int_choice_mono)
by (simp_all)

lemma cspF_Timeout_cong:
  "[| P1 =F Q1 ; P2 =F Q2 |]
      ==> P1 [> P2 =F Q1 [> Q2"
by (simp add: cspF_eq_ref_iff cspF_Timeout_mono)

(*------------------------------------------------------*
 |                       alias                          |
 *------------------------------------------------------*)

lemmas cspF_free_mono =
   cspF_Ext_choice_mono cspF_Int_choice_mono cspF_Parallel_mono
   cspF_Hiding_mono cspF_Renaming_mono cspF_Seq_compo_mono

lemmas cspF_mono = cspF_free_mono
   cspF_Act_prefix_mono cspF_Ext_pre_choice_mono 
   cspF_Rep_int_choice_mono cspF_Conditional_mono 

lemmas cspF_free_cong =
   cspF_Ext_choice_cong cspF_Int_choice_cong cspF_Parallel_cong
   cspF_Hiding_cong cspF_Renaming_cong cspF_Seq_compo_cong

lemmas cspF_cong = cspF_free_cong
   cspF_Act_prefix_cong cspF_Ext_pre_choice_cong 
   cspF_Rep_int_choice_cong cspF_Conditional_cong 

lemmas cspF_free_decompo = cspF_free_mono cspF_free_cong
lemmas cspF_decompo = cspF_mono cspF_cong

lemmas cspF_rm_head_mono = cspF_Act_prefix_mono cspF_Ext_pre_choice_mono 
lemmas cspF_rm_head_cong = cspF_Act_prefix_cong cspF_Ext_pre_choice_cong
lemmas cspF_rm_head      = cspF_rm_head_mono cspF_rm_head_cong

end
