           (*-------------------------------------------*
            |        CSP-Prover on Isabelle2004         |
            |               December 2004               |
            |                   July 2005  (modified)   |
            |              September 2005  (modified)   |
            |                                           |
            |        CSP-Prover on Isabelle2005         |
            |               November 2005  (modified)   |
            |                                           |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_F_law_dist = CSP_F_law_commut + CSP_F_law_decompo
                      + CSP_T_law_dist:

(*****************************************************************

      distribution of internal choice over ...

         1. (P1 |~| P2) [+] Q
         2. Q [+] (P1 |~| P2)
         3. (P1 |~| P2) |[X]| Q
         4. Q |[X]| (P1 |~| P2)
         5. (P1 |~| P2) -- X
         6. (P1 |~| P2) [[r]]

 *****************************************************************)

(*********************************************************
                dist law for Ext_choice (l)
 *********************************************************)

lemma cspF_Ext_choice_dist_l: 
  "(P1 |~| P2) [+] Q =F
   (P1 [+] Q) |~| (P2 [+] Q)"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_Ext_choice_dist_l)
apply (rule order_antisym)
apply (rule, simp add: in_traces in_failures, fast)+
done

(*********************************************************
                dist law for Ext_choice (r)
 *********************************************************)

lemma cspF_Ext_choice_dist_r: 
  "P [+] (Q1 |~| Q2) =F
   (P [+] Q1) |~| (P [+] Q2)"
apply (rule cspF_rw_left)
apply (rule cspF_commut)
apply (rule cspF_rw_left)
apply (rule cspF_Ext_choice_dist_l)
apply (rule cspF_decompo)
apply (rule cspF_commut)+
done

(*********************************************************
                dist law for Parallel (l)
 *********************************************************)

lemma cspF_Parallel_dist_l: 
  "(P1 |~| P2) |[X]| Q =F
   (P1 |[X]| Q) |~| (P2 |[X]| Q)"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_Parallel_dist_l)
apply (rule order_antisym)

  apply (rule, simp add: in_failures)
  apply (elim disjE conjE exE)
  apply (rule disjI1)
  apply (rule_tac x="Y" in exI)
  apply (rule_tac x="Z" in exI)
  apply (fast)
  apply (rule disjI2)
  apply (rule_tac x="Y" in exI)
  apply (rule_tac x="Z" in exI)
  apply (fast)

  apply (rule, simp add: in_failures)
  apply (elim disjE conjE exE)
  apply (rule_tac x="Y" in exI)
  apply (rule_tac x="Z" in exI)
  apply (fast)
  apply (rule_tac x="Y" in exI)
  apply (rule_tac x="Z" in exI)
  apply (fast)
done

(*********************************************************
                dist law for Parallel (r)
 *********************************************************)

lemma cspF_Parallel_dist_r: 
  "P |[X]| (Q1 |~| Q2) =F
   (P |[X]| Q1) |~| (P |[X]| Q2)"
apply (rule cspF_rw_left)
apply (rule cspF_commut)
apply (rule cspF_rw_left)
apply (rule cspF_Parallel_dist_l)
apply (rule cspF_decompo)
apply (rule cspF_commut)+
done

(*********************************************************
                dist law for Hiding
 *********************************************************)

lemma cspF_Hiding_dist: 
  "(P1 |~| P2) -- X =F
   (P1 -- X) |~| (P2 -- X)"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_Hiding_dist)
apply (rule order_antisym)
apply (rule, simp add: in_failures, fast)+
done

(*********************************************************
               dist law for Renaming
 *********************************************************)

lemma cspF_Renaming_dist: 
  "(P1 |~| P2) [[r]] =F
   (P1 [[r]]) |~| (P2 [[r]])"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_Renaming_dist)
apply (rule order_antisym)
apply (rule, simp add: in_failures, fast)+
done

(*********************************************************
         dist law for Sequential composition
 *********************************************************)

lemma cspF_Seq_compo_dist: 
  "(P1 |~| P2) ;; Q =F
   (P1 ;; Q) |~| (P2 ;; Q)"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_Seq_compo_dist)
apply (rule order_antisym)
apply (rule, simp add: in_traces in_failures, fast)+
done

(*********************************************************
                     dist laws
 *********************************************************)

lemmas cspF_dist = cspF_Ext_choice_dist_l cspF_Ext_choice_dist_r
                   cspF_Parallel_dist_l   cspF_Parallel_dist_r
                   cspF_Hiding_dist       cspF_Renaming_dist
                   cspF_Seq_compo_dist

(*****************************************************************

      distribution of internal bind over ...

         1. (!! :X .. Pf) [+] Q
         2. Q [+] (!! :X .. Pf)
         3. (!! :X .. Pf) |[X]| Q
         4. Q |[X]| (!! :X .. Pf)
         5. (!! :X .. Pf) -- X
         6. (!! :X .. Pf) [[r]]

 *****************************************************************)

(*********************************************************
                Rep_dist law for Ext_choice (l)
 *********************************************************)

lemma cspF_Ext_choice_Rep_dist0_l: 
  "X ~= {} ==> (!! :X .. Pf) [+] Q =F
               !! x:X .. (Pf x [+] Q)"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_Ext_choice_Rep_dist0_l)
apply (rule order_antisym)
apply (rule, simp add: in_traces in_failures, fast)+
done

(*********************************************************
                Rep_dist0 law for Ext_choice (r)
 *********************************************************)

lemma cspF_Ext_choice_Rep_dist0_r: 
  "X ~= {} ==> P [+] (!! :X .. Qf) =F
               !! x:X .. (P [+] Qf x)"
apply (rule cspF_rw_left)
apply (rule cspF_commut)
apply (rule cspF_rw_left)
apply (rule cspF_Ext_choice_Rep_dist0_l, simp)
apply (rule cspF_decompo, simp)
apply (rule cspF_commut)
done

(*********************************************************
                Rep_dist0 law for Parallel (l)
 *********************************************************)

lemma cspF_Parallel_Rep_dist0_l: 
  "Y ~= {} ==>
     (!! :Y .. Pf) |[X]| Q =F
     !! x:Y .. (Pf x |[X]| Q)"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_Parallel_Rep_dist0_l)
apply (rule order_antisym)

(* domF *)
  apply (rule)
  apply (simp add: in_failures)
  apply (elim conjE exE bexE)
   apply (rule_tac x="a" in bexI)
   apply (rule_tac x="Ya" in exI)
   apply (rule_tac x="Z" in exI)
   apply (fast)
   apply (simp)
  (* *)
  apply (rule)
  apply (simp add: in_failures)
  apply (elim conjE exE bexE)
   apply (rule_tac x="Ya" in exI)
   apply (rule_tac x="Z" in exI)
   apply (fast)
done

(*********************************************************
                Rep_dist0 law for Parallel (r)
 *********************************************************)

lemma cspF_Parallel_Rep_dist0_r: 
  "Y ~= {} ==>
     P |[X]| (!! :Y .. Qf) =F
     !! x:Y .. (P |[X]| Qf x)"
apply (rule cspF_rw_left)
apply (rule cspF_commut)
apply (rule cspF_rw_left)
apply (rule cspF_Parallel_Rep_dist0_l, simp)
apply (rule cspF_decompo, simp)
apply (rule cspF_commut)
done

(*********************************************************
                Rep_dist0 law for Hiding
 *********************************************************)

lemma cspF_Hiding_Rep_dist0: 
  "(!! :Y .. Pf) -- X =F
   !! x:Y .. (Pf x -- X)"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_Hiding_Rep_dist0)
apply (rule order_antisym)
apply (rule, simp add: in_failures, fast)+
done

(*********************************************************
                Rep_dist0 law for Renaming
 *********************************************************)

lemma cspF_Renaming_Rep_dist0: 
  "(!! :X .. Pf) [[r]] =F
   !! x:X .. (Pf x [[r]])"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_Renaming_Rep_dist0)
apply (rule order_antisym)
apply (rule, simp add: in_failures, fast)+
done

(*********************************************************
          Rep_dist0 law for Sequential composition
 *********************************************************)

lemma cspF_Seq_compo_Rep_dist0: 
  "(!! :X .. Pf) ;; Q =F
   !! x:X .. (Pf x ;; Q)"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_Seq_compo_Rep_dist0)
apply (rule order_antisym)
apply (rule, simp add: in_traces in_failures)
apply (force)
apply (rule, simp add: in_traces in_failures)
apply (force)
done

(*********************************************************
                     Rep_dist0 laws
 *********************************************************)

lemmas cspF_Rep_dist0 = cspF_Ext_choice_Rep_dist0_l cspF_Ext_choice_Rep_dist0_r
                        cspF_Parallel_Rep_dist0_l   cspF_Parallel_Rep_dist0_r
                        cspF_Hiding_Rep_dist0       cspF_Renaming_Rep_dist0
                        cspF_Seq_compo_Rep_dist0

(*****************************************************************

      for convenience

         1. (!!<f> :X .. Pf) [+] Q
         2. Q [+] (!!<f> :X .. Pf)
         3. (!!<f> :X .. Pf) |[X]| Q
         4. Q |[X]| (!!<f> :X .. Pf)
         5. (!!<f> :X .. Pf) -- X
         6. (!!<f> :X .. Pf) [[r]]

 *****************************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspF_Ext_choice_Rep_dist_fun_l: 
  "[| inj f ; X ~= {} |]
   ==> (!!<f> :X .. Pf) [+] Q =F !!<f> x:X .. (Pf x [+] Q)"
by (simp add: Rep_int_choice_fun_def cspF_Rep_dist0)

lemma cspF_Ext_choice_Rep_dist_fun_r: 
  "[| inj f ; X ~= {} |]
   ==> P [+] (!!<f> :X .. Qf) =F !!<f> x:X .. (P [+] Qf x)"
by (simp add: Rep_int_choice_fun_def cspF_Rep_dist0)

lemma cspF_Parallel_Rep_dist_fun_l: 
  "[| inj f ; Y ~= {} |]
   ==> (!!<f> :Y .. Pf) |[X]| Q =F !!<f> x:Y .. (Pf x |[X]| Q)"
by (simp add: Rep_int_choice_fun_def cspF_Rep_dist0)

lemma cspF_Parallel_Rep_dist_fun_r: 
  "[| inj f ; Y ~= {} |]
   ==> P |[X]| (!!<f> :Y .. Qf) =F !!<f> x:Y .. (P |[X]| Qf x)"
by (simp add: Rep_int_choice_fun_def cspF_Rep_dist0)

lemma cspF_Hiding_Rep_dist_fun: 
  "inj f ==> (!!<f> :Y .. Pf) -- X =F !!<f> x:Y .. (Pf x -- X)"
by (simp add: Rep_int_choice_fun_def cspF_Rep_dist0)

lemma cspF_Renaming_Rep_dist_fun: 
  "inj f ==> (!!<f> :X .. Pf) [[r]] =F !!<f> x:X .. (Pf x [[r]])"
by (simp add: Rep_int_choice_fun_def cspF_Rep_dist0)

lemma cspF_Seq_compo_Rep_dist_fun:
  "inj f ==> (!!<f> :X .. Pf) ;; Q =F !!<f> x:X .. (Pf x ;; Q)"
by (simp add: Rep_int_choice_fun_def cspF_Rep_dist0)

(*********************************************************
                     Rep_dist laws
 *********************************************************)

lemmas cspF_Rep_dist_fun = cspF_Ext_choice_Rep_dist_fun_l cspF_Ext_choice_Rep_dist_fun_r
                           cspF_Parallel_Rep_dist_fun_l   cspF_Parallel_Rep_dist_fun_r
                           cspF_Hiding_Rep_dist_fun       cspF_Renaming_Rep_dist_fun
                           cspF_Seq_compo_Rep_dist_fun

(*** all rules ***)

lemmas cspF_Rep_dist = cspF_Rep_dist0 cspF_Rep_dist_fun

(*****************************************************************

      additional distribution of internal bind over ...

         1. (!! :X .. (a -> P))
         2. (!! :Y .. (? :X -> P))

 *****************************************************************)

(*********************************************************
              Rep_dist law for Act_prefix
 *********************************************************)

lemma cspF_Act_prefix_Rep_dist0:
  "X ~= {} ==> 
   a -> (!! :X .. Pf) =F !! x:X .. (a -> Pf x)"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_Act_prefix_Rep_dist0)
apply (rule order_antisym)

(* <= *)
 apply (rule)
 apply (simp add: in_failures)
 apply (erule disjE)
  apply (force)

  apply (elim conjE exE bexE)
  apply (rule_tac x="aa" in bexI)
  apply (force)
  apply (simp)

(* => *)
 apply (rule)
 apply (simp add: in_failures)
 apply (elim disjE conjE bexE exE)
  apply (simp)
  apply (simp)
  apply (rule_tac x="aa" in bexI)
  apply (simp)
  apply (simp)
done

(*********************************************************
              Rep_dist0 law for Ext_pre_choice
 *********************************************************)

lemma cspF_Ext_pre_choice_Rep_dist0:
  "Y ~= {} ==> 
   ? x:X -> (!! y:Y .. (Pf y) x) =F !! y:Y .. (? :X -> (Pf y))"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_Ext_pre_choice_Rep_dist0)
apply (rule order_antisym)

(* <= *)
 apply (rule)
 apply (simp add: in_failures)
 apply (erule disjE)
  apply (force)

  apply (elim conjE exE bexE)
  apply (rule_tac x="aa" in bexI)
  apply (force)
  apply (simp)

(* => *)
 apply (rule)
 apply (simp add: in_failures)
 apply (elim disjE conjE exE bexE)
  apply (simp)
  apply (simp)
  apply (rule_tac x="a" in bexI)
  apply (simp)
  apply (simp)
done

(*****************************************************************

      for convenience

         1. (!!<f> :X .. (a -> P))
         2. (!!<f> :Y .. (? :X -> P))

 *****************************************************************)

lemma cspF_Act_prefix_Rep_dist_fun:
  "X ~= {} ==> 
   a -> (!!<f> :X .. Pf) =F !!<f> x:X .. (a -> Pf x)"
by (simp add: Rep_int_choice_fun_def cspF_Act_prefix_Rep_dist0)

lemma cspF_Ext_pre_choice_Rep_dist_fun:
  "Y ~= {} ==> 
   ? x:X -> (!!<f> y:Y .. (Pf y) x) =F !!<f> y:Y .. (? :X -> (Pf y))"
by (simp add: Rep_int_choice_fun_def cspF_Ext_pre_choice_Rep_dist0)

(*** arias ***)

lemmas cspF_Act_prefix_Rep_dist 
     = cspF_Act_prefix_Rep_dist0
       cspF_Act_prefix_Rep_dist_fun

lemmas cspF_Ext_pre_choice_Rep_dist
     = cspF_Ext_pre_choice_Rep_dist0
       cspF_Ext_pre_choice_Rep_dist_fun

(****************** to add them again ******************)

end
