           (*-------------------------------------------*
            |        CSP-Prover on Isabelle2004         |
            |               February 2005               |
            |                   June 2005 (modified)    |
            |              September 2005 (modified)    |
            |                                           |
            |        CSP-Prover on Isabelle2005         |
            |                October 2005  (modified)   |
            |                                           |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_F_law_fp = CSP_T_law_fp
                    + CSP_F_continuous + CSP_F_contraction + CSP_F_mono:

(*****************************************************************

         1. cpo fixed point theory in CSP-Prover
         2. cms fixed point theory in CSP-Prover
         3.
         4. 

 *****************************************************************)

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  Union (B ` A) = (UN x:A. B x)                      *)
(*                  Inter (B ` A) = (INT x:A. B x)                     *)

declare Union_image_eq [simp del]
declare Inter_image_eq [simp del]

(*=======================================================*
 |                                                       |
 |                        CPO                            |
 |                                                       |
 *=======================================================*)

(*-----------*
 |    Bot    |
 *-----------*)

lemma failures_prod_Bot: "(%pn. failures DIV) = Bot"
apply (simp add: prod_Bot_def)
apply (simp add: expand_fun_eq)
apply (simp add: failures_def)
apply (simp add: bottom_setF_def)
done

lemma traces_failures_prod_Bot: "(%pn. (traces DIV ,, failures DIV)) = Bot"
apply (simp add: prod_Bot_def)
apply (simp add: expand_fun_eq)
apply (simp add: bottom_domF_def)
apply (simp add: traces_def failures_def)
done

lemma semF_prod_Bot: "(%pn. [[DIV]]F) = Bot"
apply (simp add: semF_def)
apply (simp add: traces_failures_prod_Bot)
done

(*-----------*
 |   FIX P   |
 *-----------*)

lemma semF_FIX_LUB_domF_T:
   "fst ` Rep_domF ` 
   {y. EX x. (EX n. x = ([[P]]FFun ^ n) Bot) & y = x pn}
 = {t. EX n. t = (fstF ((([[P]]FFun ^ n) Bot) pn))}"
by (auto simp add: fstF_def)

lemma semF_FIX_LUB_domF_F:
   "Union (Rep_setF ` snd ` Rep_domF ` 
           {y. EX x. (EX n. x = ([[P]]FFun ^ n) Bot) & y = x pn})
    = {f. EX n. f :f (sndF ((([[P]]FFun ^ n) Bot) pn))}"
by (auto simp add: Union_image_eq memF_def sndF_def)

lemma semF_FIX: 
    "PF : ProcFun ==>
     [[(FIX PF) pn]]F
     = LUB_domF {y. EX x. (EX n. x = ([[PF]]FFun ^ n) Bot) & y = x pn}"
apply (simp add: semF_def)
apply (simp add: pairF_def)
apply (simp add: LUB_domF_def)
apply (subgoal_tac "EX x n. x = ([[P]]FFun ^ n) Bot")
apply (subgoal_tac 
 "{y. EX x. (EX n. x = ([[PF]]FFun ^ n) Bot) & y = x pn} ~= {}")
apply (simp add: Abs_domF_inject LUB_TF_in_Rep)
apply (simp add: LUB_TF_def)
apply (simp add: semF_FIX_LUB_domF_T)
apply (simp add: UnionF_def semF_FIX_LUB_domF_F)
apply (rule) 

 (* T *)
 apply (simp add: traces_FIX)
 apply (simp add: semFFun_compo_n[THEN sym])
 apply (simp add: iteration_in_ProcFun pairF_semF)
 apply (simp add: iteration_in_ProcFun tracesFun_compo_n_X)

 (* F *)
 apply (simp add: FIX_def FIXn_def)
 apply (rule order_antisym)

  (* <= *)
  apply (rule)
  apply (simp add: in_failures)
  apply (elim conjE exE)
  apply (rule_tac x="a" in exI)
  apply (simp add: semFFun_compo_n[THEN sym])
  apply (simp add: iteration_in_ProcFun failuresFun_X_pn)
  apply (simp add: iteration_in_ProcFun pairF_semF)
  apply (simp add: traces_failures_prod_Bot)

  (* => *)
  apply (rule)
  apply (simp add: semFFun_compo_n[THEN sym])
  apply (simp add: in_failures)
  apply (elim conjE exE)
  apply (rule_tac x="n" in exI)
  apply (simp add: iteration_in_ProcFun failuresFun_X_pn)
  apply (simp add: iteration_in_ProcFun pairF_semF)
  apply (simp add: traces_failures_prod_Bot)

by (auto)

(*** FIX is LUB ***)

lemma semF_FIX_isLUB:
   "PF : ProcFun ==>
    (%pn. [[(FIX PF) pn]]F) isLUB {x. EX n. x = ([[PF]]FFun ^ n) Bot}"
apply (simp add: prod_LUB_decompo)
apply (intro allI)
apply (simp add: proj_fun_def)
apply (simp add: image_def)
apply (simp add: semF_FIX)
apply (subgoal_tac 
     "{y. EX x. (EX n. x = ([[PF]]FFun ^ n) Bot) & y = x i} =
      {u. EX n. u = ([[PF]]FFun ^ n) Bot i}")
apply (simp)
apply (rule LUB_domF_isLUB)
apply (auto)
done

lemma semF_FIX_LUB:
   "PF : ProcFun ==>
    (%pn. [[(FIX PF) pn]]F) = LUB {x. EX n. x = ([[PF]]FFun ^ n) Bot}"
apply (rule sym)
apply (rule isLUB_LUB)
apply (simp add: semF_FIX_isLUB)
done

lemma semF_FIX_LFP:
   "PF : ProcFun ==> (%pn. [[(FIX PF) pn]]F) = LFP [[PF]]FFun"
apply (simp add: semF_FIX_LUB)
apply (simp add: Tarski_thm_LFP_LUB continuous_semFFun)
done

lemma semF_FIX_LFP_pn:
   "PF : ProcFun ==> [[(FIX PF) pn]]F = LFP [[PF]]FFun pn"
apply (insert semF_FIX_LFP[of PF])
apply (simp add: expand_fun_eq)
done

lemma semF_FIX_isLFP:
   "PF : ProcFun ==> (%pn. [[(FIX PF) pn]]F) isLFP [[PF]]FFun"
apply (simp add: semF_FIX_LFP)
apply (simp add: LFP_is Tarski_thm_EX continuous_semFFun)
done

lemma semF_FIX_LFP_fixed_point:
   "PF : ProcFun
    ==> [[PF]]FFun (%pn. [[(FIX PF) pn]]F) = (%pn. [[(FIX PF) pn]]F)"
apply (insert semF_FIX_isLFP[of PF], simp)
apply (simp add: isLFP_def)
done

lemma semF_FIX_LFP_least:
   "PF : ProcFun ==> 
    ALL X. [[PF]]FFun X = X --> (%pn. [[(FIX PF) pn]]F) <= X"
apply (insert semF_FIX_isLFP[of PF], simp)
apply (simp add: isLFP_def)
done

(*---------------------------------------------------------------------*
 |   FIX is the greatest fixed point of process-expression-functions   |
 |   (semantically least fixed point)                                  |
 *---------------------------------------------------------------------*)

lemma cspF_FIX_isGFP:
  "PF : ProcFun ==> 
   (FIX PF) ==F (PF (FIX PF)) & (ALL X. X ==F PF X --> X <==F (FIX PF))"
apply (simp add: refF_prod_def eqF_prod_def)
apply (simp add: refF_def eqF_def)
apply (insert semF_FIX_isLFP[of PF], simp)
apply (rule)

 (* FP *)
 apply (simp add: expand_fun_eq[THEN sym])
 apply (simp add: semFFun_X_pn)
 apply (simp add: isLFP_def)

 (* LFP *)
 apply (intro allI impI)
 apply (simp add: expand_fun_eq[THEN sym])
 apply (drule sym)
 apply (simp add: expand_fun_eq)
 apply (simp add: expand_fun_eq[THEN sym])
 apply (simp add: semFFun_X_pn)
 apply (simp add: isLFP_def)
 apply (erule conjE)
 apply (drule_tac x="(%n. [[X n]]F)" in spec)
 apply (simp add: order_prod_def)
done

(*-------------------------------------------------------*
 |                                                       |
 |           Fixpoint unwind (CSP-Prover rule)           |
 |                                                       |
 *-------------------------------------------------------*)

lemma cspF_unwind_lm:
  "PF : ProcFun ==> ALL pn. (FIX PF) pn =F PF (FIX PF) pn"
apply (simp add: eqF_def)
apply (simp add: expand_fun_eq[THEN sym])
apply (simp add: semFFun_X_pn)
apply (simp add: semF_FIX_LFP_fixed_point)
done

(*** csp rule ***)

lemma cspF_unwind:
  "PF : ProcFun ==> (FIX PF) pn =F PF (FIX PF) pn"
by (simp add: cspF_unwind_lm)

(*-------------------------------------------------------*
 |                                                       |
 |        Check fixpoint  (CSP-Prover intro rule)        |
 |                                                       |
 *-------------------------------------------------------*)

lemma cspF_fp_induct_cpo_lm:
   "[| PF : ProcFun ; Q <=F X pn ; ALL pn. X pn <=F (PF X) pn |]
    ==> Q <=F (FIX PF) pn"
apply (simp add: refF_def)
apply (simp add: fold_order_prod_def)
apply (simp add: semFFun_X_pn)
apply (simp add: semF_FIX_LFP_pn)
apply (insert cpo_fixpoint_induction_rev
       [of "[[PF]]FFun" "(%pn. [[X pn]]F)"])
apply (simp add: continuous_semFFun)
apply (simp add: order_prod_def)
apply (rotate_tac -1)
apply (drule_tac x="pn" in spec)
apply (auto)
done

(*** csp rule ***)

lemma cspF_fp_induct_cpo:
   "[| PF : ProcFun ; Q <=F X pn ; !!pn. X pn <=F (PF X) pn |]
    ==> Q <=F (FIX PF) pn"
by (simp add: cspF_fp_induct_cpo_lm)

(*=======================================================*
 |                                                       |
 |                        CMS                            |
 |                                                       |
 *=======================================================*)

lemma semF_FIX_UFP:
   "[| PF : ProcFun ; PF : nohideFun ; PF : guardFun |] ==>
    (%pn. [[(FIX PF) pn]]F) = UFP [[PF]]FFun"
apply (subgoal_tac "[[PF]]FFun hasUFP")
apply (insert semF_FIX_isLFP[of PF], simp)
apply (subgoal_tac "UFP [[PF]]FFun isUFP [[PF]]FFun")
apply (simp add: isUFP_def)
apply (simp add: isLFP_def)
apply (elim conjE exE)
apply (rotate_tac -1)
apply (drule_tac x="(%pn. [[(FIX PF) pn]]F)" in spec)
apply (simp)
apply (simp add: UFP_is)
apply (simp add: contraction_semFFun Banach_thm_EX)
done

lemma semF_FIX_UFP_pn:
   "[| PF : ProcFun ; PF : nohideFun ; PF : guardFun |] ==>
    [[(FIX PF) pn]]F = UFP [[PF]]FFun pn"
apply (insert semF_FIX_UFP[of PF])
apply (simp add: expand_fun_eq)
done

lemma semF_FIX_isUFP:
   "[| PF : ProcFun ; PF : nohideFun ; PF : guardFun |] ==>
    (%pn. [[(FIX PF) pn]]F) isUFP [[PF]]FFun"
apply (simp add: semF_FIX_UFP)
apply (simp add: contraction_semFFun Banach_thm_EX UFP_is)
done

lemma semF_FIX_UFP_unique:
   "[| PF : ProcFun ; PF : nohideFun ; PF : guardFun |] ==>
    ALL X. [[PF]]FFun X = X --> (%pn. [[(FIX PF) pn]]F) = X"
apply (insert semF_FIX_isUFP[of PF], simp)
apply (simp add: isUFP_def)
done

(*------------------------------------------------------------------*
 |   FIX is the unique fixed point of process-expression-functions  |
 *------------------------------------------------------------------*)

lemma cspF_FIX_isUFP:
   "[| PF : ProcFun ; PF : nohideFun ; PF : guardFun |] ==>
   (FIX PF) ==F (PF (FIX PF)) & (ALL X. X ==F PF X --> X ==F (FIX PF))"
apply (simp add: cspF_FIX_isGFP)

apply (simp add: eqF_prod_def)
apply (simp add: eqF_def)
apply (insert semF_FIX_isUFP[of PF], simp)

 (* UFP *)
 apply (intro allI impI)
 apply (simp add: expand_fun_eq[THEN sym])
 apply (drule sym)
 apply (simp add: expand_fun_eq)
 apply (simp add: expand_fun_eq[THEN sym])
 apply (simp add: semFFun_X_pn)
 apply (simp add: isUFP_def)
 apply (erule conjE)
 apply (drule_tac x="(%pn. [[X pn]]F)" in spec)
 apply (simp add: expand_fun_eq)
done

(*-------------------------------------------------------*
 |                                                       |
 |        Check fixpoint  (CSP-Prover intro rule)        |
 |                                                       |
 *-------------------------------------------------------*)

(* refinement *)

lemma cspF_fp_induct_cms_ref_lm:
   "[| PF : ProcFun ; PF : nohideFun ; PF : guardFun  ;
       X pn <=F Q ; ALL pn. (PF X) pn <=F X pn |]
    ==> (FIX PF) pn <=F Q"
apply (simp add: refF_def)
apply (simp add: fold_order_prod_def)
apply (simp add: semFFun_X_pn)

apply (simp add: semF_FIX_UFP_pn)
apply (insert cms_fixpoint_induction_ref
       [of "[[PF]]FFun" "(%pn. [[X pn]]F)" "UFP [[PF]]FFun"])
apply (simp add: contra_alpha_to_contst contraction_alpha_semFFun)
apply (simp add: mono_semFFun)
apply (simp add: contraction_semFFun UFP_fp Banach_thm_EX)
apply (simp add: order_prod_def)
apply (rotate_tac -1)
apply (drule_tac x="pn" in spec)
apply (simp)
done

(*** csp rule ***)

lemma cspF_fp_induct_cms_ref:
   "[| PF : ProcFun ; PF : nohideFun ; PF : guardFun  ;
       X pn <=F Q ; !! pn. (PF X) pn <=F X pn |]
    ==> (FIX PF) pn <=F Q"
by (simp add: cspF_fp_induct_cms_ref_lm)

(* equivalence *)

lemma cspF_fp_induct_cms_eq_lm:
   "[| PF : ProcFun ; PF : nohideFun ; PF : guardFun  ;
       X pn =F Q ; ALL pn. (PF X) pn =F X pn |]
    ==> (FIX PF) pn =F Q"
apply (simp add: eqF_def)
apply (simp add: expand_fun_eq[THEN sym])
apply (simp add: semFFun_X_pn)
apply (insert semF_FIX_isUFP[of PF], simp)
apply (simp add: isUFP_def)
apply (elim conjE)
apply (drule_tac x="(%pn. [[X pn]]F)" in spec)
apply (simp add: expand_fun_eq)
done

(*** csp rule ***)

lemma cspF_fp_induct_cms_eq:
   "[| PF : ProcFun ; PF : nohideFun ; PF : guardFun ;
       X pn =F Q ; !! pn. (PF X) pn =F X pn |]
    ==> (FIX PF) pn =F Q"
by (simp add: cspF_fp_induct_cms_eq_lm)

lemmas cspF_fp_induct_cms = cspF_fp_induct_cms_ref cspF_fp_induct_cms_eq

(*-------------------------------------------------------*
 |                                                       |
 |         FIX expansion (CSP-Prover intro rule)         |
 |                                                       |
 *-------------------------------------------------------*)

lemma failures_FIXn_plus_sub_lm:
 "PF : ProcFun ==>
  ALL n m pn. failures ((FIX[n] PF) pn) <= failures ((FIX[n+m] PF) pn)"
apply (rule)
apply (induct_tac n)
apply (simp add: FIXn_def)
apply (intro allI)
apply (rule)
apply (simp add: in_failures)

apply (rule allI)
apply (simp add: FIXn_def)
apply (drule_tac x="m" in spec)
apply (simp add: failuresFun_X_pn)
apply (fold order_prod_def)
apply (subgoal_tac
  "(%pn. (traces   ((PF ^ na) (%pm. DIV) pn) ,, 
          failures ((PF ^ na) (%pm. DIV) pn))) <=
   (%pn. (traces   ((PF ^ (na + m)) (%pm. DIV) pn) ,,
          failures ((PF ^ (na + m)) (%pm. DIV) pn)))")
apply (simp add: mono_failuresFun[simplified mono_def])

apply (simp add: order_prod_def)
apply (intro allI)
apply (simp add: subdomF_decompo)
apply (insert traces_FIXn_plus_sub_lm[of PF])
apply (simp add: FIXn_def)
done

lemma failures_FIXn_plus_sub: 
   "PF : ProcFun ==> failures ((FIX[n] PF) pn) <= failures ((FIX[n+m] PF) pn)"
by (simp add: failures_FIXn_plus_sub_lm)

lemma semF_FIXn_plus_sub:
 "PF : ProcFun ==> [[(FIX[n] PF) pn]]F <= [[(FIX[n+m] PF) pn]]F"
apply (simp add: semF_decompo)
apply (simp add: subdomF_decompo)
apply (simp add: traces_FIXn_plus_sub)
apply (simp add: failures_FIXn_plus_sub)
done

lemma in_failures_FIXn_plus_sub: 
  "[| PF : ProcFun ; (s,X) :f failures ((FIX[n] PF) pn) |]
   ==> (s,X) :f failures ((FIX[n+m] PF) pn)"
apply (insert failures_FIXn_plus_sub[of PF n pn m])
by (auto)

(*-----------------------------------------------------*
 |  sometimes FIX[n + f n] is useful more than FIX[n]  |
 *-----------------------------------------------------*)

lemma cspF_FIX_plus_eq: 
     "PF : ProcFun ==>
      ALL f pn. (FIX PF) pn =F (!nat n .. ((FIX[n + f n] PF) pn))"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_FIX_plus_eq)

apply (simp add: FIX_def)
apply (intro allI)
apply (rule order_antisym)

 apply (rule)
 apply (simp add: in_failures)
 apply (elim conjE exE)
 apply (rule_tac x="a" in exI)
 apply (simp add: in_failures_FIXn_plus_sub)

 apply (rule)
 apply (simp add: in_failures)
 apply (elim conjE exE)
 apply (rule_tac x="a + f a" in exI)
 apply (simp)
done

lemma cspF_FIX_plus_eq_fun: 
     "PF : ProcFun ==>
      ALL f pn. (%pn. (FIX PF) pn) ==F (%pn. (!nat n .. ((FIX[n + f n] PF) pn)))"
apply (simp add: eqF_prod_def)
apply (simp add: cspF_FIX_plus_eq)
done

(*=======================================================*
 |                                                       |
 |                Single recursion MU                    |
 |                                                       |
 *=======================================================*)

(*=======================================================*
 |                         cpo                           |
 *=======================================================*)

(*-------------------------------------------------------*
 |            unwinding (CSP-Prover intro rule)          |
 *-------------------------------------------------------*)

lemma cspF_unwind_MU:
  "PX : ProcX ==> MUX PX =F PX (MUX PX)"
apply (simp add: MUX_def)
apply (insert cspF_unwind[of "(%x n. PX (x MUpn))" "MUpn"])
apply (simp add: ProcX_def)
apply (simp add: ProcFun_def)
done

(*-------------------------------------------------------*
 |        Check fixpoint  (CSP-Prover intro rule)        |
 *-------------------------------------------------------*)

lemma cspF_fp_induct_MU_cpo:
   "[| PX : ProcX ; Q <=F PX Q |] ==> Q <=F (MUX PX)"
apply (simp add: MUX_def)
apply (rule cspF_fp_induct_cpo[of _ _ "(%X. Q)"])
apply (simp add: ProcX_def)
apply (simp add: ProcFun_def)
apply (simp)
apply (simp)
done

(*=======================================================*
 |                         cms                           |
 *=======================================================*)

(*-------------------------------------------------------*
 |        Check fixpoint  (CSP-Prover intro rule)        |
 *-------------------------------------------------------*)

lemma cspF_fp_induct_MU_cms_ref:
   "[| PX : ProcX ; PX : nohideX ; PX : guardX ; PX Q <=F Q |]
    ==> (MUX PX) <=F Q"
apply (simp add: MUX_def)
apply (rule cspF_fp_induct_cms_ref[of _ "(%X. Q)"])
apply (simp add: ProcX_def)
apply (simp add: ProcFun_def)
apply (simp add: nohideX_def)
apply (simp add: nohideFun_def)
apply (simp add: guardX_def)
apply (simp add: guardFun_def)
apply (simp)
apply (simp)
done

lemma cspF_fp_induct_MU_cms_eq:
   "[| PX : ProcX ; PX : nohideX ; PX : guardX ; PX Q =F Q |]
    ==> (MUX PX) =F Q"
apply (simp add: MUX_def)
apply (rule cspF_fp_induct_cms_eq[of _ "(%X. Q)"])
apply (simp add: ProcX_def)
apply (simp add: ProcFun_def)
apply (simp add: nohideX_def)
apply (simp add: nohideFun_def)
apply (simp add: guardX_def)
apply (simp add: guardFun_def)
apply (simp)
apply (simp)
done

lemmas cspF_fp_induct_MU_cms = cspF_fp_induct_MU_cms_ref cspF_fp_induct_MU_cms_eq

(*=======================================================*
 |                   expand by Tariski                   |
 *=======================================================*)

lemma cspF_FIX_MU_plus_eq_lm: 
     "((%x pn. (PX::'a proc => 'a proc) (x MUpn)) ^ n) (%pm. DIV) MUpn = (PX ^ n) DIV"
apply (induct_tac n)
apply (simp_all)
done

lemma cspF_FIX_MU_plus_eq: 
     "PX : ProcX ==>
      ALL f. (MUX PX) =F (!nat n .. (PX ^ (n + f n)) DIV)"
apply (insert cspF_FIX_plus_eq[of "(%x pn. PX (x MUpn))"])
apply (simp add: ProcX_def)
apply (simp add: ProcFun_def)
apply (rule allI)
apply (drule_tac x="f" in spec)
apply (simp add: MUX_def)
apply (drule_tac x="MUpn" in spec)
apply (simp add: FIXn_def)
apply (simp add: cspF_FIX_MU_plus_eq_lm)
done

lemma cspF_FIX_MU_eq: 
     "PX : ProcX ==>
      (MUX PX) =F (!nat n .. (PX ^ n) DIV)"
apply (insert cspF_FIX_MU_plus_eq[of PX])
apply (simp)
apply (drule_tac x="(%n. (0::nat))" in spec)
apply (simp)
done

(****************** to add them again ******************)

declare Union_image_eq [simp]
declare Inter_image_eq [simp]

end
