           (*-------------------------------------------*
            |        CSP-Prover on Isabelle2004         |
            |               December 2004               |
            |               February 2005  (modified)   |
            |                   June 2005  (modified)   |
            |                                           |
            |        CSP-Prover on Isabelle2005         |
            |                October 2005  (modified)   |
            |               November 2005  (modified)   |
            |                                           |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_F_tactic = CSP_F_law:

(*****************************************************************

         1. tactic
         2.
         3. 
         4. 

 *****************************************************************)

(*==============================================================*
 |                                                              |
 |    decompostion control by the flag "Not_Decompo_Flag"       |
 |                                                              |
 *==============================================================*)

(*------------------------------------------------*
 |              trans with Flag                   |
 *------------------------------------------------*)

(*** rewrite (eq) ***)

lemma cspF_rw_flag_left_eq:
  "[| R1 =F R2 ; Not_Decompo_Flag & R2 =F R3 |] ==> R1 =F R3"
by (simp add: eqF_def Not_Decompo_Flag_def)

lemma cspF_rw_flag_left_ref:
  "[| R1 =F R2 ; Not_Decompo_Flag & R2 <=F R3 |] ==> R1 <=F R3"
by (simp add: refF_def eqF_def Not_Decompo_Flag_def)

lemmas cspF_rw_flag_left = cspF_rw_flag_left_eq cspF_rw_flag_left_ref

lemma cspF_rw_flag_right_eq:
  "[| R3 =F R2 ; Not_Decompo_Flag & R1 =F R2 |] ==> R1 =F R3"
by (simp add: eqF_def Not_Decompo_Flag_def)

lemma cspF_rw_flag_right_ref:
  "[| R3 =F R2 ; Not_Decompo_Flag & R1 <=F R2 |] ==> R1 <=F R3"
by (simp add: refF_def eqF_def Not_Decompo_Flag_def)

lemmas cspF_rw_flag_right = cspF_rw_flag_right_eq cspF_rw_flag_right_ref

(*==============================================================*
 |  decompostion of Sequential composition with a flag          |
 |  It is often useful that the second process is not expanded. |
 *==============================================================*)

lemma cspF_Seq_compo_mono_flag:
  "[| P1 <=F Q1 ; 
      Not_Decompo_Flag & P2 <=F Q2 |]
           ==> P1 ;; P2 <=F Q1 ;; Q2"
by (simp add: cspF_Seq_compo_mono)

lemma cspF_Seq_compo_cong_flag:
  "[| P1 =F Q1 ; 
      Not_Decompo_Flag & P2 =F Q2 |]
           ==> P1 ;; P2 =F Q1 ;; Q2"
by (simp add: cspF_Seq_compo_cong)

lemmas cspF_free_mono_flag =
       cspF_Ext_choice_mono cspF_Int_choice_mono cspF_Parallel_mono
       cspF_Hiding_mono cspF_Renaming_mono cspF_Seq_compo_mono_flag

lemmas cspF_free_cong_flag =
       cspF_Ext_choice_cong cspF_Int_choice_cong cspF_Parallel_cong
       cspF_Hiding_cong cspF_Renaming_cong cspF_Seq_compo_cong_flag

lemmas cspF_free_decompo_flag = cspF_free_mono_flag cspF_free_cong_flag

(*================================================*
 |                                                |
 |                  Tacticals                     |
 |                                                |
 *================================================*)

lemmas cspF_dist_both = cspF_dist cspF_Rep_dist
lemmas cspF_choice_Conditional = cspF_choice_rule cspF_Conditional

ML_setup {*
    val CSPF_reflex                    = thms "cspF_reflex" ;
    val CSPF_rw_flag_left              = thms "cspF_rw_flag_left" ;
    val CSPF_rw_flag_right             = thms "cspF_rw_flag_right" ;

    val CSPF_choice_Conditional        = thms "cspF_choice_Conditional";
    val CSPF_step                      = thms "cspF_step" ;
    val CSPF_light_step                = thms "cspF_light_step" ;
    val CSPF_SKIP                      = thms "cspF_SKIP" ;
    val CSPF_dist_both                 = thms "cspF_dist_both" ;
    val CSPF_unwind                    = thms "cspF_unwind" ;
    val CSPF_unwind_MU                 = thms "cspF_unwind_MU" ;
    val CSPF_free_decompo_flag         = thms "cspF_free_decompo_flag" ;
*}

(************************************************
                  rewrite left
 ************************************************)

(*-----------------------------------------------------*
      apply (tactic {* cspF_hnf_left_tac 1 *})

                      is equal to

      apply (rule cspF_rw_flag_left ,
            ( simp
             | ((rule off_Not_Decompo_Flag)?,
                (  rule cspF_rule_Conditional
                 | rule cspF_SKIP
                 | rule cspF_step
                 | rule cspF_dist_both
                 | rule cspF_unwind
                 | rule cspF_unwind_MU))
             | (rule cspF_rw_flag_left, rule cspF_free_decompo_flag)
             | ((rule off_Not_Decompo_Flag)?, rule cspF_reflex)
             | (rule off_Not_Decompo_Flag_True))+ ,
             (simp_all off_Not_Decompo_Flag_True))

 *-----------------------------------------------------*)

(*** Head Normal Form ***)

ML_setup {*
fun cspF_hnf_left_tac n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_left n,

    REPEAT (
     FIRST 
      [(CHANGED (asm_full_simp_tac (simpset ()) n)),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                FIRST [ resolve_tac CSPF_choice_Conditional n,
                        resolve_tac CSPF_SKIP n,
                        resolve_tac CSPF_step n,
                        resolve_tac CSPF_dist_both n,
                        resolve_tac CSPF_unwind n,
                        resolve_tac CSPF_unwind_MU n ]]),

       (EVERY [ resolve_tac CSPF_rw_flag_left n , 
                resolve_tac CSPF_free_decompo_flag n ]),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                resolve_tac CSPF_reflex n ]),

       (resolve_tac OFF_Not_Decompo_Flag_True n)]) ,

    (ALLGOALS (asm_full_simp_tac (simpset () addsimps OFF_Not_Decompo_Flag_True)))])
*}

(*** step ***)

ML_setup {*
fun cspF_step_left_tac n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_left n,

    REPEAT (
     FIRST 
      [(CHANGED (asm_full_simp_tac (simpset ()) n)),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                FIRST [ resolve_tac CSPF_choice_Conditional n,
                        resolve_tac CSPF_SKIP n,
                        resolve_tac CSPF_step n ]]),

       (EVERY [ resolve_tac CSPF_rw_flag_left n , 
                resolve_tac CSPF_free_decompo_flag n ]),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                resolve_tac CSPF_reflex n ]),

       (resolve_tac OFF_Not_Decompo_Flag_True n)]) ,

    (ALLGOALS (asm_full_simp_tac (simpset () addsimps OFF_Not_Decompo_Flag_True)))])
*}

(*** light step ***)

ML_setup {*
fun cspF_light_step_left_tac n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_left n,

    REPEAT (
     FIRST 
      [(CHANGED (asm_full_simp_tac (simpset ()) n)),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                FIRST [ resolve_tac CSPF_choice_Conditional n,
                        resolve_tac CSPF_SKIP n,
                        resolve_tac CSPF_light_step n ]]),

       (EVERY [ resolve_tac CSPF_rw_flag_left n , 
                resolve_tac CSPF_free_decompo_flag n ]),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                resolve_tac CSPF_reflex n ]),

       (resolve_tac OFF_Not_Decompo_Flag_True n)]) ,

    (ALLGOALS (asm_full_simp_tac (simpset () addsimps OFF_Not_Decompo_Flag_True)))])
*}

(*** dist ***)

ML_setup {*
fun cspF_dist_left_tac n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_left n,

    REPEAT (
     FIRST 
      [(CHANGED (asm_full_simp_tac (simpset ()) n)),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                FIRST [ resolve_tac CSPF_choice_Conditional n,
                        resolve_tac CSPF_dist_both n ]]),

       (EVERY [ resolve_tac CSPF_rw_flag_left n , 
                resolve_tac CSPF_free_decompo_flag n ]),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                resolve_tac CSPF_reflex n ]),

       (resolve_tac OFF_Not_Decompo_Flag_True n)]) ,

    (ALLGOALS (asm_full_simp_tac (simpset () addsimps OFF_Not_Decompo_Flag_True)))])
*}

(*** unwind ***)

ML_setup {*
fun cspF_unwind_left_tac n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_left n,

    REPEAT (
     FIRST 
      [(CHANGED (asm_full_simp_tac (simpset ()) n)),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                FIRST [ resolve_tac CSPF_choice_Conditional n,
                        resolve_tac CSPF_unwind n,
                        resolve_tac CSPF_unwind_MU n ]]),

       (EVERY [ resolve_tac CSPF_rw_flag_left n , 
                resolve_tac CSPF_free_decompo_flag n ]),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                resolve_tac CSPF_reflex n ]),

       (resolve_tac OFF_Not_Decompo_Flag_True n)]) ,

    (ALLGOALS (asm_full_simp_tac (simpset () addsimps OFF_Not_Decompo_Flag_True)))])
*}

(*** simp ***)

ML_setup {*
fun cspF_simp_left_tac n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_left n,

    REPEAT (
     FIRST 
      [(CHANGED (asm_full_simp_tac (simpset ()) n)),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                resolve_tac CSPF_choice_Conditional n]),

       (EVERY [ resolve_tac CSPF_rw_flag_left n , 
                resolve_tac CSPF_free_decompo_flag n ]),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                resolve_tac CSPF_reflex n ]),

       (resolve_tac OFF_Not_Decompo_Flag_True n)]) ,

    (ALLGOALS (asm_full_simp_tac (simpset () addsimps OFF_Not_Decompo_Flag_True)))])
*}

(*** rule ***)

(*-----------------------------------------------------*
      apply (tactic {* cspF_simp_with_left_tac "rulename" 1 *})

                      is equal to

      apply (rule cspF_rw_flag_left ,
            ( simp
             | ((rule off_Not_Decompo_Flag)?,
                (  rule cspF_rule_Conditional
                 | rule rulename))
             | (rule cspF_rw_flag_left, rule cspF_free_decompo_flag)
             | ((rule off_Not_Decompo_Flag)?, rule cspF_reflex),
             | (rule off_Not_Decompo_Flag_True))+ ,
             (simp add: rule off_Not_Decompo_Flag_True)?)

 *-----------------------------------------------------*)

ML_setup {*
fun cspF_simp_with_left_tac a n =
let val A = thms a
in
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_left n,

    REPEAT (
     FIRST 
      [(CHANGED (asm_full_simp_tac (simpset ()) n)),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                FIRST [ resolve_tac CSPF_choice_Conditional n,
                        resolve_tac A n ]]),

       (EVERY [ resolve_tac CSPF_rw_flag_left n , 
                resolve_tac CSPF_free_decompo_flag n ]),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                resolve_tac CSPF_reflex n ]),

       (resolve_tac OFF_Not_Decompo_Flag_True n)]) ,

    (ALLGOALS (asm_full_simp_tac (simpset () addsimps OFF_Not_Decompo_Flag_True)))])
end
*}

(*** assumption ***)

ML_setup {*
fun cspF_asm_left_tac n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_left n,

    REPEAT (
     FIRST 
      [(EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                (CHANGED (assume_tac n)) ]) ,

       (EVERY [ resolve_tac CSPF_rw_flag_left n , 
                resolve_tac CSPF_free_decompo_flag n ]),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                resolve_tac CSPF_reflex n ]),

       (resolve_tac OFF_Not_Decompo_Flag_True n)]) ,

    TRY (FIRST[ resolve_tac OFF_Not_Decompo_Flag n,
                resolve_tac OFF_Not_Decompo_Flag_True n])])
*}

(************************************************
                  rewrite right
 ************************************************)

(*-----------------------------------------------------*
      apply (tactic {* cspF_hnf_right_tac 1 *})

                      is equal to

      apply (rule cspF_rw_flag_right ,
            ( simp
             | ((rule off_Not_Decompo_Flag)?,
                (  rule cspF_rule_Conditional
                 | rule cspF_SKIP
                 | rule cspF_step
                 | rule cspF_dist_both
                 | rule cspF_unwind
                 | rule cspF_unwind_MU))
             | (rule cspF_rw_flag_left, rule cspF_free_decompo_flag)
             | ((rule off_Not_Decompo_Flag)?, rule cspF_reflex)
             | (rule off_Not_Decompo_Flag_True))+ ,
             (simp_all off_Not_Decompo_Flag_True))

 *-----------------------------------------------------*)

(*** Head Normal Form ***)

ML_setup {*
fun cspF_hnf_right_tac n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_right n,

    REPEAT (
     FIRST 
      [(CHANGED (asm_full_simp_tac (simpset ()) n)),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                FIRST [ resolve_tac CSPF_choice_Conditional n,
                        resolve_tac CSPF_SKIP n,
                        resolve_tac CSPF_step n,
                        resolve_tac CSPF_dist_both n,
                        resolve_tac CSPF_unwind n,
                        resolve_tac CSPF_unwind_MU n ]]),

       (EVERY [ resolve_tac CSPF_rw_flag_left n , 
                resolve_tac CSPF_free_decompo_flag n ]),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                resolve_tac CSPF_reflex n ]),

       (resolve_tac OFF_Not_Decompo_Flag_True n)]) ,

    (ALLGOALS (asm_full_simp_tac (simpset () addsimps OFF_Not_Decompo_Flag_True)))])
*}

(*** step ***)

ML_setup {*
fun cspF_step_right_tac n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_right n,

    REPEAT (
     FIRST 
      [(CHANGED (asm_full_simp_tac (simpset ()) n)),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                FIRST [ resolve_tac CSPF_choice_Conditional n,
                        resolve_tac CSPF_SKIP n,
                        resolve_tac CSPF_step n ]]),

       (EVERY [ resolve_tac CSPF_rw_flag_left n , 
                resolve_tac CSPF_free_decompo_flag n ]),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                resolve_tac CSPF_reflex n ]),

       (resolve_tac OFF_Not_Decompo_Flag_True n)]) ,

    (ALLGOALS (asm_full_simp_tac (simpset () addsimps OFF_Not_Decompo_Flag_True)))])
*}

(*** light step ***)

ML_setup {*
fun cspF_light_step_right_tac n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_right n,

    REPEAT (
     FIRST 
      [(CHANGED (asm_full_simp_tac (simpset ()) n)),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                FIRST [ resolve_tac CSPF_choice_Conditional n,
                        resolve_tac CSPF_SKIP n,
                        resolve_tac CSPF_light_step n ]]),

       (EVERY [ resolve_tac CSPF_rw_flag_left n , 
                resolve_tac CSPF_free_decompo_flag n ]),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                resolve_tac CSPF_reflex n ]),

       (resolve_tac OFF_Not_Decompo_Flag_True n)]) ,

    (ALLGOALS (asm_full_simp_tac (simpset () addsimps OFF_Not_Decompo_Flag_True)))])
*}

(*** dist ***)

ML_setup {*
fun cspF_dist_right_tac n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_right n,

    REPEAT (
     FIRST 
      [(CHANGED (asm_full_simp_tac (simpset ()) n)),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                FIRST [ resolve_tac CSPF_choice_Conditional n,
                        resolve_tac CSPF_dist_both n ]]),

       (EVERY [ resolve_tac CSPF_rw_flag_left n , 
                resolve_tac CSPF_free_decompo_flag n ]),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                resolve_tac CSPF_reflex n ]),

       (resolve_tac OFF_Not_Decompo_Flag_True n)]) ,

    (ALLGOALS (asm_full_simp_tac (simpset () addsimps OFF_Not_Decompo_Flag_True)))])
*}

(*** unwind ***)

ML_setup {*
fun cspF_unwind_right_tac n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_right n,

    REPEAT (
     FIRST 
      [(CHANGED (asm_full_simp_tac (simpset ()) n)),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                FIRST [ resolve_tac CSPF_choice_Conditional n,
                        resolve_tac CSPF_unwind n,
                        resolve_tac CSPF_unwind_MU n ]]),

       (EVERY [ resolve_tac CSPF_rw_flag_left n , 
                resolve_tac CSPF_free_decompo_flag n ]),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                resolve_tac CSPF_reflex n ]),

       (resolve_tac OFF_Not_Decompo_Flag_True n)]) ,

    (ALLGOALS (asm_full_simp_tac (simpset () addsimps OFF_Not_Decompo_Flag_True)))])
*}

(*** simp ***)

ML_setup {*
fun cspF_simp_right_tac n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_right n,

    REPEAT (
     FIRST 
      [(CHANGED (asm_full_simp_tac (simpset ()) n)),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                resolve_tac CSPF_choice_Conditional n]),

       (EVERY [ resolve_tac CSPF_rw_flag_left n , 
                resolve_tac CSPF_free_decompo_flag n ]),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                resolve_tac CSPF_reflex n ]),

       (resolve_tac OFF_Not_Decompo_Flag_True n)]) ,

    (ALLGOALS (asm_full_simp_tac (simpset () addsimps OFF_Not_Decompo_Flag_True)))])
*}

(*** rule ***)

(*-----------------------------------------------------*
      apply (tactic {* cspF_simp_with_right_tac "rulename" 1 *})

                      is equal to

      apply (rule cspF_rw_flag_right ,
            ( simp
             | ((rule off_Not_Decompo_Flag)?,
                (  rule cspF_rule_Conditional
                 | rule rulename))
             | (rule cspF_rw_flag_left, rule cspF_free_decompo_flag)
             | ((rule off_Not_Decompo_Flag)?, rule cspF_reflex),
             | (rule off_Not_Decompo_Flag_True))+ ,
             (simp add: rule off_Not_Decompo_Flag_True)?)

 *-----------------------------------------------------*)

ML_setup {*
fun cspF_simp_with_right_tac a n =
let val A = thms a
in
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_right n,

    REPEAT (
     FIRST 
      [(CHANGED (asm_full_simp_tac (simpset ()) n)),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                FIRST [ resolve_tac CSPF_choice_Conditional n,
                        resolve_tac A n ]]),

       (EVERY [ resolve_tac CSPF_rw_flag_left n , 
                resolve_tac CSPF_free_decompo_flag n ]),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                resolve_tac CSPF_reflex n ]),

       (resolve_tac OFF_Not_Decompo_Flag_True n)]) ,

    (ALLGOALS (asm_full_simp_tac (simpset () addsimps OFF_Not_Decompo_Flag_True)))])
end
*}

(*** assumption ***)

ML_setup {*
fun cspF_asm_right_tac n =
 CHANGED (
  EVERY
   [resolve_tac CSPF_rw_flag_right n,

    REPEAT (
     FIRST 
      [(EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                (CHANGED (assume_tac n)) ]) ,

       (EVERY [ resolve_tac CSPF_rw_flag_left n , 
                resolve_tac CSPF_free_decompo_flag n ]),

       (EVERY [ TRY (resolve_tac OFF_Not_Decompo_Flag n), 
                resolve_tac CSPF_reflex n ]),

       (resolve_tac OFF_Not_Decompo_Flag_True n)]) ,

    TRY (FIRST[ resolve_tac OFF_Not_Decompo_Flag n,
                resolve_tac OFF_Not_Decompo_Flag_True n])])
*}

(************************************************
                 rewrite both
 ************************************************)

(*-----------------------------------------------------*
      Apply (tactic {* cspF_hnf_tac n *})
                      is equal to
      apply ((tactic {* cspF_hnf_left_tac n *})? ,
             (tactic {* cspF_hnf_right_tac n *})?)
 *-----------------------------------------------------*)

ML_setup {*
    fun cspF_hnf_tac n = 
      CHANGED (
       EVERY
        [TRY (cspF_hnf_left_tac n),
         TRY (cspF_hnf_right_tac n)])
*}

ML_setup {*
    fun cspF_step_tac n = 
      CHANGED (
       EVERY
        [TRY (cspF_step_left_tac n),
         TRY (cspF_step_right_tac n)])
*}

ML_setup {*
    fun cspF_light_step_tac n = 
      CHANGED (
       EVERY
        [TRY (cspF_light_step_left_tac n),
         TRY (cspF_light_step_right_tac n)])
*}

ML_setup {*
    fun cspF_dist_tac n = 
      CHANGED (
       EVERY
        [TRY (cspF_dist_left_tac n),
         TRY (cspF_dist_right_tac n)])
*}

ML_setup {*
    fun cspF_unwind_tac n = 
      CHANGED (
       EVERY
        [TRY (cspF_unwind_left_tac n),
         TRY (cspF_unwind_right_tac n)])
*}

ML_setup {*
    fun cspF_simp_tac n = 
      CHANGED (
       EVERY
        [TRY (cspF_simp_left_tac n),
         TRY (cspF_simp_right_tac n)])
*}

(*-----------------------------------------------------*
      apply (tactic {* cspF_simp_with_tac a n *})
                      is equal to
      apply ((tactic {* cspF_simp_with_left_tac a n *})? ,
             (tactic {* cspF_simp_with_right_tac a n *})?)
 *-----------------------------------------------------*)

ML_setup {*
    fun cspF_simp_with_tac a n = 
      CHANGED (
       EVERY
        [TRY (cspF_simp_with_left_tac a n),
         TRY (cspF_simp_with_right_tac a n)])
*}

(*-----------------------------------------------------*
      apply (tactic {* cspF_asm_tac n *})
                      is equal to
      apply ((tactic {* cspF_asm_left_tac n *})? ,
             (tactic {* cspF_asm_right_tac n *})?)
 *-----------------------------------------------------*)

ML_setup {*
    fun cspF_asm_tac n = 
      CHANGED (
       EVERY
        [TRY (cspF_asm_left_tac n),
         TRY (cspF_asm_right_tac n)])
*}

end
