           (*-------------------------------------------*
            |        CSP-Prover on Isabelle2004         |
            |                 August 2005               |
            |                                           |
            |        CSP-Prover on Isabelle2005         |
            |                October 2005  (modified)   |
            |                                           |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_T_full_abs = CSP_T_traces:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  disj_not1: (~ P | Q) = (P --> Q)                   *)

declare disj_not1 [simp del]

(*********************************************************
            inverse function : DomT => proc
 *********************************************************)

consts
  head_traces :: "'a domT => 'a set"
  tail_traces :: "'a domT => 'a => 'a domT"

  proc_domT   :: "nat => 'a domT => 'a proc"

  inv_semT    :: "'a domT => 'a proc"      ("inv[[_]]T")

defs
  head_traces_def :
  "head_traces T == {a. EX t. [Ev a]t @t t :t T}"
  tail_traces_def :
  "tail_traces T == (%a. {t. [Ev a]t @t t :t T}t)"

primrec
  "proc_domT 0 = (%T. DIV)"
  "proc_domT (Suc n)
     = (%T. ((! a:(head_traces T) .. a -> proc_domT n (tail_traces T a)) [+] DIV)
            |~| (IF ([Tick]t :t T) THEN SKIP ELSE DIV))"

defs
  inv_semT_def :
   "inv[[T]]T == !nat n .. proc_domT n T"

consts
  ALL_traces_proc_domT :: "'a domT => 'a trace => nat => bool"

defs
  ALL_traces_proc_domT_def :
   "ALL_traces_proc_domT T t n == t :t traces (proc_domT n T) --> t :t T"

(*********************************************************
                     lemmas
 *********************************************************)

(* tail in domT *)

lemma tail_traces_domT:
   "a : head_traces T ==> {t. [Ev a]t @t t :t T} : domT"
apply (simp add: domT_def)
apply (simp add: HC_T1_def)
apply (rule)
apply (simp add: head_traces_def)

apply (simp add: prefix_closed_def)
apply (intro allI impI)
apply (elim conjE exE)
apply (rule memT_prefix_closed)
apply (simp)
apply (simp)
done

(* tail *)

lemma in_tail_traces: 
  "a : head_traces T ==> (s :t tail_traces T a) = ([Ev a]t @t s :t T)"
apply (simp add: tail_traces_def)
apply (simp add: tail_traces_domT CollectT_open_memT)
done

(* head & tail *)

lemma head_tail_traces_only_if:
   "[Ev a]t @t s :t T ==> a : head_traces T & s :t tail_traces T a"
apply (subgoal_tac "a : head_traces T", simp)
apply (simp add: in_tail_traces head_traces_def)
apply (auto simp add: head_traces_def)
done

lemma head_tail_traces_if:
   "[| a : head_traces T ;  s :t tail_traces T a |]
    ==> [Ev a]t @t s :t T"
by (simp add: in_tail_traces)

(* iff *)

lemma head_tail_traces:
   "[Ev a]t @t s :t T = (a : head_traces T & s :t tail_traces T a)"
apply (rule iffI)
apply (simp add: head_tail_traces_only_if)
apply (simp add: head_tail_traces_if)
done

(*----------------------------*
 |       inv_semT lemma       |
 *----------------------------*)

(* only if lemma *)

lemma semT_inv_semT_only_if_lm:
  "ALL n T t. t :t traces (proc_domT n T) --> t :t T"
apply (rule allI)
apply (induct_tac n)

 (* base *)
 apply (simp add: in_traces)

 (* step *)
 apply (fold ALL_traces_proc_domT_def)    (* to prevent infinite simplification *)
 apply (simp (no_asm) add: ALL_traces_proc_domT_def)
 apply (simp add: in_traces)
 apply (intro allI impI)
 apply (elim conjE exE bexE disjE)
 apply (simp)
 apply (simp)
 apply (unfold ALL_traces_proc_domT_def)
 apply (drule_tac x="tail_traces T a" in spec)
 apply (drule_tac x="s" in spec)
 apply (simp add: head_tail_traces)
done

lemma semT_inv_semT_only_if:
  "EX n. t :t traces (proc_domT n T) ==> t :t T"
apply (erule exE)
apply (insert semT_inv_semT_only_if_lm)
apply (drule_tac x="n" in spec)
apply (drule_tac x="T" in spec)
apply (drule_tac x="t" in spec)
by (simp)

(* if lemma *)

lemma semT_inv_semT_if_lm:
  "ALL t T. t :t T
   --> t :t traces (proc_domT (lengtht t) T)"
apply (rule allI)
apply (induct_tac t rule: induct_trace)

 (* []t *)
 apply (simp)

 (* [Tick]t *)
 apply (intro allI impI)
 apply (simp add: in_traces)

 (* [Ev a] @t s *)
 apply (intro allI impI)
 apply (simp add: in_traces)
 apply (drule_tac x="tail_traces T a" in spec)
 apply (simp add: head_tail_traces)
done

lemma semT_inv_semT_if:
  "t :t T
   ==> t :t traces (proc_domT (lengtht t) T)"
by (simp add: semT_inv_semT_if_lm)

(*----------------------------*
 |   inv_semT lemma (main)    |
 *----------------------------*)

lemma semT_inv_semT: "[[ inv[[T]]T ]]T = T"
apply (simp add: inv_semT_def)
apply (rule order_antisym)

 (* <= *)
 apply (rule)
 apply (simp add: semT_def)
 apply (simp add: in_traces)
 apply (erule disjE, simp)
 apply (elim conjE exE)
 apply (rule semT_inv_semT_only_if)
 apply (rule_tac x="a" in exI) 
 apply (simp)

 (* => *)
 apply (rule)
 apply (simp add: semT_def)
 apply (simp add: in_traces)
 apply (rule disjI2)
 apply (rule_tac x="lengtht t" in exI) 
 apply (simp add: semT_inv_semT_if)
done

lemma traces_inv_semT: "traces (inv[[T]]T) = T"
by (simp add: semT_inv_semT[simplified semT_def])

(*----------------------------*
 |      Full abstraction      |
 *----------------------------*)

theorem full_abs_domT: "ALL T. EX P. [[P]]T = T"
apply (rule allI)
apply (rule_tac x="inv[[T]]T" in exI)
apply (simp add: semT_inv_semT)
done

(****************** to add them again ******************)

declare disj_not1   [simp]

end
