           (*-------------------------------------------*
            |        CSP-Prover on Isabelle2004         |
            |               December 2004               |
            |                   June 2005  (modified)   |
            |              September 2005  (modified)   |
            |                                           |
            |        CSP-Prover on Isabelle2005         |
            |                October 2005  (modified)   |
            |                                           |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_T_law = CSP_T_law_SKIP  + CSP_T_law_ref       +
                   CSP_T_law_dist  + CSP_T_law_alpha_par +
                   CSP_T_law_step  + CSP_T_law_index_par +
                   CSP_T_law_fp:

(*****************************************************************

         1. Conditional
         2. Internal ... simp
         3. External ... simp
         4. Mix      ... simp
         5. Extended reflex

 *****************************************************************)

(*-----------------------------------------------------------*
 |                                                           |
 |                 Ext_choice_Int_choice                     |
 |                                                           |
 |  These rules show the difference between models T and F.  |
 |                                                           |
 *-----------------------------------------------------------*)

lemma cspT_Ext_choice_Int_choice:
  "P1 [+] P2 =T P1 |~| P2"
apply (simp add: cspT_semantics)
apply (rule order_antisym)
apply (rule, simp add: in_traces)+
done

lemma cspT_Ext_pre_choice_Rep_int_choice:
  "? :X -> Pf =T ! x:X .. x -> Pf x"
apply (simp add: cspT_semantics)
apply (rule order_antisym)
apply (rule, simp add: in_traces)
apply (force)
apply (rule, simp add: in_traces)
apply (force)
done

lemmas cspT_Ext_Int = cspT_Ext_choice_Int_choice
                      cspT_Ext_pre_choice_Rep_int_choice

(*****************************************************************
                Extended reflex by empty sets
 *****************************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspT_Ext_pre_choice_empty: 
  "? :{} -> Pf =T ? x:{} -> DIV"
apply (simp add: cspT_semantics)
apply (simp add: traces_def)
done

lemma cspT_Ext_pre_choice_STOP:
   "? :{} -> Pf =T STOP"
apply (rule cspT_rw_right)
apply (rule cspT_STOP_step)
apply (rule cspT_Ext_pre_choice_empty)
done

lemmas cspT_Ext_pre_choice_STOP_sym = cspT_Ext_pre_choice_STOP[THEN cspT_sym]

lemma cspT_Rep_int_choice0_DIV:
   "!! :{} .. Pf =T DIV"
apply (simp add: cspT_semantics)
apply (simp add: traces_def)
done

lemma cspT_Rep_int_choice_fun_DIV:
   "inj f ==> !!<f> :{} .. Pf =T DIV"
apply (simp add: cspT_semantics)
apply (simp add: traces_def)
done

lemma cspT_Rep_int_choice1_DIV:
   "! :{} .. Pf =T DIV"
by (simp add: cspT_Rep_int_choice_fun_DIV)

lemma cspT_Rep_int_choice2_DIV:
   "!set :{} .. Pf =T DIV"
by (simp add: cspT_Rep_int_choice_fun_DIV)

lemma cspT_Rep_int_choice3_DIV:
   "!nat :{} .. Pf =T DIV"
by (simp add: cspT_Rep_int_choice_fun_DIV)

lemmas cspT_Rep_int_choice_DIV = cspT_Rep_int_choice0_DIV
                                 cspT_Rep_int_choice1_DIV
                                 cspT_Rep_int_choice2_DIV
                                 cspT_Rep_int_choice3_DIV

lemmas cspT_Rep_int_choice_DIV_sym = cspT_Rep_int_choice0_DIV[THEN cspT_sym]

lemma cspT_Ext_pre_choice_empty_reflex:
   "? :{} -> Pf =T ? :{} -> Qf"
apply (rule cspT_trans[of _ "STOP"])
apply (simp add: cspT_Ext_pre_choice_STOP)
apply (simp add: cspT_Ext_pre_choice_STOP_sym)
done

lemma cspT_Rep_int_choice0_empty_reflex:
   "!! :{} .. Pf =T !! :{} .. Qf"
apply (rule cspT_trans[of _ "DIV"])
apply (simp add: cspT_Rep_int_choice_DIV)
apply (simp add: cspT_Rep_int_choice_DIV_sym)
done

lemma cspT_Rep_int_choice_fun_empty_reflex:
   "inj f ==> !!<f> :{} .. Pf =T !!<f> :{} .. Qf"
apply (simp add: Rep_int_choice_fun_def)
apply (simp add: cspT_Rep_int_choice0_empty_reflex)
done

lemma cspT_Rep_int_choice1_empty_reflex:
   "! :{} .. Pf =T ! :{} .. Qf"
by (simp add: cspT_Rep_int_choice_fun_empty_reflex)

lemma cspT_Rep_int_choice2_empty_reflex:
   "!set :{} .. Pf =T !set :{} .. Qf"
by (simp add: cspT_Rep_int_choice_fun_empty_reflex)

lemma cspT_Rep_int_choice3_empty_reflex:
   "!nat :{} .. Pf =T !nat :{} .. Qf"
by (simp add: cspT_Rep_int_choice_fun_empty_reflex)

lemmas cspT_Rep_int_choice_empty_reflex 
     = cspT_Rep_int_choice0_empty_reflex
       cspT_Rep_int_choice1_empty_reflex
       cspT_Rep_int_choice2_empty_reflex
       cspT_Rep_int_choice3_empty_reflex

lemmas cspT_reflex_ex = cspT_Ext_pre_choice_STOP
                        cspT_Ext_pre_choice_STOP_sym
                        cspT_Rep_int_choice_DIV
                        cspT_Rep_int_choice_DIV_sym
                        cspT_Ext_pre_choice_empty_reflex
                        cspT_Rep_int_choice_empty_reflex

declare cspT_reflex_ex [simp]

(*********************************************************
                       IF bool
 *********************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspT_Conditional_split: 
  "IF b THEN P ELSE Q =T (if b then P else Q)"
apply (simp add: cspT_semantics)
apply (simp add: traces_def)
done

lemma cspT_Conditional_True:
  "IF True THEN P ELSE Q =T P"
apply (rule cspT_rw_left)
apply (rule cspT_Conditional_split)
by (simp)

lemma cspT_Conditional_False:
  "IF False THEN P ELSE Q =T Q"
apply (rule cspT_rw_left)
apply (rule cspT_Conditional_split)
by (simp)

lemmas cspT_Conditional = cspT_Conditional_True cspT_Conditional_False

(*****************************************************************
                       Internal ... simp                         
 *****************************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspT_DIV_unwind:
   "DIV =T ! x:{} .. DIV"
apply (simp add: cspT_semantics)
apply (simp add: traces_def)
done

(*** ! :{a} ***)

lemma cspT_Rep_int_choice0_unique:
  "!! :{a} .. Pf =T Pf a"
apply (simp add: cspT_semantics)
apply (rule order_antisym)

 apply (rule, simp add: in_traces, force)
 apply (rule, simp add: in_traces)
done

lemma cspT_Rep_int_choice_fun_unique:
  "inj f ==> !!<f> :{a} .. Pf =T Pf a"
apply (simp add: Rep_int_choice_fun_def)
apply (simp add: cspT_semantics)
apply (rule order_antisym)

 apply (rule, simp add: in_traces, force)
 apply (rule, simp add: in_traces)
done

lemma cspT_Rep_int_choice1_unique:
  "! :{a} .. Pf =T Pf a"
by (simp add: cspT_Rep_int_choice_fun_unique)

lemma cspT_Rep_int_choice2_unique:
  "!set :{a} .. Pf =T Pf a"
by (simp add: cspT_Rep_int_choice_fun_unique)

lemma cspT_Rep_int_choice3_unique:
  "!nat :{a} .. Pf =T Pf a"
by (simp add: cspT_Rep_int_choice_fun_unique)

lemmas cspT_Rep_int_choice_unique = cspT_Rep_int_choice0_unique
                                    cspT_Rep_int_choice1_unique
                                    cspT_Rep_int_choice2_unique
                                    cspT_Rep_int_choice3_unique

(* Int choice idempotence *)

lemma cspT_Int_choice_idem: 
  "P |~| P =T P"
apply (simp add: cspT_semantics)
apply (rule order_antisym)
 apply (rule, simp add: in_traces)+
done

lemma cspT_Int_choice_idem_weak: 
  "P =T Q ==> P |~| Q =T P"
apply (rule cspT_rw_left)
apply (rule cspT_decompo)
apply (rule cspT_reflex)
apply (erule cspT_symE)
apply (simp)
apply (simp add: cspT_Int_choice_idem)
done

lemmas cspT_Int_choice_rule = cspT_DIV_unwind 
                              cspT_Rep_int_choice_unique 
                              cspT_Int_choice_idem

lemmas cspT_Int_choice_simp = cspT_DIV_unwind 
                              cspT_Rep_int_choice_unique 
                              cspT_Int_choice_idem_weak

(*****************************************************************
                      External ... simp
 *****************************************************************)

(*** STOP [+] P ***)

lemma cspT_Ext_choice_unit_l: 
  "STOP [+] P =T P"
apply (simp add: cspT_semantics)
apply (rule order_antisym)
 apply (rule, simp add: in_traces)
 apply (force)
 apply (rule, simp add: in_traces)
done

lemma cspT_Ext_choice_unit_l_hnf: 
  "? :{} -> Qf [+] P =T P"
apply (rule cspT_rw_left)
apply (rule cspT_decompo)
apply (rule cspT_Ext_pre_choice_STOP)
apply (rule cspT_reflex)
apply (simp add: cspT_Ext_choice_unit_l)
done

lemma cspT_Ext_choice_STOP_r: 
  "P [+] STOP =T P"
apply (rule cspT_rw_left)
apply (rule cspT_Ext_choice_commut)
apply (simp add: cspT_Ext_choice_unit_l)
done

lemma cspT_Ext_choice_STOP_r_hnf: 
  "P [+] ? :{} -> Qf =T P"
apply (rule cspT_rw_left)
apply (rule cspT_Ext_choice_commut)
apply (simp add: cspT_Ext_choice_unit_l_hnf)
done

(*** weak ***)

lemma cspT_Ext_choice_unit_l_weak: 
  "Q =T STOP 
   ==> Q [+] P =T P"
apply (rule cspT_rw_left)
apply (rule cspT_decompo)
apply (assumption)
apply (rule cspT_reflex)
apply (simp add: cspT_Ext_choice_unit_l)
done

lemma cspT_Ext_choice_unit_r_weak: 
  "Q =T STOP 
   ==> P [+] Q =T P"
apply (rule cspT_rw_left)
apply (rule cspT_commut)
apply (simp add: cspT_Ext_choice_unit_l_weak)
done

lemmas cspT_Ext_choice_rule = cspT_Ext_choice_unit_l
                              cspT_Ext_choice_unit_l_hnf
                              cspT_Ext_choice_STOP_r
                              cspT_Ext_choice_STOP_r_hnf

lemmas cspT_Ext_choice_simp = cspT_Ext_choice_unit_l_weak
                              cspT_Ext_choice_unit_r_weak

(*****************************************************************
                      Mix ... simp
 *****************************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

(*** <= Timeout ***)

lemma cspT_Timeout_right:
  "[| P <=T Q1 ; P <=T Q2 |] ==> P <=T Q1 [> Q2"
apply (rule cspT_rw_right)
apply (rule cspT_dist)
apply (rule cspT_Int_choice_right)
apply (rule cspT_Ext_choice_right, simp_all)
apply (rule cspT_rw_right)
apply (rule cspT_Ext_choice_unit_l, simp_all)
done

(*** STOP [> P  =  P ***)

lemma cspT_STOP_Timeout:
  "STOP [> P =T P"
apply (rule cspT_rw_left)
apply (rule cspT_dist)
apply (rule cspT_rw_left)
apply (rule cspT_Int_choice_idem)
apply (rule cspT_Ext_choice_unit_l_weak)
by (simp)

lemma cspT_STOP_Timeout_hnf:
  "(? :{} -> Qf1 |~| ? :{} -> Qf2) [+] Q =T Q"
apply (rule cspT_rw_left)
apply (rule cspT_decompo)
apply (rule cspT_decompo)
apply (rule cspT_Ext_pre_choice_STOP)
apply (rule cspT_Ext_pre_choice_STOP)
apply (rule cspT_reflex)
apply (simp add: cspT_STOP_Timeout)
done

lemma cspT_STOP_Timeout_weak:
  "[| P1 =T STOP ;
      P2 =T STOP |]
   ==> (P1 |~| P2) [+] Q =T Q"
apply (rule cspT_rw_left)
apply (rule cspT_decompo)
apply (rule cspT_decompo)
apply (simp_all)
apply (rule cspT_reflex)
apply (simp add: cspT_STOP_Timeout)
done

lemmas cspT_Mix_choice_rule = cspT_STOP_Timeout cspT_STOP_Timeout_hnf

lemmas cspT_Mix_choice_simp = cspT_STOP_Timeout_weak

(*-----------------------------*
 |      simp rule for cspT     |
 *-----------------------------*)

lemmas cspT_choice_rule  = cspT_Int_choice_rule cspT_Ext_choice_rule cspT_Mix_choice_rule
lemmas cspT_choice_simp  = cspT_Int_choice_simp cspT_Ext_choice_simp cspT_Mix_choice_simp

end
