           (*-------------------------------------------*
            |        CSP-Prover on Isabelle2004         |
            |               December 2004               |
            |                   June 2005  (modified)   |
            |              September 2005  (modified)   |
            |                                           |
            |        CSP-Prover on Isabelle2005         |
            |                October 2005  (modified)   |
            |                                           |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_T_law_decompo = CSP_T_traces:

(*------------------------------------------------*
 |                                                |
 |      laws for monotonicity and congruence      |
 |                                                |
 *------------------------------------------------*)

(*********************************************************
                        Act_prefix mono
 *********************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspT_Act_prefix_mono:
  "[| a = b ; P <=T Q |] ==> a -> P <=T b -> Q"
apply (simp add: cspT_semantics)
apply (simp add: subdomT_iff)
apply (intro allI impI)
apply (simp add: in_traces)
apply (fast)
done

lemma cspT_Act_prefix_cong:
  "[| a = b ; P =T Q |] ==> a -> P =T b -> Q"
apply (simp add: cspT_eq_ref_iff)
apply (simp add: cspT_Act_prefix_mono)
done

(*********************************************************
                   Ext_pre_choice mono 
 *********************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspT_Ext_pre_choice_mono:
  "[| X = Y ; !! a. a:Y ==> Pf a <=T Qf a |] ==> ? :X -> Pf <=T ? :Y -> Qf"
apply (simp add: cspT_semantics)
apply (simp add: subdomT_iff)
apply (intro allI impI)
apply (simp add: in_traces)
apply (fast)
done

lemma cspT_Ext_pre_choice_cong:
  "[| X = Y ; !! a. a:Y ==> Pf a =T Qf a |] ==> ? :X -> Pf =T ? :Y -> Qf"
by (simp add: cspT_eq_ref_iff cspT_Ext_pre_choice_mono)

(*********************************************************
                      Ext choice mono
 *********************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspT_Ext_choice_mono:
  "[| P1 <=T Q1 ; P2 <=T Q2 |] ==> P1 [+] P2 <=T Q1 [+] Q2"
apply (simp add: cspT_semantics)
apply (simp add: subdomT_iff)
apply (intro allI impI)
apply (simp add: in_traces)
apply (fast)
done

lemma cspT_Ext_choice_cong:
  "[| P1 =T Q1 ; P2 =T Q2 |] ==> P1 [+] P2 =T Q1 [+] Q2"
by (simp add: cspT_eq_ref_iff cspT_Ext_choice_mono)

(*********************************************************
                      Int choice mono
 *********************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspT_Int_choice_mono:
  "[| P1 <=T Q1 ; P2 <=T Q2 |] ==> P1 |~| P2 <=T Q1 |~| Q2"
apply (simp add: cspT_semantics)
apply (simp add: subdomT_iff)
apply (intro allI impI)
apply (simp add: in_traces)
apply (fast)
done

lemma cspT_Int_choice_cong:
  "[| P1 =T Q1 ; P2 =T Q2 |] ==> P1 |~| P2 =T Q1 |~| Q2"
by (simp add: cspT_eq_ref_iff cspT_Int_choice_mono)

(*********************************************************
               replicated internal choice
 *********************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspT_Rep_int_choice_mono0:
  "[| X = Y ; !! a. a:Y ==> Pf a <=T Qf a |]
      ==> !! :X .. Pf <=T !! :Y .. Qf"
apply (simp add: cspT_semantics)
apply (simp add: subdomT_iff)
apply (intro allI impI)
apply (simp add: in_traces)
apply (fast)
done

lemma cspT_Rep_int_choice_cong0:
  "[| X = Y ; !! a. a:Y ==> Pf a =T Qf a |]
      ==> !! :X .. Pf =T !! :Y .. Qf"
by (simp add: cspT_eq_ref_iff cspT_Rep_int_choice_mono0)

(*** fun ***)

lemma cspT_Rep_int_choice_mono_fun:
  "[| inj f ; X = Y ; !! a. a:Y ==> Pf a <=T Qf a |]
      ==> !!<f> :X .. Pf <=T !!<f> :Y .. Qf"
apply (simp add: Rep_int_choice_fun_def)
apply (rule cspT_Rep_int_choice_mono0)
apply (auto)
done

lemma cspT_Rep_int_choice_cong_fun:
  "[| inj f ; X = Y ; !! a. a:Y ==> Pf a =T Qf a |]
      ==> !!<f> :X .. Pf =T !!<f> :Y .. Qf"
by (simp add: cspT_eq_ref_iff cspT_Rep_int_choice_mono_fun)

(*** Tri1 ***)

lemma cspT_Rep_int_choice_mono_Tri1:
  "[| X = Y ; !! a. a:Y ==> Pf a <=T Qf a |]
      ==> ! :X .. Pf <=T ! :Y .. Qf"
by (simp add: cspT_Rep_int_choice_mono_fun)

lemma cspT_Rep_int_choice_cong_Tri1:
  "[| X = Y ; !! a. a:Y ==> Pf a =T Qf a |]
      ==> ! :X .. Pf =T ! :Y .. Qf"
by (simp add: cspT_Rep_int_choice_cong_fun)

(*** Tri2 ***)

lemma cspT_Rep_int_choice_mono_Tri2:
  "[| X = Y ; !! a. a:Y ==> Pf a <=T Qf a |]
      ==> !set :X .. Pf <=T !set :Y .. Qf"
by (simp add: cspT_Rep_int_choice_mono_fun)

lemma cspT_Rep_int_choice_cong_Tri2:
  "[| X = Y ; !! a. a:Y ==> Pf a =T Qf a |]
      ==> !set :X .. Pf =T !set :Y .. Qf"
by (simp add: cspT_Rep_int_choice_cong_fun)

(*** Tri3 ***)

lemma cspT_Rep_int_choice_mono_Tri3:
  "[| X = Y ; !! a. a:Y ==> Pf a <=T Qf a |]
      ==> !nat :X .. Pf <=T !nat :Y .. Qf"
by (simp add: cspT_Rep_int_choice_mono_fun)

lemma cspT_Rep_int_choice_cong_Tri3:
  "[| X = Y ; !! a. a:Y ==> Pf a =T Qf a |]
      ==> !nat :X .. Pf =T !nat :Y .. Qf"
by (simp add: cspT_Rep_int_choice_cong_fun)

(*** f ***)

lemma cspT_Rep_int_choice_mono_f:
  "[| inj f ; X = Y ; !! a. a:Y ==> Pf a <=T Qf a |]
      ==> !<f> :X .. Pf <=T !<f> :Y .. Qf"
apply (simp add: Rep_int_choice_f_def)
apply (rule cspT_Rep_int_choice_mono_Tri1)
apply (auto)
done

lemma cspT_Rep_int_choice_cong_f:
  "[| inj f ; X = Y ; !! a. a:Y ==> Pf a =T Qf a |]
      ==> !<f> :X .. Pf =T !<f> :Y .. Qf"
apply (simp add: Rep_int_choice_f_def)
apply (rule cspT_Rep_int_choice_cong_Tri1)
apply (auto)
done

lemmas cspT_Rep_int_choice_mono = cspT_Rep_int_choice_mono0
                                  cspT_Rep_int_choice_mono_Tri1
                                  cspT_Rep_int_choice_mono_Tri2
                                  cspT_Rep_int_choice_mono_Tri3
                                  cspT_Rep_int_choice_mono_f

lemmas cspT_Rep_int_choice_cong = cspT_Rep_int_choice_cong0
                                  cspT_Rep_int_choice_cong_Tri1
                                  cspT_Rep_int_choice_cong_Tri2
                                  cspT_Rep_int_choice_cong_Tri3
                                  cspT_Rep_int_choice_cong_f

(*********************************************************
                   Conditional mono
 *********************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspT_Conditional_mono:
  "[| b1 = b2 ; P1 <=T Q1 ; P2 <=T Q2 |]
           ==> IF b1 THEN P1 ELSE P2 <=T
               IF b2 THEN Q1 ELSE Q2"
apply (simp add: cspT_semantics)
apply (simp add: subdomT_iff)
apply (intro allI impI)
apply (simp add: in_traces)
done

lemma cspT_Conditional_cong:
  "[| b1 = b2 ; P1 =T Q1 ;  P2 =T Q2 |]
           ==> IF b1 THEN P1 ELSE P2 =T
               IF b2 THEN Q1 ELSE Q2"
by (simp add: cspT_eq_ref_iff cspT_Conditional_mono)

(*********************************************************
                     Parallel mono 
 *********************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspT_Parallel_mono:
  "[| X = Y ; P1 <=T Q1 ; P2 <=T Q2 |]
           ==> P1 |[X]| P2 <=T Q1 |[Y]| Q2"
apply (simp add: cspT_semantics)
apply (simp add: subdomT_iff)
apply (intro allI impI)
apply (simp add: in_traces)
apply (fast)
done

lemma cspT_Parallel_cong:
  "[| X = Y ; P1 =T Q1 ; P2 =T Q2 |]
           ==> P1 |[X]| P2 =T Q1 |[Y]| Q2"
by (simp add: cspT_eq_ref_iff cspT_Parallel_mono)

(*********************************************************
                        Hiding mono
 *********************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspT_Hiding_mono:
  "[| X = Y ; P <=T Q |] ==> P -- X <=T Q -- Y"
apply (simp add: cspT_semantics)
apply (simp add: subdomT_iff)
apply (intro allI impI)
apply (simp add: in_traces)
apply (fast)
done

lemma cspT_Hiding_cong:
  "[| X = Y ; P =T Q |] ==> P -- X =T Q -- Y"
by (simp add: cspT_eq_ref_iff cspT_Hiding_mono)

(*********************************************************
                        Renaming mono
 *********************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspT_Renaming_mono:
  "[| r1 = r2 ; P <=T Q |] ==> P [[r1]] <=T Q [[r2]]"
apply (simp add: cspT_semantics)
apply (simp add: subdomT_iff)
apply (intro allI impI)
apply (simp add: in_traces)
apply (fast)
done

lemma cspT_Renaming_cong:
  "[| r1 = r2 ; P =T Q |] ==> P [[r1]] =T Q [[r2]]"
by (simp add: cspT_eq_ref_iff cspT_Renaming_mono)

(*********************************************************
               Sequential composition mono 
 *********************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspT_Seq_compo_mono:
  "[| P1 <=T Q1 ; P2 <=T Q2 |]
           ==> P1 ;; P2 <=T Q1 ;; Q2"
apply (simp add: cspT_semantics)
apply (simp add: subdomT_iff)
apply (intro allI impI)
apply (simp add: in_traces)
apply (fast)
done

lemma cspT_Seq_compo_cong:
  "[| P1 =T Q1 ; P2 =T Q2 |] ==> P1 ;; P2 =T Q1 ;; Q2"
by (simp add: cspT_eq_ref_iff cspT_Seq_compo_mono)

(*********************************************************
                        Timeout mono
 *********************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspT_Timeout_mono:
  "[| P1 <=T Q1 ; P2 <=T Q2 |]
      ==> P1 [> P2 <=T Q1 [> Q2"
apply (rule cspT_Ext_choice_mono)
apply (rule cspT_Int_choice_mono)
by (simp_all)

lemma cspT_Timeout_cong:
  "[| P1 =T Q1 ; P2 =T Q2 |]
      ==> P1 [> P2 =T Q1 [> Q2"
by (simp add: cspT_eq_ref_iff cspT_Timeout_mono)

(*------------------------------------------------------*
 |                       alias                          |
 *------------------------------------------------------*)

lemmas cspT_free_mono =
   cspT_Ext_choice_mono cspT_Int_choice_mono cspT_Parallel_mono
   cspT_Hiding_mono cspT_Renaming_mono cspT_Seq_compo_mono

lemmas cspT_mono = cspT_free_mono
   cspT_Act_prefix_mono cspT_Ext_pre_choice_mono 
   cspT_Rep_int_choice_mono cspT_Conditional_mono 

lemmas cspT_free_cong =
   cspT_Ext_choice_cong cspT_Int_choice_cong cspT_Parallel_cong
   cspT_Hiding_cong cspT_Renaming_cong cspT_Seq_compo_cong

lemmas cspT_cong = cspT_free_cong
   cspT_Act_prefix_cong cspT_Ext_pre_choice_cong 
   cspT_Rep_int_choice_cong cspT_Conditional_cong 

lemmas cspT_free_decompo = cspT_free_mono cspT_free_cong
lemmas cspT_decompo = cspT_mono cspT_cong

lemmas cspT_rm_head_mono = cspT_Act_prefix_mono cspT_Ext_pre_choice_mono 
lemmas cspT_rm_head_cong = cspT_Act_prefix_cong cspT_Ext_pre_choice_cong
lemmas cspT_rm_head      = cspT_rm_head_mono cspT_rm_head_cong

end
