           (*-------------------------------------------*
            |        CSP-Prover on Isabelle2004         |
            |               December 2004               |
            |                   July 2005  (modified)   |
            |              September 2005  (modified)   |
            |                                           |
            |        CSP-Prover on Isabelle2005         |
            |                October 2005  (modified)   |
            |                                           |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_T_law_dist = CSP_T_law_commut + CSP_T_law_decompo:

(*****************************************************************

      distribution of internal choice over ...

         1. (P1 |~| P2) [+] Q
         2. Q [+] (P1 |~| P2)
         3. (P1 |~| P2) |[X]| Q
         4. Q |[X]| (P1 |~| P2)
         5. (P1 |~| P2) -- X
         6. (P1 |~| P2) [[r]]

 *****************************************************************)

(*********************************************************
                dist law for Ext_choice (l)
 *********************************************************)

lemma cspT_Ext_choice_dist_l: 
  "(P1 |~| P2) [+] Q =T
   (P1 [+] Q) |~| (P2 [+] Q)"
apply (simp add: cspT_semantics)
apply (rule order_antisym)
apply (rule, simp add: in_traces, fast)+
done

(*********************************************************
                dist law for Ext_choice (r)
 *********************************************************)

lemma cspT_Ext_choice_dist_r: 
  "P [+] (Q1 |~| Q2) =T
   (P [+] Q1) |~| (P [+] Q2)"
apply (rule cspT_rw_left)
apply (rule cspT_commut)
apply (rule cspT_rw_left)
apply (rule cspT_Ext_choice_dist_l)
apply (rule cspT_decompo)
apply (rule cspT_commut)+
done

(*********************************************************
                dist law for Parallel (l)
 *********************************************************)

lemma cspT_Parallel_dist_l: 
  "(P1 |~| P2) |[X]| Q =T
   (P1 |[X]| Q) |~| (P2 |[X]| Q)"
apply (simp add: cspT_semantics)
apply (rule order_antisym)
apply (rule, simp add: in_traces, fast)+
done

(*********************************************************
                dist law for Parallel (r)
 *********************************************************)

lemma cspT_Parallel_dist_r: 
  "P |[X]| (Q1 |~| Q2) =T
   (P |[X]| Q1) |~| (P |[X]| Q2)"
apply (rule cspT_rw_left)
apply (rule cspT_commut)
apply (rule cspT_rw_left)
apply (rule cspT_Parallel_dist_l)
apply (rule cspT_decompo)
apply (rule cspT_commut)+
done

(*********************************************************
                dist law for Hiding
 *********************************************************)

lemma cspT_Hiding_dist: 
  "(P1 |~| P2) -- X =T
   (P1 -- X) |~| (P2 -- X)"
apply (simp add: cspT_semantics)
apply (rule order_antisym)
apply (rule, simp add: in_traces, fast)+
done

(*********************************************************
               dist law for Renaming
 *********************************************************)

lemma cspT_Renaming_dist: 
  "(P1 |~| P2) [[r]] =T
   (P1 [[r]]) |~| (P2 [[r]])"
apply (simp add: cspT_semantics)
apply (rule order_antisym)
apply (rule, simp add: in_traces, fast)+
done

(*********************************************************
         dist law for Sequential composition
 *********************************************************)

lemma cspT_Seq_compo_dist: 
  "(P1 |~| P2) ;; Q =T
   (P1 ;; Q) |~| (P2 ;; Q)"
apply (simp add: cspT_semantics)
apply (rule order_antisym)
apply (rule, simp add: in_traces, fast)+
done

(*********************************************************
                     dist laws
 *********************************************************)

lemmas cspT_dist = cspT_Ext_choice_dist_l cspT_Ext_choice_dist_r
                   cspT_Parallel_dist_l   cspT_Parallel_dist_r
                   cspT_Hiding_dist       cspT_Renaming_dist
                   cspT_Seq_compo_dist

(*****************************************************************

      distribution of internal bind over ...

         1. (!! :X .. Pf) [+] Q
         2. Q [+] (!! :X .. Pf)
         3. (!! :X .. Pf) |[X]| Q
         4. Q |[X]| (!! :X .. Pf)
         5. (!! :X .. Pf) -- X
         6. (!! :X .. Pf) [[r]]

 *****************************************************************)

(*********************************************************
                Rep_dist law for Ext_choice (l)
 *********************************************************)

lemma cspT_Ext_choice_Rep_dist0_l: 
  "X ~= {} ==> (!! :X .. Pf) [+] Q =T
               !! x:X .. (Pf x [+] Q)"
apply (simp add: cspT_semantics)
apply (rule order_antisym)
apply (rule, simp add: in_traces, fast)+
done

(*********************************************************
                Rep_dist0 law for Ext_choice (r)
 *********************************************************)

lemma cspT_Ext_choice_Rep_dist0_r: 
  "X ~= {} ==> P [+] (!! :X .. Qf) =T
               !! x:X .. (P [+] Qf x)"
apply (rule cspT_rw_left)
apply (rule cspT_commut)
apply (rule cspT_rw_left)
apply (rule cspT_Ext_choice_Rep_dist0_l, simp)
apply (rule cspT_decompo, simp)
apply (rule cspT_commut)
done

(*********************************************************
                Rep_dist0 law for Parallel (l)
 *********************************************************)

lemma cspT_Parallel_Rep_dist0_l: 
  "Y ~= {} ==>
     (!! :Y .. Pf) |[X]| Q =T
     !! x:Y .. (Pf x |[X]| Q)"
apply (simp add: cspT_semantics)
apply (rule order_antisym)

 apply (rule, simp add: in_traces)
 apply (elim disjE conjE exE bexE)
  apply (subgoal_tac "EX a. a:Y")
  apply (elim exE)
  apply (rule disjI2)
  apply (rule_tac x="a" in bexI)
  apply (simp)
  apply (rule_tac x="[]t" in exI)
  apply (rule_tac x="ta" in exI)
  apply (simp)
  apply (simp)
  apply (fast)
  (* *)
  apply (rule disjI2)
  apply (rule_tac x="a" in bexI)
  apply (fast)
  apply (simp)

 apply (rule, simp add: in_traces)
 apply (elim disjE conjE exE bexE)
  apply (rule_tac x="[]t" in exI)
  apply (rule_tac x="[]t" in exI)
  apply (simp)
  apply (fast)
done

(*********************************************************
                Rep_dist0 law for Parallel (r)
 *********************************************************)

lemma cspT_Parallel_Rep_dist0_r: 
  "Y ~= {} ==>
     P |[X]| (!! :Y .. Qf) =T
     !! x:Y .. (P |[X]| Qf x)"
apply (rule cspT_rw_left)
apply (rule cspT_commut)
apply (rule cspT_rw_left)
apply (rule cspT_Parallel_Rep_dist0_l, simp)
apply (rule cspT_decompo, simp)
apply (rule cspT_commut)
done

(*********************************************************
                Rep_dist0 law for Hiding
 *********************************************************)

lemma cspT_Hiding_Rep_dist0: 
  "(!! :Y .. Pf) -- X =T
   !! x:Y .. (Pf x -- X)"
apply (simp add: cspT_semantics)
apply (rule order_antisym)

 apply (rule, simp add: in_traces)
 apply (elim disjE conjE exE, simp, fast)

 apply (rule, simp add: in_traces)
 apply (elim disjE conjE exE bexE)
 apply (rule_tac x="[]t" in exI, simp)
 apply (rule_tac x="s" in exI, fast)
done

(*********************************************************
                Rep_dist0 law for Renaming
 *********************************************************)

lemma cspT_Renaming_Rep_dist0: 
  "(!! :X .. Pf) [[r]] =T
   !! x:X .. (Pf x [[r]])"
apply (simp add: cspT_semantics)
apply (rule order_antisym)

 apply (rule, simp add: in_traces)
 apply (elim disjE conjE exE, simp, fast)
 apply (rule, simp add: in_traces)
 apply (elim disjE conjE exE, simp, fast)
done

(*********************************************************
          Rep_dist0 law for Sequential composition
 *********************************************************)

lemma cspT_Seq_compo_Rep_dist0: 
  "(!! :X .. Pf) ;; Q =T
   !! x:X .. (Pf x ;; Q)"
apply (simp add: cspT_semantics)
apply (rule order_antisym)

 apply (rule, simp add: in_traces)
 apply (elim disjE conjE exE bexE)
 apply (simp)
 apply (fast)
 apply (force)
 apply (rule disjI2)
 apply (rule_tac x="a" in bexI)
 apply (force)
 apply (simp)

 apply (rule, simp add: in_traces)
 apply (elim disjE conjE exE bexE)
 apply (rule disjI1)
 apply (rule_tac x="[]t" in exI, simp)
 apply (fast)
 apply (fast)
done

(*********************************************************
                     Rep_dist0 laws
 *********************************************************)

lemmas cspT_Rep_dist0 = cspT_Ext_choice_Rep_dist0_l cspT_Ext_choice_Rep_dist0_r
                        cspT_Parallel_Rep_dist0_l   cspT_Parallel_Rep_dist0_r
                        cspT_Hiding_Rep_dist0       cspT_Renaming_Rep_dist0
                        cspT_Seq_compo_Rep_dist0

(*****************************************************************

      for convenience

         1. (!!<f> :X .. Pf) [+] Q
         2. Q [+] (!!<f> :X .. Pf)
         3. (!!<f> :X .. Pf) |[X]| Q
         4. Q |[X]| (!!<f> :X .. Pf)
         5. (!!<f> :X .. Pf) -- X
         6. (!!<f> :X .. Pf) [[r]]

 *****************************************************************)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspT_Ext_choice_Rep_dist_fun_l: 
  "[| inj f ; X ~= {} |]
   ==> (!!<f> :X .. Pf) [+] Q =T !!<f> x:X .. (Pf x [+] Q)"
by (simp add: Rep_int_choice_fun_def cspT_Rep_dist0)

lemma cspT_Ext_choice_Rep_dist_fun_r: 
  "[| inj f ; X ~= {} |]
   ==> P [+] (!!<f> :X .. Qf) =T !!<f> x:X .. (P [+] Qf x)"
by (simp add: Rep_int_choice_fun_def cspT_Rep_dist0)

lemma cspT_Parallel_Rep_dist_fun_l: 
  "[| inj f ; Y ~= {} |]
   ==> (!!<f> :Y .. Pf) |[X]| Q =T !!<f> x:Y .. (Pf x |[X]| Q)"
by (simp add: Rep_int_choice_fun_def cspT_Rep_dist0)

lemma cspT_Parallel_Rep_dist_fun_r: 
  "[| inj f ; Y ~= {} |]
   ==> P |[X]| (!!<f> :Y .. Qf) =T !!<f> x:Y .. (P |[X]| Qf x)"
by (simp add: Rep_int_choice_fun_def cspT_Rep_dist0)

lemma cspT_Hiding_Rep_dist_fun: 
  "inj f ==> (!!<f> :Y .. Pf) -- X =T !!<f> x:Y .. (Pf x -- X)"
by (simp add: Rep_int_choice_fun_def cspT_Rep_dist0)

lemma cspT_Renaming_Rep_dist_fun: 
  "inj f ==> (!!<f> :X .. Pf) [[r]] =T !!<f> x:X .. (Pf x [[r]])"
by (simp add: Rep_int_choice_fun_def cspT_Rep_dist0)

lemma cspT_Seq_compo_Rep_dist_fun:
  "inj f ==> (!!<f> :X .. Pf) ;; Q =T !!<f> x:X .. (Pf x ;; Q)"
by (simp add: Rep_int_choice_fun_def cspT_Rep_dist0)

(*********************************************************
                     Rep_dist laws
 *********************************************************)

lemmas cspT_Rep_dist_fun = cspT_Ext_choice_Rep_dist_fun_l cspT_Ext_choice_Rep_dist_fun_r
                           cspT_Parallel_Rep_dist_fun_l   cspT_Parallel_Rep_dist_fun_r
                           cspT_Hiding_Rep_dist_fun       cspT_Renaming_Rep_dist_fun
                           cspT_Seq_compo_Rep_dist_fun

(*** all rules ***)

lemmas cspT_Rep_dist = cspT_Rep_dist0 cspT_Rep_dist_fun

(*****************************************************************

      additional distribution of internal bind over ...

         1. (!! :X .. (a -> P))
         2. (!! :Y .. (? :X -> P))

 *****************************************************************)

(*********************************************************
              Rep_dist law for Act_prefix
 *********************************************************)

lemma cspT_Act_prefix_Rep_dist0:
  "X ~= {} ==> 
   a -> (!! :X .. Pf) =T !! x:X .. (a -> Pf x)"
apply (simp add: cspT_semantics)
apply (rule order_antisym)
(* <= *)
 apply (rule)
 apply (simp add: in_traces)
 apply (elim disjE conjE exE bexE)
 apply (simp_all)
 apply (force)
 apply (force)

(* => *)
 apply (rule)
 apply (simp add: in_traces)
 apply (elim disjE conjE exE bexE)
 apply (simp_all)
 apply (force)
done

(*********************************************************
              Rep_dist0 law for Ext_pre_choice
 *********************************************************)

lemma cspT_Ext_pre_choice_Rep_dist0:
  "Y ~= {} ==> 
   ? x:X -> (!! y:Y .. (Pf y) x) =T !! y:Y .. (? :X -> (Pf y))"
apply (simp add: cspT_semantics)
apply (rule order_antisym)
(* <= *)
 apply (rule)
 apply (simp add: in_traces)
 apply (elim disjE conjE exE bexE)
 apply (simp_all)
 apply (force)
 apply (force)

(* => *)
 apply (rule)
 apply (simp add: in_traces)
 apply (elim disjE conjE exE bexE)
 apply (simp_all)
 apply (force)
done

(*****************************************************************

      for convenience

         1. (!!<f> :X .. (a -> P))
         2. (!!<f> :Y .. (? :X -> P))

 *****************************************************************)

lemma cspT_Act_prefix_Rep_dist_fun:
  "X ~= {} ==> 
   a -> (!!<f> :X .. Pf) =T !!<f> x:X .. (a -> Pf x)"
by (simp add: Rep_int_choice_fun_def cspT_Act_prefix_Rep_dist0)

lemma cspT_Ext_pre_choice_Rep_dist_fun:
  "Y ~= {} ==> 
   ? x:X -> (!!<f> y:Y .. (Pf y) x) =T !!<f> y:Y .. (? :X -> (Pf y))"
by (simp add: Rep_int_choice_fun_def cspT_Ext_pre_choice_Rep_dist0)

(*** arias ***)

lemmas cspT_Act_prefix_Rep_dist 
     = cspT_Act_prefix_Rep_dist0
       cspT_Act_prefix_Rep_dist_fun

lemmas cspT_Ext_pre_choice_Rep_dist
     = cspT_Ext_pre_choice_Rep_dist0
       cspT_Ext_pre_choice_Rep_dist_fun

(****************** to add them again ******************)

end
