           (*-------------------------------------------*
            |        CSP-Prover on Isabelle2004         |
            |               February 2005               |
            |                   June 2005  (modified)   |
            |                 August 2005  (modified)   |
            |                                           |
            |        CSP-Prover on Isabelle2005         |
            |                October 2005  (modified)   |
            |                                           |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_T_law_fp = CSP_T_continuous + CSP_T_contraction + CSP_T_mono:

(*****************************************************************

         1. cpo fixed point theory in CSP-Prover
         2. cms fixed point theory in CSP-Prover
         3.
         4. 

 *****************************************************************)

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  Union (B ` A) = (UN x:A. B x)                      *)
(*                  Inter (B ` A) = (INT x:A. B x)                     *)

declare Union_image_eq [simp del]
declare Inter_image_eq [simp del]

(*=======================================================*
 |                                                       |
 |                        CPO                            |
 |                                                       |
 *=======================================================*)

(*-----------*
 |    Bot    |
 *-----------*)

lemma traces_prod_Bot: "(%pn. traces DIV) = Bot"
apply (simp add: prod_Bot_def)
apply (simp add: expand_fun_eq)
apply (simp add: traces_def)
apply (simp add: bottom_domT_def)
done

lemma semT_prod_Bot: "(%pn. [[DIV]]T) = Bot"
by (simp add: semT_def traces_prod_Bot)

(*-----------*
 |   FIX P   |
 *-----------*)

lemma traces_FIX:
    "PF : ProcFun ==>
     traces ((FIX PF) pn)
     = UnionT {u. EX n. u = ((tracesFun PF) ^ n) Bot pn}"
apply (simp add: FIX_def FIXn_def)
apply (subgoal_tac "{u. EX n. u = ((tracesFun PF) ^ n) Bot pn} ~= {}")
apply (rule order_antisym)

 (* <= *)
 apply (rule)
 apply (simp add: in_traces)
 apply (erule disjE, simp)
 apply (elim conjE exE)
 apply (rule_tac x="traces ((PF ^ a) (%pm. DIV) pn)" in exI)
 apply (simp)
 apply (rule_tac x="a" in exI)
 apply (simp add: tracesFun_compo_n_X[THEN sym])
 apply (simp add: iteration_in_ProcFun tracesFun_X_pn)
 apply (simp add: traces_prod_Bot)

 (* => *)
 apply (rule)
 apply (simp add: in_traces)
 apply (rule disjI2)
 apply (elim conjE exE)
 apply (simp)
 apply (rule_tac x="na" in exI)
 apply (simp add: tracesFun_compo_n_X[THEN sym])
 apply (simp add: iteration_in_ProcFun tracesFun_X_pn)
 apply (simp add: traces_prod_Bot)
by (auto)

(*** [[ ]]T ***)

lemma semT_FIX:
    "PF : ProcFun ==>
     [[(FIX PF) pn]]T
     = UnionT {u. EX n. u = ([[PF]]TFun ^ n) Bot pn}"
apply (simp add: semT_to_traces)
apply (simp add: traces_FIX)
done

(*** FIX is LUB ***)

lemma semT_FIX_isLUB:
   "PF : ProcFun ==>
    (%pn. [[(FIX PF) pn]]T) isLUB {x. EX n. x = ([[PF]]TFun ^ n) Bot}"
apply (simp add: prod_LUB_decompo)
apply (intro allI)
apply (simp add: proj_fun_def)
apply (simp add: image_def)
apply (simp add: semT_FIX)
apply (subgoal_tac 
     "{y. EX x. (EX n. x = ([[PF]]TFun ^ n) Bot) & y = x i} =
      {u. EX n. u = ([[PF]]TFun ^ n) Bot i}")
apply (simp)
apply (rule UnionT_isLUB)
apply (auto)
done

lemma semT_FIX_LUB:
   "PF : ProcFun ==>
    (%pn. [[(FIX PF) pn]]T) = LUB {x. EX n. x = ([[PF]]TFun ^ n) Bot}"
apply (rule sym)
apply (rule isLUB_LUB)
apply (simp add: semT_FIX_isLUB)
done

lemma semT_FIX_LFP:
   "PF : ProcFun ==> (%pn. [[(FIX PF) pn]]T) = LFP [[PF]]TFun"
apply (simp add: semT_FIX_LUB)
apply (simp add: Tarski_thm_LFP_LUB continuous_semTFun)
done

lemma semT_FIX_LFP_pn:
   "PF : ProcFun ==> [[(FIX PF) pn]]T = LFP [[PF]]TFun pn"
apply (insert semT_FIX_LFP[of PF])
apply (simp add: expand_fun_eq)
done

lemma semT_FIX_isLFP:
   "PF : ProcFun ==> (%pn. [[(FIX PF) pn]]T) isLFP [[PF]]TFun"
apply (simp add: semT_FIX_LFP)
apply (simp add: LFP_is Tarski_thm_EX continuous_semTFun)
done

lemma semT_FIX_LFP_fixed_point:
   "PF : ProcFun
    ==> [[PF]]TFun (%pn. [[(FIX PF) pn]]T) = (%pn. [[(FIX PF) pn]]T)"
apply (insert semT_FIX_isLFP[of PF], simp)
apply (simp add: isLFP_def)
done

lemma semT_FIX_LFP_least:
   "PF : ProcFun ==> 
    ALL X. [[PF]]TFun X = X --> (%pn. [[(FIX PF) pn]]T) <= X"
apply (insert semT_FIX_isLFP[of PF], simp)
apply (simp add: isLFP_def)
done

(*---------------------------------------------------------------------*
 |   FIX is the greatest fixed point of process-expression-functions   |
 |   (semantically least fixed point)                                  |
 *---------------------------------------------------------------------*)

lemma cspT_FIX_isGFP:
  "PF : ProcFun ==> 
   (FIX PF) ==T (PF (FIX PF)) & (ALL X. X ==T PF X --> X <==T (FIX PF))"
apply (simp add: refT_prod_def eqT_prod_def)
apply (simp add: refT_def eqT_def)
apply (insert semT_FIX_isLFP[of PF], simp)
apply (rule)

 (* FP *)
 apply (simp add: expand_fun_eq[THEN sym])
 apply (simp add: semTFun_X_pn)
 apply (simp add: isLFP_def)

 (* LFP *)
 apply (intro allI impI)
 apply (simp add: expand_fun_eq[THEN sym])
 apply (drule sym)
 apply (simp add: expand_fun_eq)
 apply (simp add: expand_fun_eq[THEN sym])
 apply (simp add: semTFun_X_pn)
 apply (simp add: isLFP_def)
 apply (erule conjE)
 apply (drule_tac x="(%n. [[X n]]T)" in spec)
 apply (simp add: order_prod_def)
done

(*-------------------------------------------------------*
 |                                                       |
 |           Fixpoint unwind (CSP-Prover rule)           |
 |                                                       |
 *-------------------------------------------------------*)

lemma cspT_unwind_lm:
  "PF : ProcFun ==> ALL pn. (FIX PF) pn =T PF (FIX PF) pn"
apply (simp add: eqT_def)
apply (simp add: expand_fun_eq[THEN sym])
apply (simp add: semTFun_X_pn)
apply (simp add: semT_FIX_LFP_fixed_point)
done

(*** csp rule ***)

lemma cspT_unwind:
  "PF : ProcFun ==> (FIX PF) pn =T PF (FIX PF) pn"
by (simp add: cspT_unwind_lm)

(*-------------------------------------------------------*
 |                                                       |
 |        Check fixpoint  (CSP-Prover intro rule)        |
 |                                                       |
 *-------------------------------------------------------*)

lemma cspT_fp_induct_cpo_lm:
   "[| PF : ProcFun ; Q <=T X pn ; ALL pn. X pn <=T (PF X) pn |]
    ==> Q <=T (FIX PF) pn"
apply (simp add: refT_def)
apply (simp add: fold_order_prod_def)
apply (simp add: semTFun_X_pn)
apply (simp add: semT_FIX_LFP_pn)
apply (insert cpo_fixpoint_induction_rev
       [of "[[PF]]TFun" "(%pn. [[X pn]]T)"])
apply (simp add: continuous_semTFun)
apply (simp add: order_prod_def)
apply (rotate_tac -1)
apply (drule_tac x="pn" in spec)
apply (auto)
done

(*** csp rule ***)

lemma cspT_fp_induct_cpo:
   "[| PF : ProcFun ; Q <=T X pn ; !!pn. X pn <=T (PF X) pn |]
    ==> Q <=T (FIX PF) pn"
by (simp add: cspT_fp_induct_cpo_lm)

(*=======================================================*
 |                                                       |
 |                        CMS                            |
 |                                                       |
 *=======================================================*)

lemma semT_FIX_UFP:
   "[| PF : ProcFun ; PF : nohideFun ; PF : guardFun |] ==>
    (%pn. [[(FIX PF) pn]]T) = UFP [[PF]]TFun"
apply (subgoal_tac "[[PF]]TFun hasUFP")
apply (insert semT_FIX_isLFP[of PF], simp)
apply (subgoal_tac "UFP [[PF]]TFun isUFP [[PF]]TFun")
apply (simp add: isUFP_def)
apply (simp add: isLFP_def)
apply (elim conjE exE)
apply (rotate_tac -1)
apply (drule_tac x="(%pn. [[(FIX PF) pn]]T)" in spec)
apply (simp)
apply (simp add: UFP_is)
apply (simp add: contraction_semTFun Banach_thm_EX)
done

lemma traces_FIX_UFP:
   "[| PF : ProcFun ; PF : nohideFun ; PF : guardFun |] ==>
    (%pn. traces((FIX PF) pn)) = UFP (tracesFun PF)"
apply (insert semT_FIX_UFP[of PF])
apply (simp add: semT_to_traces)
done

lemma semT_FIX_UFP_pn:
   "[| PF : ProcFun ; PF : nohideFun ; PF : guardFun |] ==>
    [[(FIX PF) pn]]T = UFP [[PF]]TFun pn"
apply (insert semT_FIX_UFP[of PF])
apply (simp add: expand_fun_eq)
done

lemma semT_FIX_isUFP:
   "[| PF : ProcFun ; PF : nohideFun ; PF : guardFun |] ==>
    (%pn. [[(FIX PF) pn]]T) isUFP [[PF]]TFun"
apply (simp add: semT_FIX_UFP)
apply (simp add: contraction_semTFun Banach_thm_EX UFP_is)
done

lemma semT_FIX_UFP_unique:
   "[| PF : ProcFun ; PF : nohideFun ; PF : guardFun |] ==>
    ALL X. [[PF]]TFun X = X --> (%pn. [[(FIX PF) pn]]T) = X"
apply (insert semT_FIX_isUFP[of PF], simp)
apply (simp add: isUFP_def)
done

(*------------------------------------------------------------------*
 |   FIX is the unique fixed point of process-expression-functions  |
 *------------------------------------------------------------------*)

lemma cspT_FIX_isUFP:
   "[| PF : ProcFun ; PF : nohideFun ; PF : guardFun |] ==>
   (FIX PF) ==T (PF (FIX PF)) & (ALL X. X ==T PF X --> X ==T (FIX PF))"
apply (simp add: cspT_FIX_isGFP)

apply (simp add: eqT_prod_def)
apply (simp add: eqT_def)
apply (insert semT_FIX_isUFP[of PF], simp)

 (* UFP *)
 apply (intro allI impI)
 apply (simp add: expand_fun_eq[THEN sym])
 apply (drule sym)
 apply (simp add: expand_fun_eq)
 apply (simp add: expand_fun_eq[THEN sym])
 apply (simp add: semTFun_X_pn)
 apply (simp add: isUFP_def)
 apply (erule conjE)
 apply (drule_tac x="(%pn. [[X pn]]T)" in spec)
 apply (simp add: expand_fun_eq)
done

(*-------------------------------------------------------*
 |                                                       |
 |        Check fixpoint  (CSP-Prover intro rule)        |
 |                                                       |
 *-------------------------------------------------------*)

(* refinement *)

lemma cspT_fp_induct_cms_ref_lm:
   "[| PF : ProcFun ; PF : nohideFun ; PF : guardFun  ;
       X pn <=T Q ; ALL pn. (PF X) pn <=T X pn |]
    ==> (FIX PF) pn <=T Q"
apply (simp add: refT_def)
apply (simp add: fold_order_prod_def)
apply (simp add: semTFun_X_pn)

apply (simp add: semT_FIX_UFP_pn)
apply (insert cms_fixpoint_induction_ref
       [of "[[PF]]TFun" "(%pn. [[X pn]]T)" "UFP [[PF]]TFun"])
apply (simp add: contra_alpha_to_contst contraction_alpha_semTFun)
apply (simp add: mono_semTFun)
apply (simp add: contraction_semTFun UFP_fp Banach_thm_EX)
apply (simp add: order_prod_def)
apply (rotate_tac -1)
apply (drule_tac x="pn" in spec)
apply (simp)
done

(*** csp rule ***)

lemma cspT_fp_induct_cms_ref:
   "[| PF : ProcFun ; PF : nohideFun ; PF : guardFun  ;
       X pn <=T Q ; !! pn. (PF X) pn <=T X pn |]
    ==> (FIX PF) pn <=T Q"
by (simp add: cspT_fp_induct_cms_ref_lm)

(* equivalence *)

lemma cspT_fp_induct_cms_eq_lm:
   "[| PF : ProcFun ; PF : nohideFun ; PF : guardFun  ;
       X pn =T Q ; ALL pn. (PF X) pn =T X pn |]
    ==> (FIX PF) pn =T Q"
apply (simp add: eqT_def)
apply (simp add: expand_fun_eq[THEN sym])
apply (simp add: semTFun_X_pn)
apply (insert semT_FIX_isUFP[of PF], simp)
apply (simp add: isUFP_def)
apply (elim conjE)
apply (drule_tac x="(%pn. [[X pn]]T)" in spec)
apply (simp add: expand_fun_eq)
done

(*** csp rule ***)

lemma cspT_fp_induct_cms_eq:
   "[| PF : ProcFun ; PF : nohideFun ; PF : guardFun ;
       X pn =T Q ; !! pn. (PF X) pn =T X pn |]
    ==> (FIX PF) pn =T Q"
by (simp add: cspT_fp_induct_cms_eq_lm)

lemmas cspT_fp_induct_cms = cspT_fp_induct_cms_ref cspT_fp_induct_cms_eq

(*-------------------------------------------------------*
 |                                                       |
 |         FIX expansion (CSP-Prover intro rule)         |
 |                                                       |
 *-------------------------------------------------------*)

lemma traces_FIXn_plus_sub_lm:
 "PF : ProcFun ==>
  ALL n m pn. traces ((FIX[n] PF) pn) <= traces ((FIX[n+m] PF) pn)"
apply (rule)
apply (induct_tac n)
apply (simp add: FIXn_def)
apply (intro allI)
apply (rule)
apply (simp add: in_traces)

apply (rule allI)
apply (simp add: FIXn_def)
apply (drule_tac x="m" in spec)
apply (simp add: tracesFun_X_pn)
apply (fold order_prod_def)
apply (simp add: mono_tracesFun[simplified mono_def])
done

lemma traces_FIXn_plus_sub: 
   "PF : ProcFun ==> traces ((FIX[n] PF) pn) <= traces ((FIX[n+m] PF) pn)"
by (simp add: traces_FIXn_plus_sub_lm)

lemma semT_FIXn_plus_sub:
 "PF : ProcFun ==> [[(FIX[n] PF) pn]]T <= [[(FIX[n+m] PF) pn]]T"
apply (simp add: semT_to_traces)
apply (simp add: traces_FIXn_plus_sub)
done

lemma in_traces_FIXn_plus_sub: 
  "[| PF : ProcFun ; t :t traces ((FIX[n] PF) pn) |] ==> t :t traces ((FIX[n+m] PF) pn)"
apply (insert traces_FIXn_plus_sub[of PF n pn m])
by (auto)

(*-----------------------------------------------------*
 |  sometimes FIX[n + f n] is useful more than FIX[n]  |
 *-----------------------------------------------------*)

lemma cspT_FIX_plus_eq: 
     "PF : ProcFun ==>
      ALL f pn. (FIX PF) pn =T (!nat n .. ((FIX[n + f n] PF) pn))"
apply (simp add: FIX_def)
apply (simp add: eqT_def)
apply (intro allI)
apply (simp add: semT_def)
apply (rule order_antisym)

 apply (rule)
 apply (simp add: in_traces)
 apply (rule disjI2)
 apply (elim conjE exE disjE, simp)
 apply (rule_tac x="a" in exI)
 apply (simp add: in_traces_FIXn_plus_sub)

 apply (rule)
 apply (simp add: in_traces)
 apply (elim conjE exE disjE)
 apply (simp)
 apply (rule disjI2)
 apply (rule_tac x="a + f a" in exI)
 apply (simp)
done

lemma cspT_FIX_plus_eq_fun: 
     "PF : ProcFun ==>
      ALL f pn. (%pn. (FIX PF) pn) ==T (%pn. (!nat n .. ((FIX[n + f n] PF) pn)))"
apply (simp add: eqT_prod_def)
apply (simp add: cspT_FIX_plus_eq)
done

lemma cspT_FIX_eq: 
     "PF : ProcFun ==>
      ALL pn. (FIX PF) pn =T (!nat n .. ((FIX[n] PF) pn))"
apply (insert cspT_FIX_plus_eq[of PF])
apply (simp)
apply (drule_tac x="(%n. (0::nat))" in spec)
apply (simp)
done

lemma cspT_FIX_eq_fun: 
     "PF : ProcFun ==>
      ALL pn. (%pn. (FIX PF) pn) ==T (%pn. (!nat n .. ((FIX[n] PF) pn)))"
apply (insert cspT_FIX_plus_eq_fun[of PF])
apply (simp)
apply (drule_tac x="(%n. (0::nat))" in spec)
apply (simp)
done

(*=======================================================*
 |                                                       |
 |                Single recursion MU                    |
 |                                                       |
 *=======================================================*)

(*=======================================================*
 |                         cpo                           |
 *=======================================================*)

(*-------------------------------------------------------*
 |            unwinding (CSP-Prover intro rule)          |
 *-------------------------------------------------------*)

lemma cspT_unwind_MU:
  "PX : ProcX ==> MUX PX =T PX (MUX PX)"
apply (simp add: MUX_def)
apply (insert cspT_unwind[of "(%x n. PX (x MUpn))" "MUpn"])
apply (simp add: ProcX_def)
apply (simp add: ProcFun_def)
done

(*-------------------------------------------------------*
 |        Check fixpoint  (CSP-Prover intro rule)        |
 *-------------------------------------------------------*)

lemma cspT_fp_induct_MU_cpo:
   "[| PX : ProcX ; Q <=T PX Q |] ==> Q <=T (MUX PX)"
apply (simp add: MUX_def)
apply (rule cspT_fp_induct_cpo[of _ _ "(%X. Q)"])
apply (simp add: ProcX_def)
apply (simp add: ProcFun_def)
apply (simp)
apply (simp)
done

(*=======================================================*
 |                         cms                           |
 *=======================================================*)

(*-------------------------------------------------------*
 |        Check fixpoint  (CSP-Prover intro rule)        |
 *-------------------------------------------------------*)

lemma cspT_fp_induct_MU_cms_ref:
   "[| PX : ProcX ; PX : nohideX ; PX : guardX ; PX Q <=T Q |]
    ==> (MUX PX) <=T Q"
apply (simp add: MUX_def)
apply (rule cspT_fp_induct_cms_ref[of _ "(%X. Q)"])
apply (simp add: ProcX_def)
apply (simp add: ProcFun_def)
apply (simp add: nohideX_def)
apply (simp add: nohideFun_def)
apply (simp add: guardX_def)
apply (simp add: guardFun_def)
apply (simp)
apply (simp)
done

lemma cspT_fp_induct_MU_cms_eq:
   "[| PX : ProcX ; PX : nohideX ; PX : guardX ; PX Q =T Q |]
    ==> (MUX PX) =T Q"
apply (simp add: MUX_def)
apply (rule cspT_fp_induct_cms_eq[of _ "(%X. Q)"])
apply (simp add: ProcX_def)
apply (simp add: ProcFun_def)
apply (simp add: nohideX_def)
apply (simp add: nohideFun_def)
apply (simp add: guardX_def)
apply (simp add: guardFun_def)
apply (simp)
apply (simp)
done

lemmas cspT_fp_induct_MU_cms = cspT_fp_induct_MU_cms_ref cspT_fp_induct_MU_cms_eq

(*=======================================================*
 |                   expand by Tariski                   |
 *=======================================================*)

lemma cspT_FIX_MU_plus_eq_lm: 
     "((%x pn. (PX::'a proc => 'a proc) (x MUpn)) ^ n) (%pm. DIV) MUpn = (PX ^ n) DIV"
apply (induct_tac n)
apply (simp_all)
done

lemma cspT_FIX_MU_plus_eq: 
     "PX : ProcX ==>
      ALL f. (MUX PX) =T (!nat n .. (PX ^ (n + f n)) DIV)"
apply (insert cspT_FIX_plus_eq[of "(%x pn. PX (x MUpn))"])
apply (simp add: ProcX_def)
apply (simp add: ProcFun_def)
apply (rule allI)
apply (drule_tac x="f" in spec)
apply (simp add: MUX_def)
apply (drule_tac x="MUpn" in spec)
apply (simp add: FIXn_def)
apply (simp add: cspT_FIX_MU_plus_eq_lm)
done

lemma cspT_FIX_MU_eq: 
     "PX : ProcX ==>
      (MUX PX) =T (!nat n .. (PX ^ n) DIV)"
apply (insert cspT_FIX_MU_plus_eq[of PX])
apply (simp)
apply (drule_tac x="(%n. (0::nat))" in spec)
apply (simp)
done

(****************** to add them again ******************)

declare Union_image_eq [simp]
declare Inter_image_eq [simp]

end
