           (*-------------------------------------------*
            |        CSP-Prover on Isabelle2004         |
            |               December 2004               |
            |                   July 2005  (modified)   |
            |                                           |
            |        CSP-Prover on Isabelle2005         |
            |                October 2005  (modified)   |
            |                                           |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_T_semantics = CSP_syntax + Domain_T
                       + Trace_par + Trace_hide + Trace_ren + Trace_seq:

(*****************************************************************

         1. semantic clause
         2. 
         3. 
         4. 

 *****************************************************************)

(*********************************************************
                    semantic clause
 *********************************************************)

consts
  traces  :: "'a proc => 'a domT"

primrec
  "traces(STOP)        = {[]t}t"
  "traces(SKIP)        = {[]t, [Tick]t}t"
  "traces(DIV)         = {[]t}t"
  "traces(a -> P)      = {t. t = []t | (EX s. t = [Ev a]t @t s & s :t traces(P)) }t"

  "traces(? :X -> Pf)  = {t. t = []t | 
                            (EX a s. t = [Ev a]t @t s & s :t traces(Pf a) & a : X) }t"

  "traces(P [+] Q)     = traces(P) UnT traces(Q)"
  "traces(P |~| Q)     = traces(P) UnT traces(Q)"
  "traces(!! :X .. Pf) = {t. t = []t | (EX a:X. t :t traces(Pf a)) }t"
  "traces(IF b THEN P ELSE Q) = (if b then traces(P) else traces(Q))"
  "traces(P |[X]| Q)   = {u. EX s t. u : s |[X]|tr t & s :t traces(P) & t :t traces(Q) }t"
  "traces(P -- X)      = {t. EX s. t = s --tr X & s :t traces(P) }t"
  "traces(P [[r]])     = {t. EX s. s [[r]]tr t & s :t traces(P) }t"
  "traces(P ;; Q)      = {u. (EX s. u = rmTick s & s :t traces(P) ) |
                             (EX s t. u = s @t t & s @t [Tick]t :t traces(P) &
                                      t :t traces(Q) & noTick s) }t"

declare traces.simps [simp del]

lemma traces_Rep_int_choice_fun: 
  "inj f ==> traces(!!<f> :X .. Pf) = {t. t = []t | (EX a:X. t :t traces(Pf a)) }t"
apply (simp add: Rep_int_choice_fun_def)
apply (simp add: traces.simps)
done

lemmas traces_def = traces.simps traces_Rep_int_choice_fun

lemma traces_Int_choice_Ext_choice: "traces(P |~| Q) = traces(P [+] Q)"
by (simp add: traces_def)

(*********************************************************
                     semantics
 *********************************************************)

consts
  semT    :: "'a proc => 'a domT"      ("[[_]]T")

defs
  semT_def : "[[P]]T == traces(P)"

(*********************************************************
              relations over processes
 *********************************************************)

consts
  refT :: "'a proc => 'a proc => bool"
                                           ("(0_ /<=T /_)" [50,50] 50)
  eqT  :: "'a proc => 'a proc => bool"
                                           ("(0_ /=T /_)" [50,50] 50)
  refT_prod :: "('n => 'a proc) => ('n => 'a proc) => bool"
                                           ("(0_ /<==T /_)" [50,50] 50)
  eqT_prod  :: "('n => 'a proc) => ('n => 'a proc) => bool"
                                           ("(0_ /==T /_)" [50,50] 50)

defs
  refT_def : "P1 <=T P2 == [[P2]]T <= [[P1]]T"
  eqT_def  : "P1  =T P2 == [[P1]]T  = [[P2]]T"
  refT_prod_def : "Pp1 <==T Pp2 == (ALL n. (Pp1 n) <=T (Pp2 n))"
  eqT_prod_def  : "Pp1  ==T Pp2 == (ALL n. (Pp1 n)  =T (Pp2 n))"

(*------------------------------------*
 |              X-Symbols             |
 *------------------------------------*)

syntax (output)
  "_refTX"    :: "'a proc => 'a proc => bool"
                   ("(0_ /<=T /_)" [50,50] 50)

syntax (xsymbols)
  "_refTX"    :: "'a proc => 'a proc => bool"
                   ("(0_ /\<sqsubseteq>T /_)" [50,50] 50)

translations
  "P1 \<sqsubseteq>T P2" == "P1 <=T P2"

(*------------------*
 |      csp law     |
 *------------------*)

(*** semantics ***)

lemma cspT_eqT_semantics:
  "(P =T Q) = (traces P = traces Q)"
by (simp add: eqT_def semT_def)

lemma cspT_refT_semantics:
  "(P <=T Q) = (traces Q <= traces P)"
by (simp add: refT_def semT_def)

lemma cspT_eqT_prod_semantics:
  "(f ==T g) = 
   (ALL x. traces (f x) = traces(g x))"
apply (simp add: eqT_prod_def)
apply (simp add: cspT_eqT_semantics)
done

lemma cspT_refT_prod_semantics:
  "(f <==T g) = 
   (ALL x. traces (g x) <= traces(f x))"
apply (simp add: refT_prod_def)
apply (simp add: cspT_refT_semantics)
done

lemmas cspT_semantics = cspT_eqT_semantics  cspT_refT_semantics
                        cspT_eqT_prod_semantics cspT_refT_prod_semantics

(*** eq and ref ***)

lemma cspT_eq_ref_iff:
  "(P1 =T P2) = (P1 <=T P2 & P2 <=T P1)"
by (auto simp add: refT_def eqT_def intro:order_antisym)

lemma cspT_eq_ref:
  "P1 =T P2 ==> P1 <=T P2"
by (simp add: cspT_eq_ref_iff)

lemma cspT_ref_eq:
  "[| P1 <=T P2 ; P2 <=T P1 |] ==> P1 =T P2"
by (simp add: cspT_eq_ref_iff)

(*** reflexivity ***)

lemma cspT_reflex_eq[simp]: "P =T P"
by (simp add: eqT_def)

lemma cspT_reflex_ref[simp]: "P <=T P"
by (simp add: refT_def)

lemmas cspT_reflex = cspT_reflex_eq cspT_reflex_ref

(*** symmetry ***)

lemma cspT_sym: "P1 =T P2 ==> P2 =T P1"
by (simp add: eqT_def)

lemma cspT_symE:
  "[| P1 =T P2 ; P2 =T P1 ==> Z |] ==> Z"
by (simp add: eqT_def)

(*** transitivity ***)

lemma cspT_trans_left_eq: "[| P1 =T P2 ; P2 =T P3 |] ==> P1 =T P3"
by (simp add: eqT_def)

lemma cspT_trans_left_ref: "[| P1 <=T P2 ; P2 <=T P3 |] ==> P1 <=T P3"
by (simp add: refT_def)

lemmas cspT_trans_left = cspT_trans_left_eq cspT_trans_left_ref
lemmas cspT_trans = cspT_trans_left

lemma cspT_trans_right_eq: "[| P2 =T P3 ; P1 =T P2 |] ==> P1 =T P3"
by (simp add: eqT_def)

lemma cspT_trans_right_ref: "[| P2 <=T P3 ;  P1 <=T P2 |] ==> P1 <=T P3"
by (simp add: refT_def)

lemmas cspT_trans_rught = cspT_trans_right_eq cspT_trans_right_ref

(*** rewrite (eq) ***)

lemma cspT_rw_left_eq:
  "[| P1 =T P2 ; P2 =T P3 |] ==> P1 =T P3"
by (simp add: eqT_def)

lemma cspT_rw_left_ref:
  "[| P1 =T P2 ; P2 <=T P3 |] ==> P1 <=T P3"
by (simp add: refT_def eqT_def)

lemmas cspT_rw_left = cspT_rw_left_eq cspT_rw_left_ref

lemma cspT_rw_right_eq:
  "[| P3 =T P2 ; P1 =T P2 |] ==> P1 =T P3"
by (simp add: eqT_def)

lemma cspT_rw_right_ref:
  "[| P3 =T P2 ; P1 <=T P2 |] ==> P1 <=T P3"
by (simp add: refT_def eqT_def)

lemmas cspT_rw_right = cspT_rw_right_eq cspT_rw_right_ref

(*** rewrite (ref) ***)

lemma cspT_tr_left_eq: "[| P1 =T P2 ; P2 =T P3 |] ==> P1 =T P3"
by (simp add: eqT_def)

lemma cspT_tr_left_ref: "[| P1 <=T P2 ; P2 <=T P3 |] ==> P1 <=T P3"
by (simp add: refT_def eqT_def)

lemmas cspT_tr_left = cspT_tr_left_eq cspT_tr_left_ref

lemma cspT_tr_right_eq: "[| P2 =T P3 ; P1 =T P2 |] ==> P1 =T P3"
by (simp add: eqT_def)

lemma cspT_tr_right_ref: "[| P2 <=T P3 ; P1 <=T P2 |] ==> P1 <=T P3"
by (simp add: refT_def eqT_def)

lemmas cspT_tr_right = cspT_tr_right_eq cspT_tr_right_ref

end
