           (*-------------------------------------------*
            |        CSP-Prover on Isabelle2004         |
            |               December 2004               |
            |                   July 2005  (modified)   |
            |                 August 2005  (modified)   |
            |                                           |
            |        CSP-Prover on Isabelle2005         |
            |                October 2005  (modified)   |
            |                                           |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_T_tracesfun = CSP_T_full_abs:

(*****************************************************************

         1. semantics for process functions
         2. 
         3. 
         4. 

 *****************************************************************)

(*********************************************************
                      semantics
 *********************************************************)

consts
  tracesfun  :: "('pn,'a) procfun => (('pn => 'a domT) => 'a domT)"
  tracesFun  :: "('pn,'a) procFun => (('pn => 'a domT) => ('pn => 'a domT))"

  semTfun    :: "('pn,'a) procfun => (('pn => 'a domT) => 'a domT)"  ("[[_]]Tfun")
  semTFun    :: "('pn,'a) procFun => 
                 (('pn => 'a domT) => ('pn => 'a domT))"  ("[[_]]TFun")

defs
  tracesfun_def : "tracesfun Pf == (%X. traces (Pf (%pn. inv[[X pn]]T)))"
  tracesFun_def : "tracesFun PF == (%X. (%pn. tracesfun (%X. (PF X) pn) X))"

  semTfun_def : "[[Pf]]Tfun == (%X. [[Pf (%pn. inv[[X pn]]T)]]T)"
  semTFun_def : "[[PF]]TFun == (%X. (%pn. [[(%X. (PF X) pn)]]Tfun X))"

(*********************************************************
              semantics for each operators
 *********************************************************)

lemma tracesfun_pn: 
  "tracesfun (%x. x pn) = (%x. (x pn))"
apply (simp add: tracesfun_def)
apply (simp add: traces_inv_semT)
done

lemma tracesfun_STOP: 
  "tracesfun (%x. STOP) = (%x. {[]t}t)"
by (simp add: tracesfun_def traces_def)

lemma tracesfun_SKIP: 
  "tracesfun (%x. SKIP) = (%x. {[]t, [Tick]t}t)"
by (simp add: tracesfun_def traces_def)

lemma tracesfun_DIV: 
  "tracesfun (%x. DIV) = (%x. {[]t}t)"
by (simp add: tracesfun_def traces_def)

lemma tracesfun_Act_prefix:
  "tracesfun (%x. a -> (Pf x)) = 
  (%x. {t. t = []t | (EX s. t = [Ev a]t @t s & s :t tracesfun Pf x) }t)"
by (simp add: tracesfun_def traces_def)

lemma tracesfun_Ext_pre_choice:
  "tracesfun (%x. ? a:X -> (Pff a x)) = 
  (%x. {t. t = []t | 
       (EX a s. t = [Ev a]t @t s & s :t tracesfun (Pff a) x & a : X) }t)"
by (simp add: tracesfun_def traces_def)

lemma tracesfun_Ext_choice:
  "tracesfun (%x. (Pf x) [+] (Qf x)) = (%x. tracesfun Pf x UnT tracesfun Qf x)"
by (simp add: tracesfun_def traces_def)

lemma tracesfun_Int_choice:
  "tracesfun (%x. (Pf x) |~| (Qf x)) = (%x. tracesfun Pf x UnT tracesfun Qf x)"
by (simp add: tracesfun_def traces_def)

lemma tracesfun_Rep_int_choice_Tri:
  "tracesfun (%x. !! a:X .. (Pff a x)) = 
  (%x. {t. t = []t | (EX a:X. t :t tracesfun (Pff a) x) }t)"
by (simp add: tracesfun_def traces_def)

lemma tracesfun_Rep_int_choice_fun:
  "inj f ==> tracesfun (%x. !!<f> a:X .. (Pff a x)) = 
  (%x. {t. t = []t | (EX a:X. t :t tracesfun (Pff a) x) }t)"
by (simp add: tracesfun_def traces_def)

lemmas tracesfun_Rep_int_choice = tracesfun_Rep_int_choice_Tri
                                  tracesfun_Rep_int_choice_fun

lemma tracesfun_Conditional:
  "tracesfun (%x. IF b THEN (Pf x) ELSE (Qf x))
   = (%x. if b then tracesfun Pf x else tracesfun Qf x)"
apply (case_tac "b")
by (simp_all add: tracesfun_def traces_def)

lemma tracesfun_Parallel:
  "tracesfun (%x. (Pf x) |[X]| (Qf x))
   = (%x. {u. EX s t. u : s |[X]|tr t & s :t tracesfun Pf x & t :t tracesfun Qf x }t)"
by (simp add: tracesfun_def traces_def)

lemma tracesfun_Hiding:
  "tracesfun (%x. (Pf x) -- X)
   = (%x. {t. EX s. t = s --tr X & s :t tracesfun Pf x}t)"
by (simp add: tracesfun_def traces_def)

lemma tracesfun_Renaming:
  "tracesfun (%x. (Pf x) [[r]])
   = (%x. {t. EX s. s [[r]]tr t & s :t tracesfun Pf x}t)"
by (simp add: tracesfun_def traces_def)

lemma tracesfun_Seq_compo:
  "tracesfun (%x. (Pf x) ;; (Qf x))
   = (%x. {u. (EX s. u = rmTick s & s :t tracesfun Pf x ) |
              (EX s t. u = s @t t & s @t [Tick]t :t tracesfun Pf x &
                       t :t tracesfun Qf x & noTick s) }t)"
by (simp add: tracesfun_def traces_def)

lemmas tracesfun_simp =
  tracesfun_pn tracesfun_STOP tracesfun_SKIP tracesfun_DIV
  tracesfun_Act_prefix tracesfun_Ext_pre_choice 
  tracesfun_Ext_choice tracesfun_Int_choice
  tracesfun_Rep_int_choice tracesfun_Conditional
  tracesfun_Parallel tracesfun_Hiding
  tracesfun_Renaming tracesfun_Seq_compo

lemma tracesfun_Int_choice_Ext_choice: 
  "tracesfun (%x. (Pf x) |~| (Qf x)) = tracesfun (%x. (Pf x) [+] (Qf x))"
by (simp add: tracesfun_simp)

(*********************************************************
                   in tracesfun
 *********************************************************)

lemma in_tracesfun_pn: 
  "(t :t tracesfun (%x. x pn) x) = (t :t (x pn))"
apply (simp add: tracesfun_def)
apply (simp add: traces_inv_semT)
done

lemma in_tracesfun_STOP: 
  "(t :t tracesfun (%x. STOP) x) = (t = []t)"
by (simp add: tracesfun_def in_traces)

lemma in_tracesfun_SKIP: 
  "(t :t tracesfun (%x. SKIP) x) = (t = []t | t = [Tick]t)"
by (simp add: tracesfun_def in_traces)

lemma in_tracesfun_DIV: 
  "(t :t tracesfun (%x. DIV) x) = (t = []t)"
by (simp add: tracesfun_def in_traces)

lemma in_tracesfun_Act_prefix: 
  "(t :t tracesfun (%x. a -> (Pf x)) x) = 
   (t = []t | (EX s. t = [Ev a]t @t s & s :t tracesfun Pf x))"
by (simp add: tracesfun_def in_traces)

lemma in_tracesfun_Ext_pre_choice:
  "(t :t tracesfun (%x. ? a:X -> (Pff a x)) x) = 
   (t = []t | 
    (EX a s. t = [Ev a]t @t s & s :t tracesfun (Pff a) x & a : X))"
by (simp add: tracesfun_def in_traces)

lemma in_tracesfun_Ext_choice:
  "(t :t tracesfun (%x. (Pf x) [+] (Qf x)) x) = 
   (t :t tracesfun Pf x | t:t tracesfun Qf x)"
by (simp add: tracesfun_def in_traces)

lemma in_tracesfun_Int_choice:
  "(t :t tracesfun (%x. (Pf x) |~| (Qf x)) x) = 
   (t :t tracesfun Pf x | t:t tracesfun Qf x)"
by (simp add: tracesfun_def in_traces)

lemma in_tracesfun_Rep_int_choice_Tri:
  "(t :t tracesfun (%x. !! a:X .. (Pff a x)) x) = 
   (t = []t | (EX a:X. t :t tracesfun (Pff a) x))"
by (simp add: tracesfun_def in_traces)

lemma in_tracesfun_Rep_int_choice_fun:
  "inj f ==> (t :t tracesfun (%x. !!<f> a:X .. (Pff a x)) x) = 
   (t = []t | (EX a:X. t :t tracesfun (Pff a) x))"
by (simp add: tracesfun_def in_traces)

lemmas in_tracesfun_Rep_int_choice = in_tracesfun_Rep_int_choice_Tri
                                     in_tracesfun_Rep_int_choice_fun

lemma in_tracesfun_Conditional:
  "(t :t tracesfun (%x. IF b THEN (Pf x) ELSE (Qf x)) x) =
   (if b then t :t tracesfun Pf x else t :t tracesfun Qf x)"
apply (case_tac "b")
by (simp_all add: tracesfun_def traces_def)

lemma in_tracesfun_Parallel:
  "(u :t tracesfun (%x. (Pf x) |[X]| (Qf x)) x) =
   (EX s t. u : s |[X]|tr t & s :t tracesfun Pf x & t :t tracesfun Qf x)"
by (simp add: tracesfun_def in_traces)

lemma in_tracesfun_Hiding:
  "(t :t tracesfun (%x. (Pf x) -- X) x) =
   (EX s. t = s --tr X & s :t tracesfun Pf x)"
by (simp add: tracesfun_def in_traces)

lemma in_tracesfun_Renaming:
  "(t :t tracesfun (%x. (Pf x) [[r]]) x) =
   (EX s. s [[r]]tr t & s :t tracesfun Pf x)"
by (simp add: tracesfun_def in_traces)

lemma in_tracesfun_Seq_compo:
  "(u :t tracesfun (%x. (Pf x) ;; (Qf x)) x) =
   ((EX s. u = rmTick s & s :t tracesfun Pf x ) |
    (EX s t. u = s @t t & s @t [Tick]t :t tracesfun Pf x &
                      t :t tracesfun Qf x & noTick s))"
by (simp add: tracesfun_def in_traces)

lemmas in_tracesfun =
  in_tracesfun_pn in_tracesfun_STOP in_tracesfun_SKIP in_tracesfun_DIV
  in_tracesfun_Act_prefix in_tracesfun_Ext_pre_choice 
  in_tracesfun_Ext_choice in_tracesfun_Int_choice
  in_tracesfun_Rep_int_choice in_tracesfun_Conditional
  in_tracesfun_Parallel in_tracesfun_Hiding
  in_tracesfun_Renaming in_tracesfun_Seq_compo

(**************************************************
     tracesfun, tracesFun, [[P]]Tfun, [[P]]TFun
 **************************************************)

lemma semTfun_tracesfun: "Pf : Procfun ==> [[Pf]]Tfun = tracesfun Pf"
apply (simp add: tracesfun_def semTfun_def)
apply (rule Procfun.induct[of Pf])
apply (simp_all add: traces_def semT_def)
done

lemma semTFun_tracesFun: "PF : ProcFun ==> [[PF]]TFun = tracesFun PF"
apply (simp add: tracesFun_def semTFun_def)
apply (simp add: ProcFun_def)
apply (simp add: semTfun_tracesfun)
done

lemmas semT_to_traces = semT_def semTfun_tracesfun semTFun_tracesFun

lemma proj_semTFun_semTfun:
 "(proj_fun pn o [[PF]]TFun) = [[(%X. PF X pn)]]Tfun"
apply (simp add: semTFun_def)
apply (simp add: proj_fun_def)
apply (simp add: expand_fun_eq)
done

lemma proj_tracesFun_tracesfun:
 "(proj_fun pn o tracesFun PF) = tracesfun (%X. PF X pn)"
apply (simp add: tracesFun_def)
apply (simp add: proj_fun_def)
apply (simp add: expand_fun_eq)
done

(**************************************************
        traces tracesfun tracesFun
 **************************************************)

lemma tracesfun_X:
  "Pf : Procfun ==> traces (Pf X) = tracesfun (Pf) (%pn. traces(X pn))"
apply (simp add: tracesfun_def)
apply (rule Procfun.induct[of Pf])
apply (simp_all add: traces_def)
apply (simp add: traces_inv_semT)
done

lemma tracesFun_X_pn:
  "PF : ProcFun ==> traces(PF X pn) = tracesFun(PF) (%pn. traces(X pn)) pn"
apply (simp add: tracesFun_def)
apply (simp add: ProcFun_def)
apply (simp add: tracesfun_X[THEN sym])
done

lemma tracesfun_compo_tracesFun_X:
   "[| Pf : Procfun ; QF : ProcFun |] ==>
    tracesfun(Pf o QF) X = tracesfun(Pf) (tracesFun(QF) X)"
apply (simp add: tracesfun_def)
apply (simp add: tracesfun_X)
apply (simp add: tracesFun_X_pn)
apply (simp add: traces_inv_semT)
done

lemma tracesfun_compo_tracesFun_fun:
   "[| Pf : Procfun ; QF : ProcFun |] ==>
    tracesfun(Pf o QF) = tracesfun(Pf) o tracesFun(QF)"
apply (simp add: expand_fun_eq)
apply (simp add: tracesfun_compo_tracesFun_X)
done

lemmas tracesfun_compo_tracesFun = tracesfun_compo_tracesFun_X 
                                   tracesfun_compo_tracesFun_fun 
                                   tracesfun_compo_tracesFun_X[simplified comp_def]
                                   tracesfun_compo_tracesFun_fun[simplified comp_def]

lemma tracesFun_compo_tracesFun_X:
   "[| PF : ProcFun ; QF : ProcFun |] ==>
    tracesFun(PF o QF) X = tracesFun(PF) (tracesFun(QF) X)"
apply (simp add: tracesFun_def)
apply (simp add: tracesfun_def)
apply (simp add: tracesFun_X_pn)
apply (simp add: traces_inv_semT)
done

(*** composition of process functions ***)

lemma tracesFun_compo_tracesFun_fun:
   "[| PF : ProcFun ; QF : ProcFun |] ==>
    tracesFun(PF o QF) = tracesFun(PF) o tracesFun(QF)"
apply (simp add: expand_fun_eq)
apply (simp add: tracesFun_compo_tracesFun_X)
done

lemmas tracesFun_compo_tracesFun = tracesFun_compo_tracesFun_X 
                                   tracesFun_compo_tracesFun_fun
                                   tracesFun_compo_tracesFun_X[simplified comp_def]
                                   tracesFun_compo_tracesFun_fun[simplified comp_def]

lemma tracesFun_compo_n_X:
   "PF : ProcFun ==> tracesFun(PF ^ n) X = (tracesFun(PF) ^ n) X"
apply (induct_tac n)

apply (simp add: tracesFun_def)
apply (simp add: tracesfun_def)
apply (simp add: traces_inv_semT)
apply (simp add: iteration_in_ProcFun tracesFun_compo_tracesFun)
done

(**************************************************
           [[P]]T, [[P]]Tfun, [[P]]TFun
 **************************************************)

lemma semTfun_X:
  "Pf : Procfun ==> [[Pf X]]T = [[Pf]]Tfun (%pn. [[X pn]]T)"
by (simp add: semT_to_traces tracesfun_X)

lemma semTFun_X_pn:
  "PF : ProcFun ==> [[PF X pn]]T = [[PF]]TFun (%pn. [[X pn]]T) pn"
by (simp add: semT_to_traces tracesFun_X_pn)

lemma semTfun_compo_semTFun_X:
   "[| Pf : Procfun ; QF : ProcFun |] ==>
    [[Pf o QF]]Tfun X = [[Pf]]Tfun ([[QF]]TFun X)"
by (simp add: semT_to_traces tracesfun_compo_tracesFun compo_in_Procfun)

lemma semTfun_compo_semTFun_fun:
   "[| Pf : Procfun ; QF : ProcFun |] ==>
    [[Pf o QF]]Tfun = [[Pf]]Tfun o [[QF]]TFun"
apply (simp add: expand_fun_eq)
apply (simp add: semTfun_compo_semTFun_X)
done

lemmas semTfun_compo_ProcFun = semTfun_compo_semTFun_X semTfun_compo_semTFun_fun
                               semTfun_compo_semTFun_X[simplified comp_def]
                               semTfun_compo_semTFun_fun[simplified comp_def]

lemma semTFun_compo_semTFun_X:
   "[| PF : ProcFun ; QF : ProcFun |] ==>
    [[PF o QF]]TFun X = [[PF]]TFun ([[QF]]TFun X)"
by (simp add: semT_to_traces tracesFun_compo_tracesFun compo_in_ProcFun)

(*** composition of process functions ***)

lemma semTFun_compo_semTFun_fun:
   "[| PF : ProcFun ; QF : ProcFun |] ==>
    [[PF o QF]]TFun = [[PF]]TFun o [[QF]]TFun"
apply (simp add: expand_fun_eq)
apply (simp add: semTFun_compo_semTFun_X)
done

lemmas semTFun_compo_ProcFun = semTFun_compo_semTFun_X semTFun_compo_semTFun_fun
                               semTFun_compo_semTFun_X[simplified comp_def]
                               semTFun_compo_semTFun_fun[simplified comp_def]

lemma semTFun_compo_n_X:
   "PF : ProcFun ==> [[PF ^ n]]TFun X = ([[PF]]TFun ^ n) X"
by (simp add: semT_to_traces tracesFun_compo_n_X iteration_in_ProcFun)

end
