           (*-------------------------------------------*
            |  The Dining Mathematicians in CSP-Prover  |
            |               August 2004                 |
            |             December 2004 (modified)      |
            |             November 2005 (modified)      |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory DM1_Imp_def = CSP_F:

(*****************************************************************

         1. defines Imp
         2.
         3. 
         4. 

 *****************************************************************)

(*********************************************************
                    ODD and EVEN
 *********************************************************)

consts
  EVEN :: "int => bool"
  ODD  :: "int => bool"

defs
  EVEN_def : "EVEN n == (n mod 2 = 0)"
  ODD_def  : "ODD n  == (n mod 2 = 1)"

syntax
  "_EVENs" :: "int set" ("EVENs")
  "_ODDs"   :: "int set" ("ODDs")

translations
  "EVENs"  == "Collect EVEN"
  "ODDs"   == "Collect ODD"

(*********************************************************
                         event
 *********************************************************)

datatype Event = Eat0 | Back0 | End0 | RD0 int | WR0 int
               | Eat1 | Back1 | End1 | RD1 int | WR1 int 
               | NUM int

lemma expand_Event_fun[simp]:
  "ALL Ef Eg. ((Ef::int => Event) = Eg) = (ALL n. Ef n = Eg n)"
apply (simp add: expand_fun_eq)
done
                
syntax
  "_CH0" :: "Event set" ("CH0")
  "_CH1" :: "Event set" ("CH1")
  "_OBS" :: "Event set" ("OBS")

translations
  "CH0" == "(range RD0) Un (range WR0)"
  "CH1" == "(range RD1) Un (range WR1)"
  "OBS" == "{Eat0, Back0, End0, Eat1, Back1, End1}"

(*********************************************************
                     function
 *********************************************************)

consts
  getInt :: "Event => int"

primrec
  "getInt (Eat0)   = 0"
  "getInt (Eat1)   = 0"
  "getInt (Back0)  = 0"
  "getInt (Back1)  = 0"
  "getInt (End0)   = 0"
  "getInt (End1)   = 0"
  "getInt (RD0 n)  = n"
  "getInt (RD1 n)  = n"
  "getInt (WR0 n)  = n"
  "getInt (WR1 n)  = n"
  "getInt (NUM n)  = n"

(*********************************************************
                Parallel system definition
 *********************************************************)

datatype ImpName = VAR int | TH0 | EAT0 int | TH1 | EAT1 int

consts
  ImpDef :: "(ImpName, Event) procDef"

recdef ImpDef "{}"
  "ImpDef   (X, TH0) = RD0 ? n -> IF (EVEN n)
                                  THEN Eat0  -> (X (EAT0 n))
                                  ELSE Back0 -> (X TH0)"

  "ImpDef   (X, TH1) = RD1 ? n -> IF (ODD n)
                                  THEN Eat1  -> (X (EAT1 n))
                                  ELSE Back1 -> (X TH1)"

  "ImpDef (X, (EAT0 n)) = End0 -> WR0 ! (n div 2) -> (X TH0)"

  "ImpDef (X, (EAT1 n)) = End1 -> WR1 ! (3 * n + 1) -> (X TH1)"

  "ImpDef (X, (VAR n)) = WR0 ? n -> (X (VAR n))
                   [+] WR1 ? n -> (X (VAR n))
                   [+] RD0 ! n -> (X (VAR n))
                   [+] RD1 ! n -> (X (VAR n))"

syntax
  "_ImpFun" :: "(ImpName, Event) procFun" ("ImpFun")
  "_ImpFix" :: "ImpName => Event proc"    ("ImpFix")

translations
    "ImpFun" == "curry ImpDef"
  "ImpFix n" == "(FIX ImpFun) n"

consts 
  Imp :: "int => Event proc"

defs
  Imp_def: "Imp == (%n. (ImpFix TH0 |[CH0]| ImpFix (VAR n) 
                                    |[CH1]| ImpFix TH1) -- (CH0 Un CH1))"

(*********************************************************
         To unfold "range" and "syntactic sugar", ...
 *********************************************************)

declare image_iff  [simp]
declare csp_prefix_ss_def [simp]

(*********************************************************
               nohide and guarded lemmas
 *********************************************************)

lemma Imp_ProcFun_guardFun_nohideFun[simp]:
      "ImpFun : ProcFun" "ImpFun : guardFun" "ImpFun : nohideFun"
by (simp add: anyFun, rule allI, induct_tac pn, simp_all)+

(*********************************************************
                        Lemmas
 *********************************************************)

(*** int lemmas ***)

lemma int_le_inc: "ALL (n::int) m. n <= m --> n = m | n+1 <= m"
by (auto)

lemma mod_2_not_le: "ALL (n::int). ~(2 <= n mod 2)"
by (simp add: linorder_not_le)

lemma mod_2_or: "ALL (n::int). n mod 2 = 0 | n mod 2 =1"
apply (intro allI)
apply (insert int_le_inc)
apply (drule_tac x="0" in spec)
apply (drule_tac x="n mod 2" in spec)
apply (simp)
apply (erule disjE)
apply (simp)

apply (insert int_le_inc)
apply (drule_tac x="1" in spec)
apply (drule_tac x="n mod 2" in spec)
apply (simp)
apply (erule disjE)
apply (simp)
apply (simp add: mod_2_not_le)
done

(*** ODD and EVEN lemmas ***)

lemma EVEN_not_ODD[simp]: "ALL n. EVEN n = (~ ODD n)"
apply (auto simp add: EVEN_def ODD_def )
apply (insert mod_2_or)
apply (drule_tac x="n" in spec)
apply (simp)
done

lemma ODD_add_1: "ALL n. (n mod 2 = 1) --> ((n + 1) mod 2 = (0::int))"
apply (simp add: zmod_zadd_left_eq)
done

lemma ODD_EX: "ALL m. ODD m --> (EX n. m = 2 * n + 1)"
apply (intro allI impI)
apply (simp add: ODD_def)
apply (insert ODD_add_1)
apply (drule_tac x="m" in spec)
apply (simp)
apply (simp add: zmod_eq_0_iff)
apply (erule exE)
apply (rule_tac x="q - 1" in exI)
apply (simp)
done

lemma ODD_to_EVEN[simp]: "ODD n ==> EVEN (3 * n + 1)"
apply (insert ODD_EX)
apply (drule_tac x="n" in spec)
apply (simp add: ODD_def EVEN_def)
apply (erule exE)
apply (simp add: zmod_zadd_left_eq)
apply (simp add: zmod_zmult1_eq')
done

lemma ODD_to_notODD[simp]: "ODD n ==> ~ ODD (3 * n + 1)"
apply (insert ODD_to_EVEN[of n])
apply (simp)
done

(*** range ***)

lemma fold_range: "(EX a. x = f a) = (x : range f)"
apply (auto)
done

(*********************************************************
           unfolding & folding process names
 *********************************************************)

(*** unfold VAR ***)

lemma VAR:
     "(ImpFix (VAR n))
    =F ? x:insert (RD1 n) (insert (RD0 n) ((range WR0) Un (range WR1)))
         -> IF (x : (range WR0) | x : (range WR1)) 
            THEN ImpFix (VAR (getInt x)) ELSE ImpFix (VAR n)"
apply (tactic {* cspF_unwind_left_tac 1 *})
apply (tactic {* cspF_step_left_tac 1 *})
apply (auto)
apply (tactic {* cspF_simp_tac 1 *})+
done

lemmas VAR_simp = VAR[simplified]
lemmas unfold_Imp_rules = VAR_simp
lemmas fold_Imp_rules = VAR_simp[THEN cspF_sym]

(*** unfold TH0 ***)

lemma TH0:
     "(ImpFix TH0)
   =F RD0 ? n -> IF (EVEN n) THEN Eat0  -> (ImpFix (EAT0 n))
                             ELSE Back0 -> (ImpFix TH0)"
apply (tactic {* cspF_unwind_left_tac 1 *})
done

lemmas TH0_simp = TH0[simplified]
lemmas unfold_Imp_rules = unfold_Imp_rules TH0_simp
lemmas fold_Imp_rules = fold_Imp_rules TH0_simp[THEN cspF_sym]

(*** unfold TH1 ***)

lemma TH1:
     "(ImpFix TH1)
   =F RD1 ? n -> IF (ODD n) THEN Eat1  ->  (ImpFix (EAT1 n))
                                 ELSE Back1 -> (ImpFix TH1)"
by (tactic {* cspF_unwind_left_tac 1 *})

lemmas TH1_simp = TH1[simplified]
lemmas unfold_Imp_rules = unfold_Imp_rules TH1_simp
lemmas fold_Imp_rules = fold_Imp_rules TH1_simp[THEN cspF_sym]

(*** unfold EAT0 ***)

lemma EAT0:
     "(ImpFix (EAT0 n))
   =F ? x:{End0} -> WR0 (n div 2) -> (ImpFix TH0)"
by (tactic {* cspF_unwind_left_tac 1 *})

lemmas EAT0_simp = EAT0[simplified]
lemmas unfold_Imp_rules = unfold_Imp_rules EAT0_simp
lemmas fold_Imp_rules = fold_Imp_rules EAT0_simp[THEN cspF_sym]

(*** unfold EAT1 ***)

lemma EAT1:
     "(ImpFix (EAT1 n))
   =F ? x:{End1} -> WR1 ! (3 * n + 1) -> (ImpFix TH1)"
by (tactic {* cspF_unwind_left_tac 1 *})

lemmas EAT1_simp = EAT1[simplified]
lemmas unfold_Imp_rules = unfold_Imp_rules EAT1_simp
lemmas fold_Imp_rules = fold_Imp_rules EAT1_simp[THEN cspF_sym]

end
