           (*-------------------------------------------------*
            |                 (a part of) ep2                 |
            |                  September 2004                 |
            |                   December 2004 (modified)      |
            |                   November 2005 (modified)      |
            |  Markus Roggenbach (Univ of Wales Swansea, UK)  |
            |  Yoshinao Isobe    (AIST, Japan)                |
            *-------------------------------------------------*)

theory ep2_acl = CSP_F:

(*********************************************************
          automatic unfolding syntactic-sugar
 *********************************************************)

declare csp_prefix_ss_def [simp]

(*********************************************************
              data type passed on channels
 *********************************************************)

typedecl D_SI_Init_SessionStart
typedecl D_SI_Init_SessionEnd
typedecl D_SI_Init_ConfigDataRequest
typedecl D_SI_Init_ConfigDataResponse
typedecl D_SI_Init_ConfigDataNotification
typedecl D_SI_Init_ConfigDataAcknowledge
typedecl D_SI_Init_RemoveConfigDataNotification
typedecl D_SI_Init_RemoveConfigDataAcknowledge
typedecl D_SI_Init_ActivateConfigDataNotification
typedecl D_SI_Init_ActivateConfigDataAcknowledge

datatype D_SI_Init = 
     SStart D_SI_Init_SessionStart
   | SEnd   D_SI_Init_SessionEnd
   | CDReq  D_SI_Init_ConfigDataRequest
   | CDRes  D_SI_Init_ConfigDataResponse
   | CDN    D_SI_Init_ConfigDataNotification
   | CDA    D_SI_Init_ConfigDataAcknowledge
   | RCDN   D_SI_Init_RemoveConfigDataNotification
   | RCDA   D_SI_Init_RemoveConfigDataAcknowledge
   | ACDN   D_SI_Init_ActivateConfigDataNotification
   | ACDA   D_SI_Init_ActivateConfigDataAcknowledge

typedecl TerminalState
typedecl Trigger
typedecl Message

(*********************************************************
                     event (channel)
 *********************************************************)

datatype Event = C_SI_Init D_SI_Init
               | C_TerminalDisplay Message
               | PairTT "TerminalState * Trigger"

(*********************************************************
         abstract component description level
 *********************************************************)

datatype ACName = TInit
                | TConfigurationManagement
                | AcquirerInit
                | ConfigurationManagement

consts 
  ACDef :: "(ACName, Event) procDef"

recdef ACDef "{}"
  "ACDef (X, TInit) 
    = C_SI_Init !? x: (range SStart)
      -> (X TConfigurationManagement)"

  "ACDef (X, TConfigurationManagement)
    = C_SI_Init ? x ->
       IF (x:range CDReq)
       THEN C_SI_Init !? y: (range CDRes)
            -> (X TConfigurationManagement)

       ELSE IF (x:range CDN)
       THEN C_SI_Init !? y: (range CDA)
            -> (X TConfigurationManagement)

       ELSE IF (x:range RCDN)
       THEN C_SI_Init !? y: (range RCDA)
            -> (X TConfigurationManagement)

       ELSE IF (x:range ACDN)
       THEN C_SI_Init !? y: (range ACDA)
            -> (X TConfigurationManagement)

       ELSE IF (x:range SEnd)
       THEN SKIP
       ELSE STOP"

  "ACDef (X, AcquirerInit)
    = C_SI_Init ? sessionStart: (range SStart)
      -> (X ConfigurationManagement)"

  "ACDef (X, ConfigurationManagement)
    =     C_SI_Init !? sessionEnd:(range SEnd) -> SKIP
      |~| C_SI_Init !? request:(range CDReq)
          -> C_SI_Init ? response:(range CDRes)
          -> (X ConfigurationManagement)
      |~| C_SI_Init !? notification:(range CDN)
          -> C_SI_Init ? acknowledge:(range CDA)
          -> (X ConfigurationManagement)
      |~| C_SI_Init !? notification:(range RCDN)
          -> C_SI_Init ? acknowledge:(range RCDA)
          -> (X ConfigurationManagement)
      |~| C_SI_Init !? notification:(range ACDN)
          -> C_SI_Init ? acknowledge:(range ACDA)
          -> (X ConfigurationManagement)"

syntax
  "_ACFun" :: "(ACName, Event) procFun" ("ACFun")
  "_ACFix" :: "ACName => Event proc"    ("ACFix")

translations
    "ACFun" == "curry ACDef"
  "ACFix n" == "(FIX ACFun) n"

consts 
  AC :: "Event proc"

defs
  AC_def: "AC == (ACFix TInit |[range C_SI_Init]| ACFix AcquirerInit)"

(*********************************************************
        nohide and guarded lemmas (routine work)
 *********************************************************)

lemma AC_ProcFun_guardFun_nohideFun[simp]:
      "ACFun : ProcFun"  "ACFun : guardFun"  "ACFun : nohideFun"
by (simp add: anyFun, rule allI, induct_tac pn, simp_all)+

(*********************************************************
               abstract level (deadlock free)
 *********************************************************)

syntax
  "_REQs" :: "D_SI_Init" ("REQs")
  "_RESs" :: "D_SI_Init" ("RESs")

translations
  "REQs" == "(range CDReq) Un (range CDN) Un 
             (range RCDN)  Un (range ACDN)"
  "RESs" == "(range CDRes) Un (range CDA) Un 
             (range RCDA)  Un (range ACDA)"

datatype AbsName = Abstract | Loop

consts 
  AbsDef :: "(AbsName, Event) procDef"

recdef AbsDef "{}"
  "AbsDef (X, Abstract)
          = C_SI_Init !? init:(range SStart) -> (X Loop)"

  "AbsDef (X, Loop)
          = C_SI_Init !? exit:(range SEnd) -> SKIP
            |~| 
            C_SI_Init !? request:REQs
              -> C_SI_Init !? response:RESs
                -> (X Loop)"

syntax
  "_AbsFun" :: "(AbsName, Event) procFun" ("AbsFun")
  "_AbsFix" :: "AbsName => Event proc"    ("AbsFix")

translations
    "AbsFun" == "curry AbsDef"
  "AbsFix n" == "(FIX AbsFun) n"

consts 
  Abs :: "Event proc"

defs
  Abs_def: "Abs == (AbsFix Abstract)"

(*********************************************************
        nohide and guarded lemmas (routine work)
 *********************************************************)

lemma Abs_ProcFun_guardFun_nohideFun[simp]:
      "AbsFun : ProcFun" "AbsFun : guardFun" "AbsFun : nohideFun"
by (simp add: anyFun, rule allI, induct_tac pn, simp_all)+

(*********************************************************
        relating function between AbsName and ACName
 *********************************************************)

consts 
  Abs_to_AC :: "AbsName => Event proc"

primrec
  "Abs_to_AC (Abstract)
          = (ACFix TInit |[range C_SI_Init]| ACFix AcquirerInit)"

  "Abs_to_AC (Loop)
          = (ACFix TConfigurationManagement |[range C_SI_Init]|
             ACFix ConfigurationManagement)"

(*********************************************************
           a theorem for verifying Abs <=F AC
               (i.e. AC is deadlock-free)
 *********************************************************)

theorem ep2_abs: "Abs <=F AC"
apply (simp add: Abs_def AC_def)

  (***** by fixed point induction *****)
apply (rule cspF_fp_induct_cms[of _ "Abs_to_AC"])

  (***** check guarded and no hiding operators *****)
apply (simp_all)

   (*** recursion ***)
apply (induct_tac pn)
apply (simp_all)

(*** initializing ***)

apply (tactic {* cspF_unwind_tac 1 *})
apply (tactic {* cspF_dist_tac 1 *})
apply (auto)
apply (tactic {* cspF_step_tac 1 *})
apply (rule cspF_decompo_subset)
apply (auto)
apply (tactic {* cspF_simp_tac 1 *})

(*** main loop ***)

apply (tactic {* cspF_unwind_tac 1 *})
apply (tactic {* cspF_dist_tac 1 *})+
apply (auto)

 (*** skip ***)
apply (tactic {* cspF_step_tac 1 *})
apply (rule cspF_Int_choice_left1)
apply (rule cspF_decompo_subset)
apply (auto)
apply (simp add: image_iff)
apply (tactic {* cspF_step_tac 1 *})+

 (*** C_SI_Init ? ***)
apply (rule cspF_Int_choice_left2)
apply (rule cspF_decompo_subset)
apply (auto)
apply (simp add: image_iff)
apply (tactic {* cspF_simp_tac 1 *})+

apply (tactic {* cspF_dist_tac 1 *})
apply (rule cspF_decompo_subset)
apply (auto)
apply (tactic {* cspF_step_tac 1 *})
apply (auto)
apply (tactic {* cspF_simp_tac 1 *})

 (*** CD req res ***)
apply (rule cspF_Int_choice_left2)
apply (tactic {* cspF_step_tac 1 *})
apply (rule cspF_decompo_subset)
apply (auto)
apply (simp add: image_iff)
apply (tactic {* cspF_simp_tac 1 *})+

apply (tactic {* cspF_dist_tac 1 *})
apply (rule cspF_decompo_subset)
apply (auto)
apply (tactic {* cspF_step_tac 1 *})
apply (auto)
apply (tactic {* cspF_simp_tac 1 *})

 (*** RCD ***)
apply (rule cspF_Int_choice_left2)
apply (tactic {* cspF_step_tac 1 *})
apply (rule cspF_decompo_subset)
apply (auto)
apply (simp add: image_iff)
apply (tactic {* cspF_simp_tac 1 *})+

apply (tactic {* cspF_dist_tac 1 *})
apply (rule cspF_decompo_subset)
apply (auto)
apply (tactic {* cspF_step_tac 1 *})
apply (auto)
apply (tactic {* cspF_simp_tac 1 *})

 (*** ACD ***)
apply (rule cspF_Int_choice_left2)
apply (tactic {* cspF_step_tac 1 *})
apply (rule cspF_decompo_subset)
apply (auto)
apply (simp add: image_iff)
apply (tactic {* cspF_simp_tac 1 *})+

apply (tactic {* cspF_dist_tac 1 *})
apply (rule cspF_decompo_subset)
apply (auto)
apply (tactic {* cspF_step_tac 1 *})
apply (auto)
apply (tactic {* cspF_simp_tac 1 *})
done

end
